#include "pool/Placement.h"
#include "pool/DataReferenceHandler.h"
#include "StorageSvc/pool.h"
#include "StorageSvc/DbType.h"
#include "POOLCore/Token.h"

#include "StorageSvc/Shape.h"
#include "StorageSvc/DbPrint.h"
#include "StorageSvc/DbReflex.h"
#include "StorageSvc/DbDomain.h"
#include "StorageSvc/IOODatabase.h"
#include "StorageSvc/IClassLoader.h"
#include "StorageSvc/DataCallBack.h"
#include "StorageSvc/DbObjectCallBack.h"
#include "StorageSvc/FileDescriptor.h"
#include "StorageSvc/IStorageExplorer.h"

#include "Reflex/Type.h"
#include "pool/IDataSvc.h"
#include "pool/Ref.h"
#include <stdexcept>
#include <iostream>

namespace pool  {
  void setDbLogDefaultLevel(DbPrintLvl::MsgLevel lvl);
  void setDumpLinkTables(DbPrintLvl::MsgLevel value);
  int  numTokenInstances();
  const std::string typeName(const std::type_info&);
}

namespace pool  {
  class CacheRow {
    friend class IDataSvc;

    long   key;
    Token* token;
    void*  object;
    ROOT::Reflex::Type classH;
    
  private:
    CacheRow() : key(0), token(0), object(0), classH(0) {}
  public:
    CacheRow(const CacheRow& c) {
      key    = c.key;
      classH = c.classH;
      object = c.object;
      token  = c.token;
      if ( token ) token->addRef();
    }
    CacheRow(void* _obj, const ROOT::Reflex::Type& cl, Token* _tok, long _key)
    : key(_key), token(_tok), object(_obj), classH(cl) 
    {
      if ( token ) token->addRef();
    }
    void deleteObject()   {
      if ( classH && object )   {
        classH.Destruct(object, true);
      }
      object = 0;
    }
    virtual ~CacheRow() {
      if ( token  ) {
        token->release();
      }
    }
  };

  struct DbaseDescriptors : public std::map<std::string, FileDescriptor*> {};
}
using namespace pool;

/// Standard Constructor
IDataSvc::IDataSvc() : m_storageSvc(0)   {
  m_doDeleteObjects = true;
  m_storageSvc = storageSvc();
  m_dataHandler = new DataReferenceHandler(this);
  m_files = new DbaseDescriptors;
  m_debug = 0;
}

/// Standard Destructor
IDataSvc::~IDataSvc()   {
  disconnect(0);
  if ( m_storageSvc ) {
    m_storageSvc->endSession(m_session);
  }
  deletePtr(m_files);
  deletePtr(m_dataHandler);
  releasePtr(m_storageSvc);
}

/// Enable printouts
void IDataSvc::setPrintout(long lvl)  {
  m_debug = lvl;
}

/// Initialize and return storage service object
IStorageSvc* IDataSvc::storageSvc()   {
  if ( 0 == m_storageSvc )  {
    m_storageSvc = createStorageSvc((void*)0xFeedBabe,"StorageSvc");
    if ( ! m_storageSvc ) {
      throw std::runtime_error("TestBase::storageSvc> Cannot not create a storageSvc object.");
    }
    m_storageSvc->addRef();
    DbStatus sc = m_storageSvc->startSession(UPDATE,ROOT_StorageType.type(),m_session);
    if ( !sc.isSuccess() ) {
      throw std::runtime_error("TestBase::startSession> Cannot not initialize a storageSvc object.");
    }
  }
  return m_storageSvc;
}

SessionH IDataSvc::session()   {
  storageSvc();
  return m_session;
}

/// Load class
IDataSvc::ClassH IDataSvc::loadClass(const std::string& cl_name)
{
  ClassH cl = DbReflex::forTypeName(cl_name);
  if ( !cl )  {
    DbDomainObj* dom = (DbDomainObj*)session();
    if ( dom )  {
      DbDomain domH(dom);
      if ( domH.isValid() )  {
        IClassLoader* ldr = domH.db()->classLoader();
        if ( ldr )  {
          IClassLoader::Descriptor dsc = ldr->getClass(cl_name);
          cl = dsc.first;
        }
      }
    }
  }
  return cl;
}

/// Clear the data cache
DbStatus IDataSvc::clear()   {
  int count = 0;
  DbPrint log("IDataSvc");
  log << DbPrintLvl::Debug << "There are " << m_cache.size() << " rows registered in the cache." << DbPrint::endmsg;
  log << "Number of token instances: " << numTokenInstances() << DbPrint::endmsg;
  if ( m_doDeleteObjects )  {
    for (Cache::iterator i=m_cache.begin(); i != m_cache.end(); ++i)  {
      (*i)->deleteObject();
    }
  }
  count = 0;
  log << "Deleting cache rows:[" << m_cache.size() << "]" << DbPrint::endmsg;
  log << "Number of token instances: " << numTokenInstances() << DbPrint::endmsg;
  for (Cache::iterator j=m_cache.begin(); j != m_cache.end(); ++j)  {
    delete (*j);
    //std::cout << "Deleted row # " << count++ << std::endl;
  }
  m_cache.clear();
  return DbStatus::Success;
}

/// Commit cache content  
DbStatus IDataSvc::commit(Transaction* trContext)   {
  return m_storageSvc->endTransaction(trContext, Transaction::TRANSACT_COMMIT);
}

/// Commit cache content  
DbStatus IDataSvc::rollback(Transaction* trContext)   {
  return m_storageSvc->endTransaction(trContext, Transaction::TRANSACT_ROLLBACK);
}

/// Register object for write (Internal call)
DbStatus 
IDataSvc::i_registerForWrite(Transaction* trContext, 
                                    const Placement& place,
                                    const std::type_info& typ,
                                    void* object)
{
  IStorageSvc* s = storageSvc();
  ClassH cl = DbReflex::forTypeInfo(typ);
  FileDescriptor* dsc = (*m_files)[place.dbID()];
  if ( cl && dsc )  {
    const Guid* guid = &place.shapeID();
    DataCallBack* cb = 0;
    const Shape* shape = 0;
    cb = new DbObjectCallBack(cl);
    if ( s->getShape( *dsc, cl, shape ) == IStorageSvc::SHAPE_NOT_AVAILIBLE ) {
      s->createShape( *dsc, place.contID(), cl, shape );
    }
    cb->setObject( object );
    cb->setHandler( m_dataHandler );
    if ( shape )  {
      Token* pToken = 0;
      if ( s->allocate( trContext,
                                                      *dsc,
                                                      place.contID(),
                                                      place.technology(),
                                                      shape, 
                                                      cb,
                                                      pToken ).isSuccess() ) {
        pToken->setClassID( *guid );
        m_cache.push_back(new CacheRow(object, cl, pToken, m_cache.size()));
        // We have got now a reference to the token. release the created instance
        pToken->release();
        return DbStatus::Success;
      }
    }
  }
  std::cout << "Error registerForWrite(...)" << std::endl;
  return DbStatus::Error;
}

/// Start Transaction
DbStatus IDataSvc::startTransaction(ConnectionH dbcH,TransactionH& trH)  {
  return m_storageSvc->startTransaction(dbcH, trH);
}

/// Connect to a logical database unit.
DbStatus 
IDataSvc::connect(const std::string& refDbName,long accessmode,ConnectionH& dbcH) {
  FileDescriptor* dsc = (*m_files)[refDbName];
  if ( 0 == dsc ) {
    DbStatus sc;
    dsc = new FileDescriptor(refDbName, refDbName);
    sc = m_storageSvc->existsConnection(m_session, accessmode, *dsc);
    if ( sc.value() == IStorageSvc::CONNECTION_NOT_EXISTING )
      std::cout << "Database:" << refDbName << " cannot be accessed [Database Not Existing]" << std::endl;
    else if ( sc.value() == IStorageSvc::INVALID_SESSION_TOKEN )
      std::cout << "Database:" << refDbName << " cannot be accessed [Invalid Session]" << std::endl;
    else if ( sc.isSuccess() )
      std::cout << "Database:" << refDbName << " exists..." << std::endl;

    sc = m_storageSvc->connect(m_session, accessmode, *dsc);
    if ( sc.isSuccess() )  {
      if ( accessmode & CREATE )  {
        IStorageExplorer* exp = 0;
        sc = m_storageSvc->queryInterface(IStorageExplorer::interfaceID(),(void**)&exp);
        if ( !exp ) {
          std::cout << "Cannot access storage explorer object." << std::endl;
          return Error;
        }
      }
      (*m_files)[refDbName] = dsc;
      dbcH = dsc->dbc();
      return sc;
    }
  }
  dbcH = dsc->dbc();
  return Success;
}

/// Reconnect to a logical database unit.
DbStatus IDataSvc::reconnect(ConnectionH& dbcConn,long accessmode)  {
  for (DbaseDescriptors::iterator i=m_files->begin(); i != m_files->end(); ++i)  {
    if ( (*i).second->dbc() == dbcConn )  {
      return m_storageSvc->reconnect(*((*i).second), accessmode);
    }
  }
  return Error;
}

/// Disconnect from a logical database unit.
DbStatus IDataSvc::disconnect(ConnectionH dbcH)  {
  DbStatus sc = Success;
  for (DbaseDescriptors::iterator i=m_files->begin(); i != m_files->end(); ++i)  {
    if ( (*i).second )  {
      if ( 0 == dbcH )  {
        DbStatus ir = m_storageSvc->disconnect(*((*i).second));
        if ( !ir.isSuccess() )  {
          sc = ir;
        }
        delete (*i).second;
        (*i).second = 0;
      }
      else if ( (*i).second->dbc() == dbcH )  {
        DbStatus ir = m_storageSvc->disconnect(*((*i).second));
        if ( !ir.isSuccess() )  {
          sc = ir;
        }
        delete (*i).second;
        m_files->erase(i);
        return sc;
      }
    }
  }
  if ( 0 == dbcH )  {
    m_files->clear();
  }
  return sc;
}

namespace {
  struct cast_obj  {
    virtual ~cast_obj() {};
  };
}

DbStatus IDataSvc::i_readObject( const Token* pTok,void*&  object,ClassH& clH)  {
  DbStatus sc;
  clH = 0;
  object = 0;
  if ( pTok )  {
    const Shape* shape = 0;
    const std::string& db = pTok->dbID();
    DbaseDescriptors::iterator i = m_files->find(pTok->dbID());
    if ( i == m_files->end() ) {
      ConnectionH dbcH = 0;
      sc = connect(db,READ, dbcH);
      i = m_files->find(pTok->dbID());
    }
    if ( i != m_files->end() ) {
      FileDescriptor* dsc = (*i).second;
      if ( m_storageSvc->getShape(*dsc,pTok->classID(),shape) == IStorageSvc::IS_PERSISTENT_SHAPE ) {
        ClassH cl = DbReflex::forGuid(pTok->classID());
        std::auto_ptr<DataCallBack> cb(new DbObjectCallBack(cl));
        cb->setHandler(m_dataHandler);
        sc = m_storageSvc->read(*dsc,*pTok,cb.get(),shape,clH);
        if ( sc.isSuccess() )  {
          object = (void*)cb->object();
          if ( 0 == clH )  {
            cast_obj* cast = (cast_obj*)object;
            cl = DbReflex::forTypeInfo(typeid(*cast));
            if ( cl )  {
              clH = cl;
            }
          }
          return sc;
        }
      }
    }
  }
  return Error;
}

/// Read single object according to token
DbStatus IDataSvc::find_or_load_object(Token* pTok, void*& object)  {
  std::vector<CacheRow*>::iterator i;
  Token* pNewTok = pTok;
  regToken(pTok, pNewTok);
//  if ( pTok != pNewTok ) pTok->release();
  for (i=m_cache.begin(); i != m_cache.end(); ++i)  {
    CacheRow* row = *i;
    Token* t = row->token;
    if ( *t == *pNewTok )   {
      if ( 0 == row->object )    {
        /// Token is already registered, but object not yet loaded
        ROOT::Reflex::Type clH = 0;
        DbStatus sc = i_readObject(pNewTok, object, clH);
        if ( sc.isSuccess() ) {
          row->object = object;
          row->classH = clH;
        }
        return sc;
      }
      // Token and object already present
      object = row->object;
      return DbStatus::Success;
    }
  }
  /*
   *   Neither token nor object known to cache:
   */
  ROOT::Reflex::Type clH;
  DbStatus sc = i_readObject(pTok, object, clH);
  if ( sc.isSuccess() ) {
    m_cache.push_back(new CacheRow(object,clH,pTok,m_cache.size()));
    return sc;
  }
  return DbStatus::Error;
}

/// Update object
DbStatus IDataSvc::i_update(Transaction*          trContext,
                            const std::type_info& typ,
                            void*                 pObj)
{
  Token* pTok = 0;
  DbStatus sc = findToken(pObj, pTok);
  if ( sc.isSuccess() )  {
    IStorageSvc* s = storageSvc();
    ClassH cl = DbReflex::forTypeInfo(typ);
    FileDescriptor* dsc = (*m_files)[pTok->dbID()];
    if ( cl && dsc )  {
      // We do not do any transformations here!
      const Guid& guid    = pTok->classID();
      const Guid& shapeID = guid;
      DataCallBack* cb = 0;
      cb = new DbObjectCallBack(cl);
      cb->setObject( pObj );
      cb->setHandler( m_dataHandler );
      const Shape* shape = 0;
      sc = s->getShape(*dsc,shapeID,shape);
      if(sc == IStorageSvc::IS_PERSISTENT_SHAPE) {
        sc = s->update(trContext,*dsc,shape,cb,*pTok);
        if ( sc.isSuccess() ) {
          return sc;
        }
      }
      sc = Error;
    }
  }
  return sc;
}

/// Update object
DbStatus IDataSvc::i_destroy(Transaction*          trContext,
                             const std::type_info& /* typ */,
                             void*                 pObj)
{
  std::vector<CacheRow*>::iterator i;
  for (i=m_cache.begin(); i != m_cache.end(); ++i)  {
    void* object = (*i)->object;
    if ( object == pObj )   {
      Token* pTok = (*i)->token;
      IStorageSvc* s = storageSvc();
      FileDescriptor* dsc = (*m_files)[pTok->dbID()];
      if ( dsc )  {
        DbStatus sc = s->destroy(trContext, *dsc, *pTok);
        if ( sc.isSuccess() ) {
          // Should delete and remove the object from the cache....
          (*i)->deleteObject();
          return sc;
        }
      }
      return Error;
    }
  }
  return Error;
}

/// Find an existing object in the cache
DbStatus IDataSvc::findToken(const void* pObj, Token*& pTok)    {
  std::vector<CacheRow*>::iterator i;
  for (i=m_cache.begin(); i != m_cache.end(); ++i)  {
    void* object = (*i)->object;
    if ( object == pObj )   {
      pTok = (*i)->token;
      return DbStatus::Success;
    }
  }
  pTok = 0;
  return DbStatus::Error;
}

/// Find an existing object in the cache
DbStatus IDataSvc::regToken(Token* pTok, Token*& refpTok)   {
  std::vector<CacheRow*>::iterator i;
  for (i=m_cache.begin(); i != m_cache.end() && 0 != pTok; ++i)  {
    Token* p = (*i)->token;
    if ( *p == *pTok )   {
      refpTok = p;
      return DbStatus::Success;
    }
  }
  DbPrint log("IDataSvc");
  log << DbPrintLvl::Debug << "Push token:" << pTok->toString() << DbPrint::endmsg;
  m_cache.push_back(new CacheRow(0,0,pTok,m_cache.size()));
  refpTok = pTok;
  return DbStatus::Success;
}
