#include "AttributeList/AttributeList.h"
#include "AttributeList/AttributeExceptions.h"
#include "AttributeValueConstructor.h"
#include "CoralBase/MessageStream.h"
#include <ostream>

static pool::AttributeListSpecification empty_spec;

pool::AttributeList::AttributeList() : m_attr_spec(&empty_spec)
{}

pool::AttributeList::~AttributeList()
{}


pool::AttributeList::AttributeList(const pool::AttributeListSpecification& spec)
{
  initialize(spec);
} 


pool::AttributeList::AttributeList(boost::shared_ptr<const AttributeListSpecification> ptr) : m_attr_spec_ptr(ptr)
{
  assert(ptr.get());
  initialize(*ptr.get());
} 

void pool::AttributeList::initialize(const pool::AttributeListSpecification& spec)
{
  m_attr_spec = &spec;

  for(pool::AttributeListSpecification::const_iterator it=spec.begin(); it!=spec.end(); ++it)
  {
    boost::shared_ptr<pool::AttributeBaseValue> pv(pool::AttributeImplFactory::theFactory().createAttributeImpl(it->type_name()));
    m_attr_vector.push_back( pool::AttributeValueConstructor(pv, *it) );
  }
  
}

pool::AttributeList::AttributeList(const AttributeList& rhs)
{
  if( this == &rhs) return;

  m_attr_spec = rhs.m_attr_spec;

  for(const_iterator it=rhs.begin(); it!=rhs.end(); ++it)
  {
    m_attr_vector.push_back( Attribute(*it) );
  }
}

const pool::Attribute& pool::AttributeList::operator[](std::string name) const
{
  size_t idx = m_attr_spec->index(name);
  if( idx == pool::IDX_NOT_FOUND)
    throw pool::attribute_not_found(name);
  else
    return m_attr_vector[idx];    
}

pool::Attribute& pool::AttributeList::operator[](std::string name)
{
  size_t idx = m_attr_spec->index(name);
  if( idx == pool::IDX_NOT_FOUND)
    throw pool::attribute_not_found(name);
  else
    return m_attr_vector[idx]; 
}

int pool::AttributeList::size() const
{
  return m_attr_vector.size();
}

pool::AttributeList::const_iterator pool::AttributeList::begin() const
{
  return m_attr_vector.begin();
}

pool::AttributeList::const_iterator pool::AttributeList::end() const
{
  return m_attr_vector.end();
}

pool::AttributeList::iterator pool::AttributeList::begin()
{
  return m_attr_vector.begin();
}

pool::AttributeList::iterator pool::AttributeList::end()
{
  return m_attr_vector.end();
}

const pool::AttributeListSpecification&
pool::AttributeList::attributeListSpecification() const
{
  return *m_attr_spec;
}

bool pool::AttributeList::operator == (const AttributeList& rhs) const
{
  const AttributeListSpecification& spec = attributeListSpecification();

  if( spec != rhs.attributeListSpecification()) return false;

  try
  {
    for(AttributeListSpecification::const_ordered_index_iterator it = spec.ordered_index_begin(); 
        it != spec.ordered_index_end(); ++it)
    {
      const AttributeSpecification& at_spec = spec[it->second];

      const Attribute& at = (*this)[at_spec.name()];
      const Attribute& ar = rhs[at_spec.name()];

      if(at != ar)
        return false;
    }
  }
  catch(const std::exception&x) {
    coral::MessageStream msg_stream( "AttributeList" );
    msg_stream << coral::Error << "unexpected exception, Internal Error \n" <<  x.what() << coral::MessageStream::endmsg;
    return false;	
  }

  return true;
}

bool pool::AttributeList::operator != (const AttributeList& rhs) const
{
  return ! ((*this) == rhs);
}

void  pool::AttributeList::print(std::ostream& os) const
{
  bool first = true;

  os << "attribute list: size = " << m_attr_vector.size() << " [ ";

  for(const_iterator it=begin(); it!=end(); ++it)
  {
    if(!first) 
      os << ", ";

    it->print(os); os << " as ";  it->spec().print(os); 
    first = false;
  }
  os << " ]";
}
