#include "CollectionBase/TokenList.h"
#include "CollectionBase/TokenListSpecification.h"

#include "POOLCore/Token.h"
#include "POOLCore/Exception.h"

#include <ostream>
#include <sstream>


pool::TokenList::TokenList()
  : m_tokenListSpec( new pool::TokenListSpecification ),
    m_ownTokenListSpec( true ),
    m_tokenForName(),
    m_tokenForIndex()
{}


pool::TokenList::TokenList( const pool::TokenListSpecification& tokenListSpec,
                            bool sharedTokenSpec )
  : m_tokenListSpec( sharedTokenSpec ? const_cast< pool::TokenListSpecification* >( &tokenListSpec ) : new pool::TokenListSpecification( tokenListSpec ) ),
    m_ownTokenListSpec( !sharedTokenSpec ),
    m_tokenForName(),
    m_tokenForIndex()
{
  if ( sharedTokenSpec ) m_tokenListSpec->addRef();

  for ( pool::TokenListSpecification::const_iterator iTokenName = m_tokenListSpec->begin();
    iTokenName != m_tokenListSpec->end(); ++iTokenName )
  {
    pool::Token* token = new pool::Token;
    m_tokenForName.insert( std::make_pair( *iTokenName, token ) );
    int index = static_cast<int>( this->size() - 1 );
    m_tokenForIndex.insert( std::make_pair( index, token ) );
  } 
}


pool::TokenList::TokenList( const pool::TokenList& rhs )
  : m_tokenListSpec( rhs.m_ownTokenListSpec ? new pool::TokenListSpecification( *( rhs.m_tokenListSpec ) ) : rhs.m_tokenListSpec ),
    m_ownTokenListSpec( rhs.m_ownTokenListSpec ),
    m_tokenForName(),
    m_tokenForIndex()
{
  if ( ! m_ownTokenListSpec ) m_tokenListSpec->addRef();

  // Copy the metadata Tokens.
  pool::TokenListSpecification::const_iterator iTokenName = m_tokenListSpec->begin();
  for ( std::map< std::string, pool::Token* >::const_iterator iData = rhs.m_tokenForName.begin();
    iData != rhs.m_tokenForName.end(); ++iData, ++iTokenName ) 
  {
    pool::Token* token = new pool::Token;
    iData->second->setData( token );
    m_tokenForName.insert( std::make_pair( *iTokenName, token ) );
    int index = static_cast<int>( this->size() - 1 );
    m_tokenForIndex.insert( std::make_pair( index, token ) );
  }
}


pool::TokenList::~TokenList()
{
  for ( std::map< std::string, pool::Token* >::iterator iData = m_tokenForName.begin();
    iData != m_tokenForName.end(); ++iData ) 
  {
    iData->second->release();
  }

  if ( m_ownTokenListSpec )
  {
    delete m_tokenListSpec;
    m_tokenListSpec = 0;
  }
}


pool::TokenList&
pool::TokenList::operator=( const pool::TokenList& rhs )
{
  // Clear all data.
  for ( std::map< std::string, pool::Token* >::iterator iData = m_tokenForName.begin();
    iData != m_tokenForName.end(); ++iData )
  {
    iData->second->release();
  }
  m_tokenForName.clear();
  m_tokenForIndex.clear();
  m_tokenListSpec->release();

  // Set the Token names specification.
  m_tokenListSpec = rhs.m_ownTokenListSpec ? new pool::TokenListSpecification( *( rhs.m_tokenListSpec ) ) : rhs.m_tokenListSpec;
  m_ownTokenListSpec = rhs.m_ownTokenListSpec;
  if ( ! m_ownTokenListSpec ) m_tokenListSpec->addRef();

  // Copy the metadata Tokens.
  pool::TokenListSpecification::const_iterator iTokenName = m_tokenListSpec->begin();
  for ( std::map< std::string, pool::Token* >::const_iterator iData = rhs.m_tokenForName.begin();
    iData != rhs.m_tokenForName.end(); ++iData, ++iTokenName ) 
  {
    pool::Token* token = new pool::Token;
    iData->second->setData( token );
    m_tokenForName.insert( std::make_pair( *iTokenName, token ) );
    int index = static_cast<int>( this->size() - 1 );
    m_tokenForIndex.insert( std::make_pair( index, token ) );
  }

  return *this;
}


bool
pool::TokenList::operator==( const pool::TokenList& rhs ) const
{
  if ( this->size() != rhs.size() ) return false;
  for ( size_t i = 0; i < this->size(); ++i )
    if ( (*this)[i] != rhs[i] )
      return false;
  return true;
}


void
pool::TokenList::extend( const std::string& name )
{
  // If Token name specification is shared make a copy of it and the associated data.
  if ( ! m_ownTokenListSpec ) 
  {
    pool::TokenListSpecification* newTokenSpec = new pool::TokenListSpecification( *m_tokenListSpec );
    m_tokenListSpec->release();
    m_tokenListSpec = newTokenSpec;
    m_ownTokenListSpec = true;
    for ( std::map< std::string, pool::Token* >::iterator iData = m_tokenForName.begin();
      iData != m_tokenForName.end(); ++iData ) 
    {
      pool::Token* token = new pool::Token;
      iData->second->setData( token );
      iData->second->release();
      token->setData( iData->second );
    }
  }

  m_tokenListSpec->extend( name );

  pool::Token* token = new pool::Token;
  m_tokenForName.insert( std::make_pair( name, token ) );
  int index = static_cast<int>( this->size() - 1 );
  m_tokenForIndex.insert( std::make_pair( index, token ) );
}


void
pool::TokenList::copyData( const pool::TokenList& rhs )
{
  // Copy the metadata Tokens.
  std::map< std::string, pool::Token* >::const_iterator iDataRhs = rhs.m_tokenForName.begin();
  for ( std::map< std::string, pool::Token* >::iterator iData = m_tokenForName.begin();
    iData != m_tokenForName.end(); ++iData, ++iDataRhs )
  {
    iDataRhs->second->setData( iData->second );
  }
}


pool::TokenList&
pool::TokenList::merge( pool::TokenList& rhs )
{
  for ( std::map< std::string, pool::Token* >::iterator iDataRhs = rhs.m_tokenForName.begin();
        iDataRhs != rhs.m_tokenForName.end(); ++iDataRhs )
  {
    const std::string tokenNameRhs = iDataRhs->first; 

    bool found = false;
    for ( std::map< std::string, pool::Token* >::iterator iData = m_tokenForName.begin();
      iData != m_tokenForName.end(); ++iData ) 
    {
      if ( iData->first == tokenNameRhs ) 
      {
        m_tokenForName.insert( std::make_pair( tokenNameRhs, iDataRhs->second ) );
        found = true;
        break;
      }
    }
    if ( ! found ) 
    {
      this->extend( tokenNameRhs );
      m_tokenForName.insert( std::make_pair( tokenNameRhs, iDataRhs->second ) );
    }
  }

  return *this;
}




pool::Token&
pool::TokenList::operator[]( const std::string& name )
{
   return const_cast<Token&>( ((const TokenList&)*this)[name] );
}


pool::Token&
pool::TokenList::operator[]( unsigned int index )
{
   return const_cast<Token&>( ((const TokenList&)*this)[index] );
}


const pool::Token&
pool::TokenList::operator[]( const std::string& name ) const
{
   std::map< std::string, pool::Token* >::const_iterator iData = m_tokenForName.find( name );

   if( iData == m_tokenForName.end() )   {
      std::string errorMsg = "Cannot find Token with name `" + name + "' in Token list.";
      throw pool::Exception( errorMsg,
			     "TokenList::operator[](name)",
			     "CollectionBase" );
   }
   return *( iData->second );
}


const pool::Token&
pool::TokenList::operator[]( unsigned int index ) const
{
   std::map< int, pool::Token* >::const_iterator iData = m_tokenForIndex.find( index );

   if( iData == m_tokenForIndex.end() ) {
      std::ostringstream errorMsg;
      errorMsg << "Cannot find Token with index `" << index << "' in Token list.";
      throw pool::Exception( errorMsg.str(),
			     "TokenList::operator[](index)",
			     "CollectionBase" );
   }
   return *( iData->second );
}



std::ostream&
pool::TokenList::toOutputStream( std::ostream& os ) const
{  
  for ( pool::TokenListSpecification::const_iterator iTokenName = m_tokenListSpec->begin();
        iTokenName != m_tokenListSpec->end(); ++iTokenName )
  {
    pool::Token token( ( m_tokenForName.find( *iTokenName ) )->second );
    std::string tokenAsString = token.toString();
    if ( iTokenName != m_tokenListSpec->begin() ) os << ", ";
    os << "[";
    os << *iTokenName << " (pool::Token) : " << tokenAsString;
    os << "]";
  }

  return os;
}

