#ifndef POOL_ICACHESVC_H
#define POOL_ICACHESVC_H

#include "RVersion.h"

#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
namespace ROOT { namespace Reflex {  class Type;  } }
#else
namespace Reflex {  class Type;  }
#endif

namespace pool    {

  class Token;
  class CacheEntry;
  class ObjectDeletePolicy;
  class Placement;
  class IPersistencySvc;
  class IRegisteredRef;
  class SharedAnyPtr;
 
  /** @class ICacheSvc ICacheSvc.h 
      Abstract interface for retrieving/registering 
      objects with the pool storage manager.

      @author Giacomo Govi
      @date 15/07/2002
  */
  class ICacheSvc  {

    public:

    // virtual destructor
    virtual ~ICacheSvc(){}

    // retrieve the object pointer from a given cache line
    virtual const SharedAnyPtr& getObject(CacheEntry* aCacheEntry) = 0;
    
    // reset (delete) the object pointer of a specified cache line 
    virtual bool resetObject(CacheEntry* aCacheEntry)=0;
  
    // check if the object in a specified cache line has been loaded
    virtual bool isObjectLoaded(const CacheEntry* aCacheEntry) = 0;

    // retrieve the token from a given cache line
    virtual Token* getToken(CacheEntry* aCacheEntry) = 0;

    // register a cacheline to be submitted to write procedure
    virtual bool registerForWrite(const Placement& place,
				      CacheEntry* aCacheEntry) = 0;

    // register a cacheline to be submitted to update procedure
    virtual bool registerForUpdate(CacheEntry* aCacheEntry) = 0;

    // register a cacheline to be submitted to delete procedure
    virtual bool registerForDelete(CacheEntry* aCacheEntry) = 0;

    // set up a reference to the specified cache line
    virtual bool linkCacheEntry(CacheEntry* aCacheEntry)=0;

    // set up a reference to the specified cache line, with type checking
#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
    virtual bool linkCacheEntry(CacheEntry* aCacheEntry,
				const ROOT::Reflex::Type& clientClass)=0;
#else
    virtual bool linkCacheEntry(CacheEntry* aCacheEntry,
				const Reflex::Type& clientClass)=0;
#endif

    // remove the reference to the specified cache line
    virtual bool unlinkCacheEntry(CacheEntry* aCacheEntry)=0;

    // retrieve the cache line for a specified token. 
    virtual CacheEntry* registerToken(const Token& aToken) = 0;

    // retrieve the cache line for a specified object pointer. 
    virtual CacheEntry* registerObject( const SharedAnyPtr& aPtr)=0;

    // add a specified cache line in the cache
    virtual bool registerCacheEntry( CacheEntry *aCacheEntry) = 0; 

    // retrieve the current delete policy for the objects
    virtual const ObjectDeletePolicy& objectDeletePolicy() = 0;

    // clear the cache content
    virtual void clearCache() = 0;

    // delete the objects from all the cache entries
    virtual void resetCache() = 0;

    // retrive the number of cache entry elements
    virtual unsigned int cacheSize() = 0;

    // returns the number of cache entries marked (write,update or delete)
    virtual unsigned int numberOfMarkedObjects()= 0;

    // register a client 
    virtual void registerRef( IRegisteredRef* ref ) = 0;

    // unregister a client 
    virtual void unregisterRef( IRegisteredRef* ref ) = 0;
 
  protected:

    ICacheSvc(){}
    ICacheSvc(const ICacheSvc&);
    ICacheSvc& operator=(const ICacheSvc&);


  };   // class ICacheSvc
}      // namespace pool

#endif  // POOL_ICACHESVC_H












