#define POOL_CACHESVC_CPP 1

#include <vector>
#include "CacheSvc.h"
#include "Cache.h"
#include "CacheEntryImp.h"
#include "ReferenceHandler.h"
#include "PersistencySvc/IPersistencySvc.h"
#include "DataSvc/AnyType.h"
#include "DataSvc/CacheSvcException.h"
#include "DataSvc/IRegisteredRef.h"

pool::CacheSvc::CacheSvc(IPersistencySvc* aPersSvc, const ObjectDeletePolicy& aDeletePolicy ):
  m_deletePolicy( aDeletePolicy ),
  m_cache( new Cache(m_deletePolicy) ),
  m_persSvc( aPersSvc ),
  m_refHandler( 0 ),
  m_refs()
{ 
  m_refHandler = new ReferenceHandler(this);
  if(!m_persSvc){
    delete m_cache;
    m_cache = 0;
    delete m_refHandler;
    m_refHandler = 0;
    throw CacheSvcException(
          "Cannot construct CacheSvc: null PersistencySvc pointer");
  }
}

pool::CacheSvc::~CacheSvc() { 
  delete m_refHandler;
  clearCache();
  for ( std::set< pool::IRegisteredRef* >::iterator iRef = m_refs.begin();
	iRef != m_refs.end(); ++iRef ) {
    (*iRef)->unregisterCacheSvc( this );
  }
  delete m_cache;
}

const pool::SharedAnyPtr& pool::CacheSvc::getObject(CacheEntry* anEntry){
  CacheEntryImp* aCacheEntry = dynamic_cast<CacheEntryImp*>(anEntry); 
  if(!aCacheEntry) {
    throw CacheSvcException(
          "Cannot resolve the specified cache line");
  }
  const SharedAnyPtr& theObject = aCacheEntry->object();
  if( 0 == theObject.get() && !aCacheEntry->isMarkedWrite() ){
    _retrieveObject(aCacheEntry);
  }
  return theObject;
}

bool pool::CacheSvc::resetObject(CacheEntry* anEntry){
  if(!m_cache){
    throw CacheSvcException("CacheSvc has not been fully constructed");
  }
  bool ret = false;
  CacheEntryImp* aCacheEntry = dynamic_cast<CacheEntryImp*>(anEntry);
  if(aCacheEntry){
    bool marked = aCacheEntry->isMarkedWrite() ||
                  aCacheEntry->isMarkedUpdate()||
                  aCacheEntry->isMarkedDelete();
    if(!marked){
      m_cache->deleteObject(aCacheEntry);
    }
    //else{
    //  cannot be handle yet: the persistency svc should provide a rollback
    //}
    ret = true;
  }
  return ret;
}

bool pool::CacheSvc::isObjectLoaded(const CacheEntry* anEntry){
  bool ret = false;
  const CacheEntryImp* aCacheEntry = dynamic_cast<const CacheEntryImp*>(anEntry);
  if(aCacheEntry){
    ret =  aCacheEntry->object().get() != 0;
  }
  return ret;
}

pool::Token* pool::CacheSvc::getToken(CacheEntry* anEntry){
  Token *tok = 0;
  CacheEntryImp* aCacheEntry = dynamic_cast<CacheEntryImp*>(anEntry);
  if(aCacheEntry){
    tok = const_cast<Token*>(aCacheEntry->token());
  }
  return tok;
}

bool pool::CacheSvc::registerForWrite(const Placement& place,
					     CacheEntry* anEntry){
  if(!m_persSvc || !m_cache){
    throw CacheSvcException("CacheSvc has not been fully constructed");
  }
  CacheEntryImp* aCacheEntry = dynamic_cast<CacheEntryImp*>(anEntry);
  bool ret = false;
  if(aCacheEntry){
    bool already_marked = aCacheEntry->isMarkedWrite();
    if(!already_marked && !aCacheEntry->token()){  
      const void* objectPtr = aCacheEntry->object().get();
     const TypeH& cl =  aCacheEntry->object().get().getClass();
      if(!cl){
        std::ostringstream os;
        os << "Class "
           << aCacheEntry->object().get().getType().name()
           << " not found in the dictionary";
        throw CacheSvcException(os.str()); 
      }
      if(objectPtr){
        Token *token = 
           m_persSvc->registerForWrite(place, objectPtr, cl, m_refHandler);
        if(token){
          if(m_cache->updateEntry(aCacheEntry,*token)){
            ret = true;
            aCacheEntry->setMarkWrite();
          } 
          token->release();
        }
      } 
    } else if(already_marked){
      ret = true;
    }
  }
  return ret;
}

bool pool::CacheSvc::registerForUpdate(CacheEntry* anEntry){
  if(!m_persSvc || !m_cache){
    throw CacheSvcException("CacheSvc has not been fully constructed");
  }
  CacheEntryImp* aCacheEntry = dynamic_cast<CacheEntryImp*>(anEntry);
  bool ret = false;
  if(aCacheEntry){
    bool already_marked = aCacheEntry->isMarkedWrite() ||
                        aCacheEntry->isMarkedUpdate()||
                        aCacheEntry->isMarkedDelete();
    ret = true;
    if(!already_marked){
      const Token* tok = aCacheEntry->token();
      if(tok){
        if( 0 == aCacheEntry->object().get()){
          ret = _retrieveObject(aCacheEntry);
        }
        if(ret){
          const void* objectPtr = aCacheEntry->object().get();
          const TypeH& objectClass =  aCacheEntry->object().get().getClass();
          if(!objectClass){
            std::ostringstream os;
            os << "Class "
               << aCacheEntry->object().get().getType().name()
               << " not found in the dictionary";
            throw CacheSvcException(os.str()); 
          }
          ret = m_persSvc->updateObject(objectPtr,objectClass,
					  *tok,m_refHandler);
          if(ret) {
            aCacheEntry->setMarkUpdate();
          }
        }
      } else {
        ret = false;
      }
    } 
  }
  return ret;
}

bool pool::CacheSvc::registerForDelete(CacheEntry* anEntry){
  if(!m_persSvc || !m_cache){
    throw CacheSvcException("CacheSvc has not been fully constructed");
  }
  CacheEntryImp* aCacheEntry = dynamic_cast<CacheEntryImp*>(anEntry);
  bool ret = false;
  if(aCacheEntry){
    bool already_marked = aCacheEntry->isMarkedWrite() ||
                        aCacheEntry->isMarkedUpdate()||
                        aCacheEntry->isMarkedDelete();
    bool ret = true;
    if(!already_marked){
      const Token* tok = aCacheEntry->token();
      if(tok){
	ret = m_persSvc->destroyObject(*tok);
        if(ret) {
          aCacheEntry->setMarkDelete();
        }
      } else {
        ret = false;
      }
    }
  }
  return ret;
}

bool pool::CacheSvc::resetMarkedObjects(){
  if(!m_cache){
    throw CacheSvcException("CacheSvc has not been fully constructed");
  }
  // list all the cache entries 
  ICache::const_token_iterator iter = m_cache->token_iter_begin();
  std::vector<CacheEntryImp*> entryList;
  std::vector<CacheEntryImp*> deleteList;
  while(iter != m_cache->token_iter_end()){
    CacheEntryImp* entry = (*iter).second;
    // select objects marked for write
    if(entry){
      if(!entry->isMarkedDelete()){
        entryList.push_back(entry);
      }
      if(entry->isMarkedDelete()){
        deleteList.push_back(entry);
      }
    }
    iter++;
  }
  std::vector<CacheEntryImp*>::const_iterator vi = entryList.begin();
  bool removeOk = true;
  while(vi!=entryList.end() && removeOk){
    CacheEntryImp* entry = (*vi);
    if(m_cache->findEntry(entry)){
      // delete the entry if not referenced by any Ref
      if(entry->isMarked()){
        entry->unMark();   
      }
      if(entry->refCount()==0){
        removeOk = _removeCacheEntry(entry);
      }       
      // reset object after commit - temporarely disabled - required for RDMS
      //} else {
      //  if(entry->object()){
      //    // remove the (unique) entry in the mapping pointer->token, if any! 
      //    m_cache->removeObject(entry);
      // 
      //
      //  }
      //}
    }
    vi++;          
  }
  vi = deleteList.begin();
  while(vi!=deleteList.end() && removeOk){
    CacheEntryImp* entry = (*vi);
    if(m_cache->findEntry(entry)){
      // delete the entry if not referenced by any Ref
      entry->unMark();
      if(entry->refCount()==0){
        removeOk = _removeCacheEntry(entry);
      } else {
        for ( std::set<pool::IRegisteredRef*>::iterator iRef = m_refs.begin();
	    iRef != m_refs.end(); ++iRef ) {
          (*iRef)->unregisterCacheEntry( entry );
        }
        m_cache->removeEntry(entry);
      }
    }
    vi++;          
  }
  return removeOk;
}

bool pool::CacheSvc::unmarkObjects(){
  if(!m_cache){
    throw CacheSvcException("CacheSvc has not been fully constructed");
  }
  // list all the cache entries 
  ICache::const_token_iterator iter = m_cache->token_iter_begin();
  std::vector<CacheEntryImp*> markedList;
  std::vector<CacheEntryImp*> markedWriteList;
  while(iter != m_cache->token_iter_end()){
    CacheEntryImp* entry = (*iter).second;
    // select objects marked for write
    if(entry){
      bool marked = entry->isMarkedUpdate() ||
	            entry->isMarkedDelete();
      if(marked){
        markedList.push_back(entry);
      }
      if(entry->isMarkedWrite()){
        markedWriteList.push_back(entry);
      }
    }
    iter++;
  }
  bool removeOk = true;
  std::vector<CacheEntryImp*>::const_iterator vi = markedList.begin();
  while(vi!=markedList.end() && removeOk){
    CacheEntryImp* entry = (*vi);
    if(entry->refCount()==0){
      removeOk = _removeCacheEntry(entry);   
    } else {
      entry->setMarkUpdate(false);
      entry->setMarkDelete(false);
    }
    vi++;          
  }
  vi = markedWriteList.begin();
  while(vi!=markedWriteList.end() && removeOk){
    CacheEntryImp* entry = (*vi);
    if(entry->refCount()==0){
       removeOk = _removeCacheEntry(entry);
    } else {
      entry->setMarkWrite(false);
      removeOk = m_cache->removeToken(entry);
    }    
    vi++;          
  }
  return removeOk;
}

bool pool::CacheSvc::linkCacheEntry(CacheEntry* anEntry){
  CacheEntryImp* aCacheEntry = dynamic_cast<CacheEntryImp*>(anEntry);
  bool ret = false;
  if(aCacheEntry){ 
    aCacheEntry->incrementCount();
    ret = true;
  }
  return ret;
}

bool pool::CacheSvc::linkCacheEntry(CacheEntry* anEntry,
                                   const TypeH& clientClass){
  CacheEntryImp* aCacheEntry = dynamic_cast<CacheEntryImp*>(anEntry);
  bool ret = false;
  if(aCacheEntry && clientClass) {
    AnyType theType(clientClass);
    const SharedAnyPtr& obj = aCacheEntry->object();
    AnyType cacheEntryType = obj.get().getType();
    if(!obj){
      aCacheEntry->incrementCount();   
      ret =  true;
    } else {
      if(theType.isBaseFor(cacheEntryType)){
        aCacheEntry->incrementCount();   
        ret =  true;
      }
    }  
  }
  return ret;
}

bool pool::CacheSvc::unlinkCacheEntry(CacheEntry* anEntry){
  bool ret  = false;
  if(!m_cache){
    throw CacheSvcException("CacheSvc has not been fully constructed");
  }
  CacheEntryImp* aCacheEntry = dynamic_cast<CacheEntryImp*>(anEntry);
  if(aCacheEntry){
    if(m_cache->findEntry(aCacheEntry)){
      aCacheEntry->decrementCount();
      ret = true;
      if( aCacheEntry->refCount() == 0) {
        ret = _removeCacheEntry(aCacheEntry);
      }
    }
  }
  return ret;
}

pool::CacheEntry* pool::CacheSvc::registerToken(const Token& aToken) {
  if(!m_cache){
    throw CacheSvcException("CacheSvc has not been fully constructed");
  }
  return m_cache->getEntry(aToken);
}

pool::CacheEntry* pool::CacheSvc::registerObject( const SharedAnyPtr& anObjPtr ) {
  if(!m_cache){
    throw CacheSvcException("CacheSvc has not been fully constructed");
  }
  CacheEntryImp* theCacheEntry = 0;
  if(anObjPtr){ 
    theCacheEntry = m_cache->getEntry(anObjPtr);
  }
  return theCacheEntry;
}

bool pool::CacheSvc::registerCacheEntry(CacheEntry* anEntry){
  if(!m_cache){
    throw CacheSvcException("CacheSvc has not been fully constructed");
  }
  CacheEntryImp* aCacheEntry = dynamic_cast<CacheEntryImp*>(anEntry);
  bool ret = false;
  if(aCacheEntry) {
    ret =  m_cache->addEntry(aCacheEntry);
  }
  return ret;
}

const pool::ObjectDeletePolicy& pool::CacheSvc::objectDeletePolicy(){
  return m_deletePolicy;
}

void pool::CacheSvc::clearCache(){
  if(!m_cache){
    throw CacheSvcException("CacheSvc has not been fully constructed");
  }

  typedef std::vector<CacheEntryImp*> CacheEntries;
  CacheEntries entries;
  Cache::const_list_iterator t_i;
  CacheEntries::iterator i;

    /**
      * It is necessary to increment any valid reference to a chacheline,
      * because otherwise the "reset" may trigger internally implicit cache line
      * removals. Such implicit removals would be fatal, because no
      * iterator is defined over a changing STL container.
      * Hence:
      *   1) Increase ref count for all cache entries knwon by their token
      *   2) Increase ref count for all cache entries knwon by their object
      *      We do not care about entries in both sets
      *   3) Clear the cache table
      *   4) Reset all known cache entries.
      *      Note: This call may be called any number times for the same
      *            cache line
      *   5) Decrement the reference counters and remove all cacheentries
      *      no longer referenced.
      */
  for (t_i=m_cache->list_begin();
                                      t_i!=m_cache->list_end();t_i++){
    CacheEntryImp *entry = const_cast<CacheEntryImp*>(
                           static_cast<const CacheEntryImp*>((*t_i).thePtr));      
    entry->incrementCount();
    entries.push_back(entry);
  }
   
  for (i=entries.begin();i!=entries.end();++i){
    (*i)->reset();
  }

  for (i=entries.begin();i!=entries.end();++i){
    for ( std::set< pool::IRegisteredRef* >::iterator iRef = m_refs.begin();
          iRef != m_refs.end(); ++iRef ) {
      (*iRef)->unregisterCacheEntry( *i );
    }
    delete *i;
  }
        
  m_cache->clearTable();
  
}

void pool::CacheSvc::resetCache(){
  if(!m_cache){
    throw CacheSvcException("CacheSvc has not been fully constructed");
  }

  typedef std::vector<CacheEntryImp*> CacheEntries;
  CacheEntries entries;
  Cache::const_list_iterator t_i;
  CacheEntries::iterator i;

  for (t_i=m_cache->list_begin(); t_i!=m_cache->list_end();t_i++){
    CacheEntryImp *entry = const_cast<CacheEntryImp*>(
                           static_cast<const CacheEntryImp*>((*t_i).thePtr)); 
    entry->incrementCount();     
    entries.push_back(entry);

  }

  for (i=entries.begin();i!=entries.end();++i){
    resetObject(*i);
  }
  for (i=entries.begin();i!=entries.end();++i){
    (*i)->decrementCount();
    if ( 0 == (*i)->refCount() ) {
      m_cache->removeEntry(*i);
    }
  }
 
  
}

unsigned int pool::CacheSvc::cacheSize(){
  if(!m_cache){
    throw CacheSvcException("CacheSvc has not been fully constructed");
  }

  return m_cache->cacheSize();
}

unsigned int pool::CacheSvc::numberOfMarkedObjects(){
  if(!m_cache){
    throw CacheSvcException("CacheSvc has not been fully constructed");
  }

  Cache::const_list_iterator t_i;
  unsigned int numberOfMarkedObjects = 0;
  for (t_i=m_cache->list_begin(); t_i!=m_cache->list_end(); t_i++){
    CacheEntryImp *entry = const_cast<CacheEntryImp*>(
                           static_cast<const CacheEntryImp*>((*t_i).thePtr));      
    if(entry->isMarked()) numberOfMarkedObjects++;
  }  
  return numberOfMarkedObjects;
}

void
pool::CacheSvc::registerRef( IRegisteredRef* ref )
{
  m_refs.insert( ref );
}

void
pool::CacheSvc::unregisterRef( IRegisteredRef* ref )
{
  m_refs.erase( ref );
}

bool pool::CacheSvc::_retrieveObject(CacheEntry* anEntry){
  CacheEntryImp* aCacheEntry = dynamic_cast<CacheEntryImp*>(anEntry);
  if(!m_cache || !m_persSvc) {
    throw CacheSvcException("CacheSvc has not been fully constructed");
  }
  bool ret = false;
  if(aCacheEntry){
    const Token* theToken = aCacheEntry->token();
    if(theToken) {
      const std::pair< void*, const TypeH >& objStruct = 
                                                m_persSvc->readObject(*theToken, m_refHandler);
      void* theObjPtr = objStruct.first;
      // if everything ok, store the pointer in the cache
      if(theObjPtr){
        AnyPtr tmp(objStruct); 
        SharedAnyPtr thePtr(tmp);
        ret = m_cache->updateEntry(aCacheEntry,thePtr);
      }
    }    
  } 
  return ret;
}

bool pool::CacheSvc::_removeCacheEntry(CacheEntry* anEntry){  
  if(!m_cache) {
    throw CacheSvcException("CacheSvc has not been fully constructed");
  }
  bool ret = false;
  CacheEntryImp* aCacheEntry = dynamic_cast<CacheEntryImp*>(anEntry);
  if(aCacheEntry){
    ret = true;
    if(!aCacheEntry->isMarked()){
      m_cache->removeEntry(aCacheEntry);
    }
  }
  return ret;
}

void pool::CacheSvc::dumpCacheContent(){
  if(!m_cache){
    throw CacheSvcException("CacheSvc has not been fully constructed");
  }

  Cache::const_list_iterator t_i;
  std::cout << cacheSize() << " entries listed"<< std::endl;
  for (t_i=m_cache->list_begin(); t_i!=m_cache->list_end(); t_i++){
    CacheEntryImp *entry = const_cast<CacheEntryImp*>(
                           static_cast<const CacheEntryImp*>((*t_i).thePtr));      
    std::cout << "Entry " << entry << " - obj=" << entry->object().get().ptr()
		<< " - tok=" << entry->token() 
                << " - ref=" << entry->refCount() <<std::endl;
  }  
}

pool::ICache* pool::CacheSvc::cache() {
  return m_cache;
}
