#ifndef POOL_ICACHE_H
#define POOL_ICACHE_H

// Include files
#include "CacheTable.h"

namespace pool    {

  class SharedAnyPtr;
  class Token;
  class CacheEntryImp;

  /** @class ICache ICache.h
      Class for the management of the cache, deals with the registration/ 
      removal/ crosscheck of cache lines against object pointers and Tokens.
      @author Giacomo Govi
      @date 25/07/2002
  **/
  class ICache {

    public:

      // object pointer table iterator type
      typedef CacheTable::const_ptr_iterator const_ptr_iterator;

      // token table iterator type
      typedef CacheTable::const_token_iterator const_token_iterator;

      // location list iterator type
      typedef CacheTable::const_list_iterator const_list_iterator;

      // virtual dtor
      virtual ~ICache(){}

      // retrieve the cache line correpondent to a given object pointer.
      virtual CacheEntryImp* getEntry(const SharedAnyPtr& aPtr) = 0;

      // retrieve the cache line correpondent to a given token.
      virtual CacheEntryImp* getEntry(const Token& aToken) = 0;

      // verify if a specified line belongs to the cache
      virtual bool findEntry(const CacheEntryImp* aCacheEntryImp)=0;

      // add the sepcified line to the cache
      virtual bool addEntry(CacheEntryImp *line) = 0;

      // update the specified line with the specified object
      virtual bool updateEntry(CacheEntryImp* aCacheEntryImp, const SharedAnyPtr& aPtr)=0; 
		                         
      // update the specified line with the specified token
      virtual bool updateEntry(CacheEntryImp* aCacheEntryImp, const Token& aToken)=0; 

      // update the specified line resetting the object
      virtual bool removeObject(CacheEntryImp* aCacheEntryImp)=0;

      // update the specified line resetting and deleting the object on memory 
      virtual bool deleteObject(CacheEntryImp* aCacheEntryImp) = 0;

      // update the specified line resetting the token
      virtual bool removeToken(CacheEntryImp* aCacheEntryImp)=0;

      // remove the specified line from the cache 
      virtual bool removeEntry(CacheEntryImp* aCacheEntryImp)=0;

      // iterators on the cache location list
      virtual const_list_iterator list_begin()=0;
      virtual const_list_iterator list_end()=0;

      // iterators on the object pointer table
      virtual const_ptr_iterator ptr_iter_begin()=0;
      virtual const_ptr_iterator ptr_iter_end()=0;

      // iterators on the token table
      virtual const_token_iterator token_iter_begin()=0;
      virtual const_token_iterator token_iter_end()=0;

      // clear the cache tables, without to delete lines and contents
      virtual void clearTable()=0;

      // return the size of the cache lines list.
      virtual unsigned int cacheSize()=0;

  };   // class ICache
}      // namespace pool

#endif  // POOL_ICACHE_H



