#ifndef OBJECTRELATIONALACCESS_OBJECTRELATIONALMAPPINGDESCRIPTORELEMENT_H
#define OBJECTRELATIONALACCESS_OBJECTRELATIONALMAPPINGDESCRIPTORELEMENT_H

#include <vector>
#include <map>
#include <string>

namespace pool {

  class ObjectRelationalMappingElement;
  class ObjectRelationalTableRegister;

  // internal class for the handling of the user-defined mapping specification
  class ObjectRelationalMappingDescriptorElement {

  public:

    // the top level tag of the XML schema 
    static const std::string& mappingTag();

    // the class tag
    static const std::string& classTag();

    // the container tag
    static const std::string& containerTag();

    // the dependency tag
    static const std::string& dependencyTag();

    // the global mapping version attribute
    static const std::string& versionAttr();

    // the class mapping version attribute
    static const std::string& mappingVersionAttr();

    // the class existing mapping (to use) attribute
    static const std::string& useMappingAttr();

    // the class existing base mapping (to use) attribute
    static const std::string& baseMappingAttr();

    // the default version attribute
    static const std::string& defaultVersionAttrValue();

    // the name attribute
    static const std::string& nameAttr();

    // the table name attribute
    static const std::string& tableNameAttr();

    // the id column list attribute 
    static const std::string& idColumnsAttr();

    // the ref id column attribute 
    static const std::string& refIdColumnAttr();
    
    // the metadata column attribute 
    static const std::string& metadataColumnAttr();

    // the column name attribute
    static const std::string& columnAttr();

    // the position column attribute (for containers)
    static const std::string& positionColumnAttr();
    
    static const std::string& refTableColumnAttr();

    // the column name attribute
    static const std::string& poolContainerNameAttr();
    
  public:

    // constructor
    explicit ObjectRelationalMappingDescriptorElement(const std::string& type);

    // constructor
    explicit ObjectRelationalMappingDescriptorElement(const ObjectRelationalMappingElement& element);

    // destructor
    virtual ~ObjectRelationalMappingDescriptorElement();

    // copy constructor
    ObjectRelationalMappingDescriptorElement(const ObjectRelationalMappingDescriptorElement& aMappingDescriptorElement);

    // assignment operator
    ObjectRelationalMappingDescriptorElement& operator=(const ObjectRelationalMappingDescriptorElement& aMappingDescriptorElement);

    // add an item to the attribute list
    void setAttribute(const std::string& attributeName,const std::string& attributeValue);

    // map of attributes
    const std::map<std::string,std::string>& attributes() const;

    // attribute lookup
    std::string attribute(const std::string& attributeName) const;

    // name of the element
    std::string name() const; 

    // set the type of the element
    void setType(std::string type);

    // type of the element
    std::string type() const;

    // type as ObjectRelationalMapping type
    std::string objectRelationalMappingElementType() const;

    // add a child and retrieve its pointer
    ObjectRelationalMappingDescriptorElement* addChild(const std::string& type);

    // import an external child. Takes over the ownership of it
    void importChild(ObjectRelationalMappingDescriptorElement* element);

    // retrieve the list of childrens
    const std::vector<ObjectRelationalMappingDescriptorElement*>& children() const;

    // add a class dependency and retrieve its pointer
    ObjectRelationalMappingDescriptorElement* addDependency();
    
    // import an external dependency. Takes over the ownership of it
    void importDependency(ObjectRelationalMappingDescriptorElement* element);

    // retrieve the list of depenencies
    const std::vector<ObjectRelationalMappingDescriptorElement*>& dependencies() const;

    // alter the specified ObjectRelationalMapping element with the described directives 
    bool alter(ObjectRelationalMappingElement& element,
               ObjectRelationalTableRegister& tableRegister,
               bool ignoreMismatches=false) const;

    // retrieve the parent element
    ObjectRelationalMappingDescriptorElement* parent() const;

  private:

    // alter the column names of the specified ObjectRelationalMapping element, according to the described directives
    void alterColumnNames(ObjectRelationalMappingElement& toElement,
                          ObjectRelationalTableRegister& tableRegister) const;

    // set the described directives according to the content of the specified ObjectRelationalMapping element
    void setColumnNames(const ObjectRelationalMappingElement& fromElement);

    // type of the element
    std::string m_type; 

    // attribute map
    std::map<std::string,std::string> m_attributes;

    // children list
    std::vector<ObjectRelationalMappingDescriptorElement*> m_children;

    // dependencies list
    std::vector<ObjectRelationalMappingDescriptorElement*> m_dependencies;    

    // parent element
    ObjectRelationalMappingDescriptorElement* m_parent;

    // default value for the table, to be retrieved during the alter method execution
    std::string m_defaultTableName;

  };


}

#endif // OBJECTRELATIONALACCESS_OBJECTRELATIONALMAPPINGDESCRIPTORELEMENT_H
