#include "ObjectRelationalAccess/ObjectRelationalMappingPersistency.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingSchema.h"
#include "ObjectRelationalAccess/ObjectRelationalMapping.h"
#include "ObjectRelationalAccess/ObjectRelationalException.h"
#include "ObjectRelationalAccess/ObjectRelationalNames.h"
#include "MappingMaterialization.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/TableDescription.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/IColumn.h"
#include "RelationalAccess/ITablePrivilegeManager.h"
#include "RelationalAccess/ITableDataEditor.h"
#include "RelationalAccess/ITableDescription.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ICursor.h"
#include "RelationalAccess/SchemaException.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/AttributeSpecification.h"
#include "CoralBase/MessageStream.h"

#include <sstream>
#include <memory>

std::string
pool::ObjectRelationalMappingPersistency::mappingVersionForClass( const std::string& className, int iteration )
{
  std::ostringstream os;
  os << className;
  os << "_automatic_default";
  if ( iteration > 0 ) {
    os << "_";
    if ( iteration < 10 ) os << "0";
    if ( iteration < 100 ) os << "0";
    os << iteration;
  }
  return os.str();
}

void
pool::ObjectRelationalMappingPersistency::storeMappingElement( const std::string& topElementId,
                                                               const pool::ObjectRelationalMappingElement& element,
                                                               coral::ITableDataEditor& dataEditor,
                                                               coral::AttributeList& dataBuffer )
{
  int iParamIndex = 0;
  size_t columns = element.columnNames().size();
  for ( std::vector< std::string >::const_iterator iColumn = element.columnNames().begin();
        iColumn != element.columnNames().end(); ++iColumn, ++iParamIndex ) {
    std::stringstream rowId;
    rowId << topElementId;
    if( columns > 1 ) rowId << "[" << iParamIndex << "]";
    dataBuffer[ pool::ObjectRelationalMappingSchema::elementIdFieldInMappingElementTable() ].data<std::string>()= rowId.str();
    dataBuffer[ pool::ObjectRelationalMappingSchema::elementTypeFieldInMappingElementTable() ].data<std::string>()= pool::ObjectRelationalMappingElement::elementTypeAsString( element.elementType() );
    if ( element.scopeName().empty() ) {
      dataBuffer[ pool::ObjectRelationalMappingSchema::scopeNameFieldInMappingElementTable() ].data<std::string>()= " " ;
    } else {
      dataBuffer[ pool::ObjectRelationalMappingSchema::scopeNameFieldInMappingElementTable() ].data<std::string>()= element.scopeName();
    }
    dataBuffer[ pool::ObjectRelationalMappingSchema::variableNameFieldInMappingElementTable() ].data<std::string>()= element.variableName();
    dataBuffer[ pool::ObjectRelationalMappingSchema::variableParIndexFieldInMappingElementTable() ].data<int>() = iParamIndex;
    dataBuffer[ pool::ObjectRelationalMappingSchema::variableTypeFieldInMappingElementTable() ].data<std::string>()= element.variableType();
    dataBuffer[ pool::ObjectRelationalMappingSchema::tableNameFieldInMappingElementTable() ].data<std::string>() = element.tableName();
    dataBuffer[ pool::ObjectRelationalMappingSchema::columnNameFieldInMappingElementTable() ].data<std::string>() = *iColumn;
    try 
    {
      dataEditor.insertRow( dataBuffer );
    }
    catch (const coral::SchemaException& e)
    {
      throw pool::ObjectRelationalException( e.what(),
                                             "ObjectRelationalMappingPersistency::storeMappingElement" );
    }
  }

  unsigned int elementIndex = 0;
  for ( pool::ObjectRelationalMappingElement::const_iterator iElement = element.begin();
        iElement != element.end(); ++iElement ) {
    std::stringstream elementId;
    elementId << topElementId << "_" << elementIndex;
    pool::ObjectRelationalMappingPersistency::storeMappingElement( elementId.str(),iElement->second, dataEditor, dataBuffer );
    elementIndex++;
  }
}

pool::ObjectRelationalMappingPersistency::ObjectRelationalMappingPersistency( coral::ISchema& schema ):
  m_schema( schema ),
  m_storedMappings(),
  m_versionsLoaded( false ),
  m_tempMappingTable()
{}


pool::ObjectRelationalMappingPersistency::~ObjectRelationalMappingPersistency()
{
}

void
pool::ObjectRelationalMappingPersistency::initializeForWriting(){
  createIfNecessaryTheMappingTables();
  loadMappingVersions();  
}

//* private methods

void
pool::ObjectRelationalMappingPersistency::createIfNecessaryTheMappingTables(){
  ObjectRelationalMappingSchema mappingSchema(m_schema);
  if(!mappingSchema.existTables()){
    mappingSchema.createTables();
  }
}

void
pool::ObjectRelationalMappingPersistency::loadMappingVersions()
{
  if ( m_versionsLoaded ) return;
  std::auto_ptr<coral::IQuery> query( m_schema.tableHandle( pool::ObjectRelationalMappingSchema::mappingVersionTable() ).newQuery() );
  query->addToOutputList( pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable() );
  coral::ICursor& cursor = query->execute();
  while ( cursor.next() ) {
    const coral::AttributeList& currentRow = cursor.currentRow();
    m_storedMappings.insert(currentRow[pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable()].data<std::string>());
  }
  m_versionsLoaded = true;
}


void pool::ObjectRelationalMappingPersistency::makeTableList(const ObjectRelationalMappingElement& element,std::set<std::string>& tableList)
{
  tableList.insert(element.tableName());
  for(ObjectRelationalMappingElement::const_iterator iChild=element.begin();iChild!=element.end();iChild++)
  {
    makeTableList(iChild->second,tableList);
  }
}

void
pool::ObjectRelationalMappingPersistency::buildMappingElement( ObjectRelationalMappingElement& parentElement, const std::string& scopeName ) const{
  std::map<std::string,std::map<std::string,MappingDescription> >::iterator iMap = m_tempMappingTable.find(scopeName);
  if(iMap!=m_tempMappingTable.end()){
    for(std::map<std::string,MappingDescription>::const_iterator iMapEntry=iMap->second.begin();iMapEntry!=iMap->second.end();iMapEntry++){
      std::string variableName = iMapEntry->first;
      ObjectRelationalMappingElement& element = parentElement.appendSubElement( iMapEntry->second.elementType,
                                                                                variableName,
                                                                                iMapEntry->second.variableType,
                                                                                iMapEntry->second.tableName );
      element.setColumnNames(iMapEntry->second.columns);
      std::string nextScope(scopeName);
      nextScope.append("::").append(variableName);
      buildMappingElement( element, nextScope);
    }
    m_tempMappingTable.erase(scopeName);
  }
}

//* public methods

std::map<std::string,std::string>
pool::ObjectRelationalMappingPersistency::retrieveMappingVersions(bool tmpTable) const
{
  std::string tableName = pool::ObjectRelationalMappingSchema::mappingVersionTable();
  if( tmpTable ) tableName = pool::ObjectRelationalMappingSchema::mappingVersionTmpTable();
  
  std::map<std::string,std::string> mappingList;
  try {
    if( !m_schema.existsTable( tableName ) ){
      throw pool::ObjectRelationalException( "Expected  table \""+tableName+"\" does not exist in the schema." ,
                                             "ObjectRelationalMappingPersistency::retrieveMappingVersions" );
      
    }
    std::auto_ptr<coral::IQuery> query( m_schema.tableHandle( tableName ).newQuery() );
    query->setDistinct();
    query->addToOutputList( pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable() );
    query->addToOutputList( pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable() );
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      const coral::AttributeList& currentRow = cursor.currentRow();
      std::string containerName = currentRow[pool::ObjectRelationalMappingSchema::containerNameFieldInMappingVersionTable()].data<std::string>();
      std::string mappingVersion = currentRow[pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable()].data<std::string>();
      mappingList.insert(std::make_pair(mappingVersion,containerName));
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::retrieveMappingVersions" );
  }
  return mappingList;
}

bool
pool::ObjectRelationalMappingPersistency::storeMappingVersions( const std::map<std::string,std::string>& data, bool tmpTable) const {
  std::string tableName = pool::ObjectRelationalMappingSchema::mappingVersionTable();
  if( tmpTable ) tableName = pool::ObjectRelationalMappingSchema::mappingVersionTmpTable();
  try {

    if( !m_schema.existsTable( tableName ) ){
      throw pool::ObjectRelationalException( "Expected table \""+tableName+"\" does not exist in the schema." ,
                                             "ObjectRelationalMappingPersistency::storeMappingVersions" );
      
    }
    
    // Store an entry into the version table
    coral::ITable& versionTable = m_schema.tableHandle( tableName );
    coral::AttributeList dataForVersionTable;
    dataForVersionTable.extend<std::string>(pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable());
    dataForVersionTable.extend<std::string>(pool::ObjectRelationalMappingSchema::containerNameFieldInMappingVersionTable());
    for(std::map<std::string,std::string>::const_iterator iV = data.begin();
        iV!=data.end(); ++iV){
      coral::AttributeList::iterator iAttribute = dataForVersionTable.begin();
      iAttribute->data< std::string >() = iV->first;
      ++iAttribute;
      iAttribute->data< std::string >() = iV->second;
      versionTable.dataEditor().insertRow( dataForVersionTable );
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::storeMappingVersions" );
  }
  return true;
}

std::string
pool::ObjectRelationalMappingPersistency::newMappingVersionForClass( const std::string& className ){
  std::map<std::string,std::string> existingMappingVersions = retrieveMappingVersions();
  std::string newMappingVersion = "";
  for ( int iteration = 0;; ++iteration ) {
    newMappingVersion = mappingVersionForClass( className, iteration );
    bool found = false;
    for ( std::map<std::string,std::string>::reverse_iterator iVersion = existingMappingVersions.rbegin();
          iVersion != existingMappingVersions.rend(); ++iVersion ) {
      if ( iVersion->first == newMappingVersion ) {
        found = true;
        break;
      }
    }
    if ( ! found ) break;
  }
  return newMappingVersion;
}

std::map<std::pair<std::string,std::string>,std::string>
pool::ObjectRelationalMappingPersistency::retrieveMappingForClassVersions(bool tmpTable) const{
  std::string tableName = pool::ObjectRelationalMappingSchema::classVersionTable();
  if( tmpTable ) tableName = pool::ObjectRelationalMappingSchema::classVersionTmpTable();
  std::map<std::pair<std::string,std::string>,std::string> classVersionMap;
  try {

    if( !m_schema.existsTable( tableName ) ){
      throw pool::ObjectRelationalException( "Expected  table \""+tableName+"\" does not exist in the schema." ,
                                             "ObjectRelationalMappingPersistency::retrieveMappingForClassVersions" );
      
    }
    std::auto_ptr<coral::IQuery> query( m_schema.tableHandle( tableName ).newQuery() );
    query->addToOutputList( pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable() );
    query->addToOutputList( pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable() );
    query->addToOutputList( pool::ObjectRelationalMappingSchema::mappingVersionFieldInClassVersionTable() );
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      const coral::AttributeList& currentRow = cursor.currentRow();
      std::string classVersion = currentRow[pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable()].data<std::string>();
      std::string containerName = currentRow[pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable()].data<std::string>();
      std::string mappingVersion = currentRow[pool::ObjectRelationalMappingSchema::mappingVersionFieldInClassVersionTable()].data<std::string>();
      classVersionMap.insert(std::make_pair(std::make_pair(classVersion,containerName),mappingVersion));
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::retrieveMappingForClassVersions" );
  }
  return classVersionMap;
}

std::string
pool::ObjectRelationalMappingPersistency::selectMappingVersion( const std::string& classVersion, const std::string& containerName ) const
{
  std::string result("");
  try {
    std::auto_ptr<coral::IQuery> query( m_schema.tableHandle( pool::ObjectRelationalMappingSchema::classVersionTable() ).newQuery() );
    query->addToOutputList( pool::ObjectRelationalMappingSchema::mappingVersionFieldInClassVersionTable() );
    std::ostringstream condition;
    condition << pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable() << " =:" <<
      pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable() << " AND " <<
      pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable();
    condition << " =:" << pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable();

    coral::AttributeList condData;
    condData.extend<std::string>(pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable());
    condData.extend<std::string>(pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable());
    coral::AttributeList::iterator iAttribute = condData.begin();
    iAttribute->data< std::string >() = classVersion;
    ++iAttribute;
    iAttribute->data< std::string >() = containerName;
    query->setCondition( condition.str(), condData );
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      const coral::AttributeList& currentRow = cursor.currentRow();
      result = currentRow[pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable()].data<std::string>();
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::selectMappingVersion" );
  }
  return result;
}

std::string
pool::ObjectRelationalMappingPersistency::containerForMappingVersion( const std::string& mappingVersion) const {
  std::string currentContainerName("");
  try {
    std::auto_ptr<coral::IQuery> query( m_schema.tableHandle( pool::ObjectRelationalMappingSchema::mappingVersionTable() ).newQuery() );
    query->setDistinct();
    query->addToOutputList( pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable() );
    std::ostringstream condition;
    condition << pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable() << " =:"<<
      pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable();
    coral::AttributeList condData;
    condData.extend<std::string>(pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable());
    coral::AttributeList::iterator iAttribute = condData.begin();
    iAttribute->data< std::string >() = mappingVersion;
    query->setCondition( condition.str(), condData );
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      const coral::AttributeList& currentRow = cursor.currentRow();
      currentContainerName = currentRow[pool::ObjectRelationalMappingSchema::containerNameFieldInMappingVersionTable()].data<std::string>();
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::containerForMappingVersion" );
  }
  return currentContainerName;
}

std::vector<std::string>
pool::ObjectRelationalMappingPersistency::mappingsForContainer( const std::string& containerName ) const {
  std::vector<std::string> mappings;
  try {
    std::auto_ptr<coral::IQuery> query( m_schema.tableHandle( pool::ObjectRelationalMappingSchema::mappingVersionTable() ).newQuery() );
    query->setDistinct();
    query->addToOutputList( pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable());
    std::ostringstream condition;
    condition << pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable() << " =:"<<
      pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable();
    coral::AttributeList condData;
    condData.extend<std::string>(pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable());
    coral::AttributeList::iterator iAttribute = condData.begin();
    iAttribute->data< std::string >() = containerName;
    query->setCondition( condition.str(), condData );
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      const coral::AttributeList& currentRow = cursor.currentRow();
      mappings.push_back(currentRow[pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable()].data<std::string>());
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::mappingsForContainer" );
  }
  return mappings;
}

bool
pool::ObjectRelationalMappingPersistency::storeMappingForClassVersions( std::map<std::pair<std::string,std::string>,std::string> data, bool tmpTable) const {
  std::string tableName = pool::ObjectRelationalMappingSchema::classVersionTable();
  if( tmpTable ) tableName = pool::ObjectRelationalMappingSchema::classVersionTmpTable();
  try {

    if( !m_schema.existsTable( tableName ) ){
      throw pool::ObjectRelationalException( "Expected  table \""+tableName+"\" does not exist in the schema." ,
                                             "ObjectRelationalMappingPersistency::storeMappingForClassVersions" );
      
    }
    
    // Store an entry into the class version table
    coral::ITable& classVersionTable = m_schema.tableHandle( tableName );
    coral::AttributeList dataForClassVersionTable;
    dataForClassVersionTable.extend<std::string>(pool::ObjectRelationalMappingSchema::mappingVersionFieldInClassVersionTable());
    dataForClassVersionTable.extend<std::string>(pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable());
    dataForClassVersionTable.extend<std::string>(pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable());
    for(std::map<std::pair<std::string,std::string>,std::string>::const_iterator iCv = data.begin();
        iCv != data.end(); iCv++){
      coral::AttributeList::iterator iAttr = dataForClassVersionTable.begin();
      iAttr->data< std::string >() = iCv->second;
      ++iAttr;
      iAttr->data< std::string >() = iCv->first.first;
      ++iAttr;
      iAttr->data< std::string >() = iCv->first.second;
      classVersionTable.dataEditor().insertRow( dataForClassVersionTable );      
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::storeMappingForClassVersions" );
  }
  return true;
}

bool
pool::ObjectRelationalMappingPersistency::assignMapping( const std::string& classVersion, const std::string& mappingVersion, const std::string& toContainerName)
{
  coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
  // first check if the specified mapping is used by an other container
  std::string currentCont = containerForMappingVersion(mappingVersion);
  if(currentCont.empty()){
    log << coral::Error << "Mapping \""<<mappingVersion<<"\" for container \""<<toContainerName<<"\" does not exist in the database."<<coral::MessageStream::endmsg;
    return false;
  } else {
    if(currentCont != toContainerName){
      log << coral::Error << "Mapping \""<<mappingVersion<<"\" is already assigned to container \""<<currentCont<<"\"."<<coral::MessageStream::endmsg;
      return false;
    }
  }
  try {
    std::auto_ptr<coral::IQuery> query0( m_schema.tableHandle( pool::ObjectRelationalMappingSchema::mappingVersionTable() ).newQuery() );
    query0->addToOutputList( pool::ObjectRelationalMappingSchema::containerNameFieldInMappingVersionTable() );
    std::ostringstream cond;
    cond << pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable()<< " =:"<<
      pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable();
    coral::AttributeList condData;
    condData.extend<std::string>(pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable());
    coral::AttributeList::iterator iAttr = condData.begin();
    iAttr->data< std::string >() = mappingVersion;
    query0->setCondition( cond.str(), condData);
    std::string currentContainerName("");
    coral::ICursor& cursor0 = query0->execute();
    while ( cursor0.next() ) {
      const coral::AttributeList& currentRow = cursor0.currentRow();      
      currentContainerName = currentRow[pool::ObjectRelationalMappingSchema::containerNameFieldInMappingVersionTable()].data<std::string>();
    }
    if(currentContainerName.empty()){
      log << coral::Error << "Mapping version \""+mappingVersion+"\" has not been found in the database." << coral::MessageStream::endmsg;
      return false;
    } else {
      if(toContainerName != currentContainerName){
        log << coral::Error << "Mapping version=\""<<mappingVersion<<"\" is already assigned to container \""<<currentContainerName<<"\"." << coral::MessageStream::endmsg;
        return false;
      }
    }
    
    // first check if the classVersion and the container are already represented
    std::auto_ptr<coral::IQuery> query1( m_schema.tableHandle( pool::ObjectRelationalMappingSchema::classVersionTable() ).newQuery() );
    query1->addToOutputList( pool::ObjectRelationalMappingSchema::mappingVersionFieldInClassVersionTable() );
    std::ostringstream condition;
    condition << pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable() << " =:"<<
      pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable();
    condition << " AND ";
    condition << pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable() << " =:"<<
      pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable();
    coral::AttributeList conditionData;
    conditionData.extend<std::string>(pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable());
    conditionData.extend<std::string>(pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable());
    coral::AttributeList::iterator iAttribute = conditionData.begin();
    iAttribute->data< std::string >() = classVersion;
    ++iAttribute;
    iAttribute->data< std::string >() = toContainerName;    
    query1->setCondition( condition.str(), conditionData );
    std::set<std::string> classVersions;
    coral::ICursor& cursor1 = query1->execute();
    std::string currentVersion("");
    while ( cursor1.next() ) {
      const coral::AttributeList& currentRow = cursor1.currentRow();
      currentVersion = currentRow[pool::ObjectRelationalMappingSchema::mappingVersionFieldInClassVersionTable()].data<std::string>();
    }
    bool update = false;
    if(!currentVersion.empty()){
      if(currentVersion!=mappingVersion){
        update = true;
      } else {
        log << coral::Debug << "Mapping version=\""<<mappingVersion<<"\" is already assigned to container \""<<toContainerName<<"\" for class version \""<<
          classVersion<<"\"." << coral::MessageStream::endmsg;
        return true;
      }
    }
    
    coral::ITable& classVersionTable = m_schema.tableHandle( pool::ObjectRelationalMappingSchema::classVersionTable() );
    coral::AttributeList inputData;
    inputData.extend<std::string>(pool::ObjectRelationalMappingSchema::mappingVersionFieldInClassVersionTable());
    inputData.extend<std::string>(pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable());
    inputData.extend<std::string>(pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable());
    coral::AttributeList::iterator iInAttr = inputData.begin();
    iInAttr->data< std::string >() = mappingVersion;
    ++iInAttr;
    iInAttr->data< std::string >() = classVersion;
    ++iInAttr;
    iInAttr->data< std::string >() = toContainerName;
    if(update){
      std::string setClause = pool::ObjectRelationalMappingSchema::mappingVersionFieldInClassVersionTable()+" =:"+
        pool::ObjectRelationalMappingSchema::mappingVersionFieldInClassVersionTable();
      std::string whereClause = pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable()+" =:"+
        pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable()+" AND "+
        pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable()+" =:"+
        pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable();
      classVersionTable.dataEditor().updateRows( setClause,whereClause, inputData  );
    } else {
      classVersionTable.dataEditor().insertRow( inputData );
    }
    log << coral::Debug << "Mapping version \""<< mappingVersion << "\" for class version \""<<classVersion<<
      "\" has been assigned to pool container \"" << toContainerName << "\""<<coral::MessageStream::endmsg;
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::assignMapping" );
  }
  return true;
}

std::map<std::string, std::string>
pool::ObjectRelationalMappingPersistency::containerTableMap()
{
  std::map<std::string,std::string> containerTableMap;
  try {
    std::auto_ptr<coral::IQuery> query(m_schema.newQuery());
    query->addToTableList( pool::ObjectRelationalMappingSchema::mappingElementTable(),"T0");
    query->addToTableList( pool::ObjectRelationalMappingSchema::mappingVersionTable(), "T1");
    query->setDistinct();
    query->addToOutputList( "T0."+pool::ObjectRelationalMappingSchema::tableNameFieldInMappingElementTable() );
    query->addToOutputList( "T1."+pool::ObjectRelationalMappingSchema::containerNameFieldInMappingVersionTable());
    std::ostringstream condition;
    condition << "T0."<<pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable()<<"="<<
      "T1."<<pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable();
    coral::AttributeList condData;
    query->setCondition(condition.str(),condData);
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      const coral::AttributeList& currentRow = cursor.currentRow();
      std::string tableName = currentRow[ "T0."+pool::ObjectRelationalMappingSchema::tableNameFieldInMappingElementTable()].data<std::string>();
      std::string containerName = currentRow[ "T1."+pool::ObjectRelationalMappingSchema::containerNameFieldInMappingVersionTable()].data<std::string>();
      containerTableMap.insert(std::make_pair(tableName,containerName));
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::containerTableMap" );
  }
  return containerTableMap;
}

void
pool::ObjectRelationalMappingPersistency::retrieveMapping(const std::string& version,
                                                          std::map<std::pair<std::string,std::string>,std::pair<std::string,std::string> >& mappingStruct) const
{
  coral::ITable& mappingTable = m_schema.tableHandle( pool::ObjectRelationalMappingSchema::mappingElementTable() );
  try {
    std::auto_ptr<coral::IQuery> query(mappingTable.newQuery());
    query->addToOutputList( pool::ObjectRelationalMappingSchema::scopeNameFieldInMappingElementTable());
    query->addToOutputList( pool::ObjectRelationalMappingSchema::variableNameFieldInMappingElementTable());
    query->addToOutputList( pool::ObjectRelationalMappingSchema::tableNameFieldInMappingElementTable() );
    query->addToOutputList( pool::ObjectRelationalMappingSchema::columnNameFieldInMappingElementTable());
    std::ostringstream condition;
    condition << pool::ObjectRelationalMappingSchema::elementTypeFieldInMappingElementTable()<<"= '"<<
      pool::ObjectRelationalMappingElement::primitiveMappingElementType()<<"' AND "<<
      pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable()<<"= :"<<
      pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable();
    coral::AttributeList condData;
    condData.extend<std::string>(pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable());
    coral::AttributeList::iterator iAttribute = condData.begin();
    iAttribute->data< std::string >() = version;
    query->setCondition( condition.str(), condData );
    query->addToOrderList("T0."+pool::ObjectRelationalMappingSchema::scopeNameFieldInMappingElementTable());
    query->addToOrderList("T0."+pool::ObjectRelationalMappingSchema::variableNameFieldInMappingElementTable());
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      const coral::AttributeList& currentRow = cursor.currentRow();
      std::string scopeName = currentRow[ pool::ObjectRelationalMappingSchema::scopeNameFieldInMappingElementTable()].data<std::string>();
      std::string variableName = currentRow[ pool::ObjectRelationalMappingSchema::variableNameFieldInMappingElementTable() ].data<std::string>();
      std::string tableName = currentRow[ pool::ObjectRelationalMappingSchema::tableNameFieldInMappingElementTable() ].data<std::string>();
      std::string columnName = currentRow[ pool::ObjectRelationalMappingSchema::columnNameFieldInMappingElementTable()].data<std::string>();
      mappingStruct.insert(std::make_pair(std::make_pair(scopeName,variableName),std::make_pair(tableName,columnName)));
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::retrieveMapping" );
  }
}

pool::ObjectRelationalMapping*
pool::ObjectRelationalMappingPersistency::retrieveMapping( const std::string& version, bool tmpTable ) const
{
  bool empty = true;
  std::string tableName = pool::ObjectRelationalMappingSchema::mappingElementTable();
  if( tmpTable ) tableName = pool::ObjectRelationalMappingSchema::mappingElementTmpTable();
  
  std::map< std::string, MappingDescription> topLevelObjects;
  try {
    if( !m_schema.existsTable( tableName ) ){
      throw pool::ObjectRelationalException( "Expected table \""+tableName+"\" does not exist in the schema." ,
                                             "ObjectRelationalMappingPersistency::readMapping" );
    }
    coral::ITable& mappingTable = m_schema.tableHandle( tableName );
    std::auto_ptr<coral::IQuery> query(mappingTable.newQuery());
    
    query->addToOutputList( pool::ObjectRelationalMappingSchema::elementTypeFieldInMappingElementTable());
    query->addToOutputList( pool::ObjectRelationalMappingSchema::scopeNameFieldInMappingElementTable());
    query->addToOutputList( pool::ObjectRelationalMappingSchema::variableNameFieldInMappingElementTable());
    query->addToOutputList( pool::ObjectRelationalMappingSchema::variableTypeFieldInMappingElementTable());
    query->addToOutputList( pool::ObjectRelationalMappingSchema::tableNameFieldInMappingElementTable() );
    query->addToOutputList( pool::ObjectRelationalMappingSchema::columnNameFieldInMappingElementTable());
    std::ostringstream condition;
    condition << pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable()<<"= :"<<
      pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable();
    coral::AttributeList condData;
    condData.extend<std::string>(pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable());
    coral::AttributeList::iterator iAttribute = condData.begin();
    iAttribute->data< std::string >() = version;
    query->setCondition( condition.str(), condData );
    query->addToOrderList( pool::ObjectRelationalMappingSchema::scopeNameFieldInMappingElementTable());
    query->addToOrderList( pool::ObjectRelationalMappingSchema::variableNameFieldInMappingElementTable());
    query->addToOrderList( pool::ObjectRelationalMappingSchema::variableParIndexFieldInMappingElementTable());
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      empty = false;
      const coral::AttributeList& currentRow = cursor.currentRow();
      std::string elementType = currentRow[ pool::ObjectRelationalMappingSchema::elementTypeFieldInMappingElementTable() ].data<std::string>();
      std::string scopeName = currentRow[ pool::ObjectRelationalMappingSchema::scopeNameFieldInMappingElementTable()].data<std::string>();
      std::string variableName = currentRow[ pool::ObjectRelationalMappingSchema::variableNameFieldInMappingElementTable() ].data<std::string>();
      std::string variableType = currentRow[ pool::ObjectRelationalMappingSchema::variableTypeFieldInMappingElementTable() ].data<std::string>();
      std::string tableName = currentRow[ pool::ObjectRelationalMappingSchema::tableNameFieldInMappingElementTable() ].data<std::string>();
      std::string columnName = currentRow[ pool::ObjectRelationalMappingSchema::columnNameFieldInMappingElementTable()].data<std::string>();
      MappingDescription mapDescr;
      mapDescr.variableType = variableType;
      mapDescr.elementType = elementType;
      mapDescr.tableName = tableName;
      mapDescr.columns = std::vector<std::string>(1,columnName);
      if(scopeName == " "){  
        std::map<std::string,MappingDescription>::iterator iM=topLevelObjects.find(variableName);
        if(iM==topLevelObjects.end()){
          topLevelObjects.insert(std::make_pair(variableName, mapDescr) );
        } else {
          iM->second.columns.push_back(columnName);
        }
      } else {
        std::map<std::string,std::map<std::string,MappingDescription> >::iterator iMap = m_tempMappingTable.find(scopeName);
        if(iMap==m_tempMappingTable.end()) {
          std::map<std::string,MappingDescription> entry;
          entry.insert(std::make_pair(variableName,mapDescr));
          m_tempMappingTable.insert(std::make_pair(scopeName,entry));
        } else {                    
          std::map<std::string,MappingDescription>::iterator iInnerMap = iMap->second.find(variableName);
          if(iInnerMap==iMap->second.end()){
            iMap->second.insert(std::make_pair(variableName,mapDescr));
          } else {
            iInnerMap->second.columns.push_back(columnName);
          }
        }
      
      }
    
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::readMapping" );
  }

  
  pool::ObjectRelationalMapping* mapping = 0;
  if(!empty){
    mapping = new pool::ObjectRelationalMapping( version );
    for(std::map<std::string, MappingDescription>::const_iterator iObj=topLevelObjects.begin();iObj!=topLevelObjects.end();iObj++){
      ObjectRelationalMappingElement* element = 0;
      if(iObj->second.elementType == ObjectRelationalMappingElement::objectMappingElementType()){
        element = &mapping->setElement(iObj->first,iObj->second.tableName);
      } else {
        element = &mapping->appendDependentElement(iObj->first,iObj->second.tableName);
      }
      element->setColumnNames(iObj->second.columns);
      buildMappingElement(*element,iObj->first);
    }
  }
  m_tempMappingTable.clear();
  return mapping;
}

pool::ObjectRelationalMapping*
pool::ObjectRelationalMappingPersistency::retrieveMappingFromOldSchema( const std::string& version ) const {
  coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );

  bool empty = true;
  std::map<std::string, MappingDescription> topLevelObjects;
  std::auto_ptr<coral::IQuery> query(m_schema.newQuery());
  try {
    query->addToTableList( pool::ObjectRelationalMappingOldSchema::mappingElementTable(),"T0");
    query->addToTableList( pool::ObjectRelationalMappingOldSchema::mappingElementColumnsTable(), "T1");
    query->addToOutputList( "T1."+pool::ObjectRelationalMappingOldSchema::columnFieldInMappingElementColumnsTable());
    query->addToOutputList( "T0."+pool::ObjectRelationalMappingOldSchema::elementTypeFieldInMappingElementTable());
    query->addToOutputList( "T0."+pool::ObjectRelationalMappingOldSchema::variableTypeFieldInMappingElementTable());
    query->addToOutputList( "T0."+pool::ObjectRelationalMappingOldSchema::variableNameFieldInMappingElementTable());
    query->addToOutputList( "T0."+pool::ObjectRelationalMappingOldSchema::scopeNameFieldInMappingElementTable());
    query->addToOutputList( "T0."+pool::ObjectRelationalMappingOldSchema::tableNameFieldInMappingElementTable() );
    std::ostringstream condition;
    condition << "T1."<<pool::ObjectRelationalMappingOldSchema::versionFieldInMappingElementTable()<<"="<<
      "T0."<<pool::ObjectRelationalMappingOldSchema::versionFieldInMappingElementTable()<<" AND "<<
      "T1."<<pool::ObjectRelationalMappingOldSchema::variableNameFieldInMappingElementTable()<<"="
      "T0."<<pool::ObjectRelationalMappingOldSchema::variableNameFieldInMappingElementTable()<<" AND "<<
      "T1."<<pool::ObjectRelationalMappingOldSchema::scopeNameFieldInMappingElementTable()<<"="<<
      "T0."<<pool::ObjectRelationalMappingOldSchema::scopeNameFieldInMappingElementTable()<<" AND "<<
      "T0."<<pool::ObjectRelationalMappingOldSchema::versionFieldInMappingElementTable()<<"= :"<<
      pool::ObjectRelationalMappingOldSchema::versionFieldInMappingElementTable();
    coral::AttributeList condData;
    condData.extend<std::string>(pool::ObjectRelationalMappingOldSchema::versionFieldInMappingElementTable());
    coral::AttributeList::iterator iAttribute = condData.begin();
    iAttribute->data< std::string >() = version;
    query->setCondition( condition.str(), condData );
    query->addToOrderList("T0."+pool::ObjectRelationalMappingOldSchema::scopeNameFieldInMappingElementTable());
    query->addToOrderList("T0."+pool::ObjectRelationalMappingOldSchema::variableNameFieldInMappingElementTable());
    query->addToOrderList("T1."+pool::ObjectRelationalMappingOldSchema::positionFieldInMappingElementColumnsTable());
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      empty = false;
      const coral::AttributeList& currentRow = cursor.currentRow();
      std::string columnName = currentRow[ "T1."+pool::ObjectRelationalMappingOldSchema::columnFieldInMappingElementColumnsTable()].data<std::string>();
      std::string elementType = currentRow[ "T0."+pool::ObjectRelationalMappingOldSchema::elementTypeFieldInMappingElementTable() ].data<std::string>();
      std::string variableType = currentRow[ "T0."+pool::ObjectRelationalMappingOldSchema::variableTypeFieldInMappingElementTable() ].data<std::string>();
      std::string variableName = currentRow[ "T0."+pool::ObjectRelationalMappingOldSchema::variableNameFieldInMappingElementTable() ].data<std::string>();
      std::string scopeName = currentRow[ "T0."+pool::ObjectRelationalMappingOldSchema::scopeNameFieldInMappingElementTable()].data<std::string>();
      std::string tableName = currentRow[ "T0."+pool::ObjectRelationalMappingOldSchema::tableNameFieldInMappingElementTable() ].data<std::string>();
      MappingDescription mapDescr;
      mapDescr.variableType = variableType;
      mapDescr.elementType = elementType;
      mapDescr.tableName = tableName;
      mapDescr.columns = std::vector<std::string>(1,columnName);
      if(scopeName == " "){
        std::map<std::string,MappingDescription>::iterator iM=topLevelObjects.find(variableName);
        if(iM==topLevelObjects.end()){
          topLevelObjects.insert(std::make_pair(variableName, mapDescr) );
        } else {
          iM->second.columns.push_back(columnName);
        }
      } else {
        std::map<std::string,std::map<std::string,MappingDescription> >::iterator iMap = m_tempMappingTable.find(scopeName);
        if(iMap==m_tempMappingTable.end()) {
          std::map<std::string,MappingDescription> entry;
          entry.insert(std::make_pair(variableName,mapDescr));
          m_tempMappingTable.insert(std::make_pair(scopeName,entry));
        } else {                    
          std::map<std::string,MappingDescription>::iterator iInnerMap = iMap->second.find(variableName);
          if(iInnerMap==iMap->second.end()){
            iMap->second.insert(std::make_pair(variableName,mapDescr));
          } else {
            iInnerMap->second.columns.push_back(columnName);
          }
        }
      
      }
    
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::loadMappingInOneGo" );
  }

  pool::ObjectRelationalMapping* mapping = 0;
  if(!empty){
    mapping = new pool::ObjectRelationalMapping( version );
    for(std::map<std::string, MappingDescription>::const_iterator iObj=topLevelObjects.begin();iObj!=topLevelObjects.end();iObj++){
      ObjectRelationalMappingElement* element = 0;
      if(iObj->second.elementType == ObjectRelationalMappingElement::objectMappingElementType()){
        element = &mapping->setElement(iObj->first,iObj->second.tableName);
      } else {
        element = &mapping->appendDependentElement(iObj->first,iObj->second.tableName);
      }
      element->setColumnNames(iObj->second.columns);
      const_cast<pool::ObjectRelationalMappingPersistency*>(this)->buildMappingElement(*element,iObj->first);
    }
  }
  m_tempMappingTable.clear();
  return mapping;  
}

bool
pool::ObjectRelationalMappingPersistency::storeMapping( const ObjectRelationalMapping& mapping, bool tmpTable) {
  std::string tableName = pool::ObjectRelationalMappingSchema::mappingElementTable();
  if( tmpTable ) tableName = pool::ObjectRelationalMappingSchema::mappingElementTmpTable();
  try {
    if( !m_schema.existsTable( tableName ) ){
      throw pool::ObjectRelationalException( "Expected  table \""+tableName+"\" does not exist in the schema." ,
                                             "ObjectRelationalMappingPersistency::storeMapping" );
      
    }
    
    // Store the mapping elements.
    coral::ITable& mappingTable = m_schema.tableHandle( tableName );
    coral::AttributeList  dataBuffer;
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::elementIdFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::elementTypeFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::scopeNameFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::variableNameFieldInMappingElementTable() );
    dataBuffer.extend< int >( pool::ObjectRelationalMappingSchema::variableParIndexFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::variableTypeFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::tableNameFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::columnNameFieldInMappingElementTable() );

    dataBuffer[ pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable() ].data<std::string>()= mapping.version();

    unsigned int iElementIndex = 0;
    std::stringstream mainElementId;
    mainElementId << iElementIndex;
    storeMappingElement( mainElementId.str(),mapping.element(),mappingTable.dataEditor(),dataBuffer );
    iElementIndex++;
    for ( std::map< std::string, pool::ObjectRelationalMappingElement >::const_iterator iDep = mapping.dependentElements().begin();
          iDep != mapping.dependentElements().end(); iDep++ ){
      std::stringstream elementId;
      elementId << iElementIndex;
      storeMappingElement( elementId.str(),iDep->second,mappingTable.dataEditor(),dataBuffer );
      iElementIndex++;
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::storeMapping" );    
  }
  return true;
}

bool
pool::ObjectRelationalMappingPersistency::storeMappingElement( const std::string& version,
                                                               const std::string& elementId,
                                                               const ObjectRelationalMappingElement& mappingElement ){
  std::string tableName = pool::ObjectRelationalMappingSchema::mappingElementTable();
  try {
    if( !m_schema.existsTable( tableName ) ){
      throw pool::ObjectRelationalException( "Expected  table \""+tableName+"\" does not exist in the schema." ,
                                             "ObjectRelationalMappingPersistency::storeMappingElement" );
      
    }
    
    // Store the mapping elements.
    coral::ITable& mappingTable = m_schema.tableHandle( tableName );
    coral::AttributeList  dataBuffer;
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::elementIdFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::elementTypeFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::scopeNameFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::variableNameFieldInMappingElementTable() );
    dataBuffer.extend< int >( pool::ObjectRelationalMappingSchema::variableParIndexFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::variableTypeFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::tableNameFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::columnNameFieldInMappingElementTable() );

    dataBuffer[ pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable() ].data<std::string>()= version;

    storeMappingElement( elementId,mappingElement,mappingTable.dataEditor(),dataBuffer );
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::storeMappingElement" );    
  }
  return true;  
}

bool
pool::ObjectRelationalMappingPersistency::storeMapping( const pool::ObjectRelationalMapping& mapping,
                                                        const std::string& classVersion,
                                                        std::string containerName  )
{
  coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );

  // Check whether the mapping version already exists.
  if ( ! m_versionsLoaded) loadMappingVersions();
  if ( m_storedMappings.find( mapping.version() ) != m_storedMappings.end() ) {
    log << coral::Error << "A mapping with version name \"" << mapping.version()
        << "\" is already stored in the database. No new mapping is stored." << coral::MessageStream::endmsg;
    return false;
  }

  try {
    
    // Store an entry into the version table
    coral::ITable& versionTable = m_schema.tableHandle( pool::ObjectRelationalMappingSchema::mappingVersionTable() );
    coral::AttributeList dataForVersionTable;
    dataForVersionTable.extend<std::string>(pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable());
    dataForVersionTable.extend<std::string>(pool::ObjectRelationalMappingSchema::containerNameFieldInMappingVersionTable());
    coral::AttributeList::iterator iAttribute = dataForVersionTable.begin();
    iAttribute->data< std::string >() = mapping.version();
    ++iAttribute;
    iAttribute->data< std::string >() = containerName;
    versionTable.dataEditor().insertRow( dataForVersionTable );

    // Store an entry into the class version table
    coral::ITable& classVersionTable = m_schema.tableHandle( pool::ObjectRelationalMappingSchema::classVersionTable() );
    coral::AttributeList dataForClassVersionTable;
    dataForClassVersionTable.extend<std::string>(pool::ObjectRelationalMappingSchema::mappingVersionFieldInClassVersionTable());
    dataForClassVersionTable.extend<std::string>(pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable());
    dataForClassVersionTable.extend<std::string>(pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable());
    coral::AttributeList::iterator iAttr = dataForClassVersionTable.begin();
    iAttr->data< std::string >() = mapping.version();
    ++iAttr;
    iAttr->data< std::string >() = classVersion;
    ++iAttr;
    iAttr->data< std::string >() = containerName;
    std::string existingVersion = this->selectMappingVersion( classVersion, containerName );
    if(existingVersion.empty()){
      classVersionTable.dataEditor().insertRow( dataForClassVersionTable );
    } else {
      std::string setClause = pool::ObjectRelationalMappingSchema::mappingVersionFieldInClassVersionTable()+" =:"+
        pool::ObjectRelationalMappingSchema::mappingVersionFieldInClassVersionTable();
      std::string whereClause = pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable()+" =:"+
        pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable()+" AND "+
        pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable()+" =:"+
        pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable();
      classVersionTable.dataEditor().updateRows(setClause,whereClause,dataForClassVersionTable);
    }

    // Store the mapping elements.
    coral::ITable& mappingTable = m_schema.tableHandle( pool::ObjectRelationalMappingSchema::mappingElementTable() );
    coral::AttributeList  dataBuffer;
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::elementIdFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::elementTypeFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::scopeNameFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::variableNameFieldInMappingElementTable() );
    dataBuffer.extend< int >( pool::ObjectRelationalMappingSchema::variableParIndexFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::variableTypeFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::tableNameFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::columnNameFieldInMappingElementTable() );

    dataBuffer[ pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable() ].data<std::string>()= mapping.version();

    unsigned int iElementIndex = 0;
    std::stringstream mainElementId;
    mainElementId << iElementIndex;
    pool::ObjectRelationalMappingPersistency::storeMappingElement( mainElementId.str(),
                                                                   mapping.element(),
                                                                   mappingTable.dataEditor(),
                                                                   dataBuffer );
    iElementIndex++;
    for ( std::map< std::string, pool::ObjectRelationalMappingElement >::const_iterator iDep = mapping.dependentElements().begin();
          iDep != mapping.dependentElements().end(); iDep++ ){
      std::stringstream elementId;
      elementId << iElementIndex;
      pool::ObjectRelationalMappingPersistency::storeMappingElement( elementId.str(),
                                                                     iDep->second,
                                                                     mappingTable.dataEditor(),
                                                                     dataBuffer );
      iElementIndex++;
    }
    
    m_storedMappings.insert(mapping.version());
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::storeMapping" );    
  }
  return true;
}

bool
pool::ObjectRelationalMappingPersistency::removeMapping( const std::string& version )
{
  coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
  if ( ! m_versionsLoaded) loadMappingVersions();
  if ( m_storedMappings.find( version ) == m_storedMappings.end() ) {
    // Print an error here
    log << coral::Warning << "A mapping with version name \"" << version
        << "\" does not exist in the database. Cannot remove it." << coral::MessageStream::endmsg;
    return false;
  }
  
  // Remove the mapping from the list
  m_storedMappings.erase( version );

  // Remove all rows in the tables with the version.
  coral::AttributeList whereData;
  whereData.extend<std::string>(pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable());
  whereData.begin()->data<std::string>() = version;

  try {
    std::string conditionForClassVersionTable = pool::ObjectRelationalMappingSchema::mappingVersionFieldInClassVersionTable() + " = :" + pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable();
    m_schema.tableHandle( pool::ObjectRelationalMappingSchema::classVersionTable() ).dataEditor().deleteRows( conditionForClassVersionTable, whereData );
    std::string conditionForElementTable = pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable() + "= :" + pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable() ;
    m_schema.tableHandle( pool::ObjectRelationalMappingSchema::mappingElementTable() ).dataEditor().deleteRows( conditionForElementTable, whereData );
    std::string conditionForVersionTable = pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable() + " = :" + pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable();
    m_schema.tableHandle( pool::ObjectRelationalMappingSchema::mappingVersionTable() ).dataEditor().deleteRows( conditionForVersionTable, whereData );
  } catch ( const coral::Exception& e){
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::removeMapping" );    
  }
  return true;
}


bool
pool::ObjectRelationalMappingPersistency::materializeMapping( const pool::ObjectRelationalMapping& mapping,
                                                              bool evolve,
                                                              bool dryRun )
{
  // Checking whether the mapping exists.
  bool ret = false;
  try {
    pool::ObjectRelationalAccess::MappingMaterialization materialization( m_schema );
    ret = materialization.materialize( mapping, evolve, dryRun );
    if(dryRun){
      coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
      const std::set<std::string>& logs = materialization.operationList();
      if(logs.empty()) log << coral::Warning << "# No changes on the schema are required." <<coral::MessageStream::endmsg;
      for( std::set<std::string>::const_iterator iLog = logs.begin();
           iLog!=logs.end();++iLog){
        log << coral::Warning << "# Schema change: "<<*iLog<<coral::MessageStream::endmsg;
      }
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::materializeMapping" );
  }
  return ret;
}

bool
pool::ObjectRelationalMappingPersistency::materializeMappingElement( const pool::ObjectRelationalMappingElement& mappingElement,
                                                                     bool evolve,
                                                                     bool dryRun )
{
  // Checking whether the mapping exists.
  bool ret = false;
  try {
    pool::ObjectRelationalAccess::MappingMaterialization materialization( m_schema );
    ret = materialization.materializeElement( mappingElement, evolve, dryRun );
    if(dryRun){
      coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
      const std::set<std::string>& logs = materialization.operationList();
      if(logs.empty()) log << coral::Warning << "# No changes on the schema are required." <<coral::MessageStream::endmsg;
      for( std::set<std::string>::const_iterator iLog = logs.begin();
           iLog!=logs.end();++iLog){
        log << coral::Warning << "# Schema change: "<<*iLog<<coral::MessageStream::endmsg;
      }
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::materializeMapping" );
  }
  return ret;
}


bool 
pool::ObjectRelationalMappingPersistency::removeMaterializedTables(const std::string& version, bool dryRun ){

  coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
  try {
    if ( ! m_versionsLoaded) loadMappingVersions();
    if ( m_storedMappings.find( version ) == m_storedMappings.end() ) {
      // Print an error here
      log << coral::Warning << "A mapping with version name \"" << version
          << "\" does not exist in the database. Cannot remove it." << coral::MessageStream::endmsg;
      return false;
    }
    pool::ObjectRelationalMapping* fullMapping = retrieveMapping( version );
    if(!fullMapping) return false;
    std::set<std::string> tableList;
    makeTableList(fullMapping->element(),tableList);
    for( std::map< std::string, pool::ObjectRelationalMappingElement >::const_iterator iDep = fullMapping->dependentElements().begin();
         iDep != fullMapping->dependentElements().end(); iDep++ ){
      makeTableList(iDep->second,tableList);
    }
    if(tableList.size()>0){
      log << coral::Info << " Dropping tables referenced by mapping version \"" << version << "\"."<<coral::MessageStream::endmsg;
    } else {
      log << coral::Info << " No table found referenced by mapping version \"" << version << "\"."<<coral::MessageStream::endmsg;
      delete fullMapping;
      return false;
    }
    
    for(std::set<std::string>::reverse_iterator iTable=tableList.rbegin();iTable!=tableList.rend();iTable++){

      // check if some other mapping version is referencing the same table
      std::auto_ptr<coral::IQuery> query( m_schema.tableHandle( pool::ObjectRelationalMappingSchema::mappingElementTable() ).newQuery() );
      query->addToOutputList( pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable() );
      std::ostringstream condition;
      condition << pool::ObjectRelationalMappingSchema::tableNameFieldInMappingElementTable() << " =:" <<
        pool::ObjectRelationalMappingSchema::tableNameFieldInMappingElementTable();
      coral::AttributeList condData;
      condData.extend<std::string>(pool::ObjectRelationalMappingSchema::tableNameFieldInMappingElementTable());
      coral::AttributeList::iterator iAttribute = condData.begin();
      iAttribute->data< std::string >() = *iTable;
      query->setCondition( condition.str(), condData );
      std::set<std::string> versions;
      coral::ICursor& cursor = query->execute();
      while ( cursor.next() ) {
        const coral::AttributeList& currentRow = cursor.currentRow();
        versions.insert(currentRow[pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable()].data<std::string>());
      }
      versions.erase( version );
      coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
      if(m_schema.existsTable(*iTable)){
        if(versions.empty()){
          log << coral::Warning << "Dropping table \"" << *iTable <<"\"" << coral::MessageStream::endmsg;
          if(!dryRun) m_schema.dropTable(*iTable);
        } else{
          log << coral::Info << "Table \"" << *iTable <<"\" is currenlty used by other mapping versions." << coral::MessageStream::endmsg;
        }
      }
    }
    delete fullMapping;
  } catch (const coral::Exception& e){
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::removeMaterializedTables" );
  }
  return true;
}

