#include "ObjectRelationalAccess/ObjectRelationalMapping.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingUtilities.h"
#include "ObjectRelationalAccess/ObjectRelationalDatabaseDescriptor.h"
#include "ObjectRelationalAccess/ObjectRelationalException.h"
#include "MappingBuilder.h"
#include "XMLDatabaseManagerDriver.h"
#include "RelationalAccess/ISessionProxy.h"
#include "RelationalAccess/ITransaction.h"
#include <fstream>
#include <iomanip>
#include <memory>

pool::ObjectRelationalMappingUtilities::ObjectRelationalMappingUtilities() : ObjectRelationalUtilities(),
                                                                             m_mappingBuffer(0),
                                                                             m_versionBuffer(0){}

pool::ObjectRelationalMappingUtilities::ObjectRelationalMappingUtilities(coral::ISessionProxy* session) : ObjectRelationalUtilities(session),
                                                                             m_mappingBuffer(0),
                                                                             m_versionBuffer(0){}


pool::ObjectRelationalMappingUtilities::~ObjectRelationalMappingUtilities(){
  cleanBuffers();
}

void pool::ObjectRelationalMappingUtilities::cleanBuffers(){
  if(m_versionBuffer) delete m_versionBuffer;
  if(m_mappingBuffer) {
    for(std::map<std::string,ObjectRelationalMapping*>::const_iterator iM = m_mappingBuffer->begin();
        iM != m_mappingBuffer->end(); ++iM ){
      delete (iM->second);
    }
    delete m_mappingBuffer;
  }  
}

void pool::ObjectRelationalMappingUtilities::buildAndMaterializeMapping(const std::string& fileName,
                                                                        const std::string& outFile,
                                                                        bool evolve,
                                                                        bool dryRun ){
  initialize();
  coral::ISessionProxy* session = this->session();
  if(!session) throw pool::ObjectRelationalException("No valid connection to database.",
                                                     "ObjectRelationalMappingUtilities::buildAndMaterializeMapping");
  pool::MappingBuilder builder( session->nominalSchema() );
  std::auto_ptr<ObjectRelationalDatabaseDescriptor> dbstruct(loadDescriptor(fileName));
  // loop over containers
  for(std::vector<ObjectRelationalContainerDescriptor*>::const_iterator iContainer=dbstruct->containers().begin();
      iContainer!=dbstruct->containers().end();
      iContainer++){
    if( (*iContainer)->name().empty() ) throw ObjectRelationalException("Container description found with empty \"name\" attribute.","ObjectRelationalMappingUtilities::buildAndMaterializeMapping");
    for(std::vector<ObjectRelationalClassDescriptor*>::const_iterator iClass=(*iContainer)->classStores().begin();
        iClass!=(*iContainer)->classStores().end();
        iClass++){
      const ObjectRelationalMappingDescriptor* mappDescr = (*iClass)->mapping();
      if(mappDescr){
        builder.buildMapping(mappDescr);
      }
    }
  }
  builder.finalizeMapping(evolve,dryRun);
  if(!outFile.empty()){
    builder.exportMapping(outFile);
  }
}

void pool::ObjectRelationalMappingUtilities::buildAndMaterializeMapping(const ObjectRelationalDatabaseDescriptor& dbstruct,
                                                                        const std::string& outFile,
                                                                        bool evolve,
                                                                        bool dryRun ){
  initialize();
  coral::ISessionProxy* session = this->session();
  if(!session) throw pool::ObjectRelationalException("No valid connection to database.",
                                                     "ObjectRelationalMappingUtilities::buildAndMaterializeMapping");
  pool::MappingBuilder builder( session->nominalSchema() );
  // loop over containers
  for(std::vector<ObjectRelationalContainerDescriptor*>::const_iterator iContainer=dbstruct.containers().begin();
      iContainer!=dbstruct.containers().end();
      iContainer++){
    if( (*iContainer)->name().empty() ) throw ObjectRelationalException("Container description found with empty \"name\" attribute.","ObjectRelationalMappingUtilities::buildAndMaterializeMapping");
    for(std::vector<ObjectRelationalClassDescriptor*>::const_iterator iClass=(*iContainer)->classStores().begin();
        iClass!=(*iContainer)->classStores().end();
        iClass++){
      const ObjectRelationalMappingDescriptor* mappDescr = (*iClass)->mapping();
      if(mappDescr){
        builder.buildMapping(mappDescr);
      }
    }
  }
  builder.finalizeMapping(evolve,dryRun);
  if(!outFile.empty()){
    builder.exportMapping(outFile);
  }
}


void pool::ObjectRelationalMappingUtilities::buildAndMaterializeMappingFromBuffer(const char* buffer,
                                                                                  bool evolve,
                                                                                  bool dryRun){
  initialize();
  coral::ISessionProxy* session = this->session();
  if(!session) throw pool::ObjectRelationalException("No valid connection to database.",
                                                     "ObjectRelationalMappingUtilities::buildAndMaterializeMappingFromBuffer");
  pool::MappingBuilder builder( session->nominalSchema() );
  std::auto_ptr<ObjectRelationalDatabaseDescriptor> dbstruct(loadDescriptorFromBuffer(buffer));
  // loop over containers
  for(std::vector<ObjectRelationalContainerDescriptor*>::const_iterator iContainer=dbstruct->containers().begin();
      iContainer!=dbstruct->containers().end();
      iContainer++){
    if( (*iContainer)->name().empty() ) throw ObjectRelationalException("Container description found with empty \"name\" attribute.","ObjectRelationalMappingUtilities::buildAndMaterializeMappingFromBuffer");
    for(std::vector<ObjectRelationalClassDescriptor*>::const_iterator iClass=(*iContainer)->classStores().begin();
        iClass!=(*iContainer)->classStores().end();
        iClass++){
      const ObjectRelationalMappingDescriptor* mappDescr = (*iClass)->mapping();
      if(mappDescr){
        builder.buildMapping(mappDescr);
      }
    }
  }
  builder.finalizeMapping(evolve,dryRun);
}

void pool::ObjectRelationalMappingUtilities::retrieveMapping(const std::string& version,const std::string& fileName){
  initialize();
  coral::ISessionProxy* session = this->session();
  if(!session) throw pool::ObjectRelationalException("No valid connection to database.","ObjectRelationalMappingUtilities::retrieveMapping");
  pool::MappingBuilder builder( session->nominalSchema() );
  builder.retrieveMapping(version,fileName);
}

void pool::ObjectRelationalMappingUtilities::retrieveMapping(const std::string& fileName, bool fromOldSchema){
  initialize();
  coral::ISessionProxy* session = this->session();
  if(!session) throw pool::ObjectRelationalException("No valid connection to database.","ObjectRelationalMappingUtilities::retrieveMapping");
  pool::MappingBuilder builder( session->nominalSchema() );
  builder.retrieveMapping(fileName,fromOldSchema);
}

std::string
pool::ObjectRelationalMappingUtilities::getContainerMapping(const std::string& containerName, std::string classVersion){
  initialize();
  coral::ISessionProxy* session = this->session();
  if(!session) throw pool::ObjectRelationalException("No valid connection to database.","ObjectRelationalMappingUtilities::retrieveMapping");
  pool::MappingBuilder builder( session->nominalSchema() );
  std::stringstream mapStr;
  builder.retrieveContainerMapping( mapStr, containerName, classVersion );
  return mapStr.str();
}

void
pool::ObjectRelationalMappingUtilities::retrieveContainerMapping(const std::string& containerName, const std::string& fileName, std::string classVersion){
  initialize();
  coral::ISessionProxy* session = this->session();
  if(!session) throw pool::ObjectRelationalException("No valid connection to database.","ObjectRelationalMappingUtilities::retrieveMapping");
  pool::MappingBuilder builder( session->nominalSchema() );
  std::ofstream outFile;
  outFile.open (fileName.c_str(), std::fstream::out);
  builder.retrieveContainerMapping( outFile, containerName, classVersion );
  outFile.close();
}

std::string
pool::ObjectRelationalMappingUtilities::getAllContainerMappings(const std::string& containerName){
  initialize();
  coral::ISessionProxy* session = this->session();
  if(!session) throw pool::ObjectRelationalException("No valid connection to database.","ObjectRelationalMappingUtilities::retrieveMapping");
  pool::MappingBuilder builder( session->nominalSchema() );
  std::stringstream mapStr;
  builder.retrieveAllContainerMappings( mapStr, containerName );
  return mapStr.str();
}

void
pool::ObjectRelationalMappingUtilities::retrieveAllContainerMappings(const std::string& containerName, const std::string& fileName){
  initialize();
  coral::ISessionProxy* session = this->session();
  if(!session) throw pool::ObjectRelationalException("No valid connection to database.","ObjectRelationalMappingUtilities::retrieveMapping");
  pool::MappingBuilder builder( session->nominalSchema() );
  std::ofstream outFile;
  outFile.open (fileName.c_str(), std::fstream::out);
  builder.retrieveAllContainerMappings( outFile, containerName );
  outFile.close();
}

std::map<std::pair<std::string,std::string>, std::string >
pool::ObjectRelationalMappingUtilities::mappingForClassVersions(){
  initialize();
  coral::ISessionProxy* session = this->session();
  if(!session) throw pool::ObjectRelationalException("No valid connection to database.","ObjectRelationalMappingUtilities::mappingForClassVersions");
  pool::MappingBuilder builder( session->nominalSchema() );
  return builder.mappingForClassVersions();  
}

std::map<std::string,std::string>
pool::ObjectRelationalMappingUtilities::listMappings(){
  initialize();
  coral::ISessionProxy* session = this->session();
  if(!session) throw pool::ObjectRelationalException("No valid connection to database.","ObjectRelationalMappingUtilities::listMappings");
  pool::MappingBuilder builder( session->nominalSchema() );
  return builder.mappingVersions();
}

std::map<std::string, std::string>
pool::ObjectRelationalMappingUtilities::containerTableMap(){
   initialize();
  coral::ISessionProxy* session = this->session();
  if(!session) throw pool::ObjectRelationalException("No valid connection to database.","ObjectRelationalMappingUtilities::listMappings");
  pool::MappingBuilder builder( session->nominalSchema() );
  return builder.containerTableMap(); 
}

void
pool::ObjectRelationalMappingUtilities::selectMapping(const std::string& version,
                                                      std::map<std::pair<std::string,std::string>,std::pair<std::string,std::string> >& mappingStruct){
  initialize();
  coral::ISessionProxy* session = this->session();
  if(!session) throw pool::ObjectRelationalException("No valid connection to database.","ObjectRelationalMappingUtilities::listMappings");
  pool::MappingBuilder builder( session->nominalSchema() );
  builder.selectMapping(version,mappingStruct);
}

bool pool::ObjectRelationalMappingUtilities::existsMapping(const std::string& version){
  initialize();
  coral::ISessionProxy* session = this->session();
  if(!session) throw pool::ObjectRelationalException("No valid connection to database.","ObjectRelationalMappingUtilities::existsMapping");
  pool::MappingBuilder builder( session->nominalSchema() );
  std::map<std::string,std::string> versionList = builder.mappingVersions();
  return (versionList.find(version)!=versionList.end());
}

bool pool::ObjectRelationalMappingUtilities::removeMapping(const std::string& version, bool removeTables, bool dryRun){
  initialize();
  coral::ISessionProxy* session = this->session();
  if(!session) throw pool::ObjectRelationalException("No valid connection to database.","ObjectRelationalMappingUtilities::removeMapping");
  pool::MappingBuilder builder( session->nominalSchema() );
  return builder.removeMapping(version,removeTables,dryRun);
}

bool
pool::ObjectRelationalMappingUtilities::upgradeMappingSchema(){
  initialize();
  coral::ISessionProxy* session = this->session();
  if(!session) throw pool::ObjectRelationalException("No valid connection to database.","ObjectRelationalMappingUtilities::upgradeMappingSchema");
  pool::MappingBuilder builder( session->nominalSchema() );
  return builder.upgradeMappingSchema();
}

pool::ObjectRelationalDatabaseDescriptor*
pool::ObjectRelationalMappingUtilities::loadDescriptor(const std::string fileName){
  XMLDatabaseManagerDriver driver;
  ObjectRelationalDatabaseDescriptor*  dbstruct = driver.read(fileName);
  return dbstruct;
}

pool::ObjectRelationalDatabaseDescriptor*
pool::ObjectRelationalMappingUtilities::loadDescriptorFromBuffer(const char* buffer){
  XMLDatabaseManagerDriver driver;
  ObjectRelationalDatabaseDescriptor*  dbstruct = driver.readFromBuffer(buffer);
  return dbstruct;
}

bool
pool::ObjectRelationalMappingUtilities::loadMappingInformation( const std::string& containerName, const std::string& classVersion, bool all){
  initialize();
  coral::ISessionProxy* session = this->session();
  if(!session) throw pool::ObjectRelationalException("No valid connection to database.","ObjectRelationalMappingUtilities::upgradeMappingSchema");
  pool::MappingBuilder builder( session->nominalSchema() );
  cleanBuffers();
  m_versionBuffer = new std::map<std::string,std::pair<std::string,std::string> >;
  m_mappingBuffer = new std::map<std::string,ObjectRelationalMapping*>;  
  if(all){
    std::map<std::string,std::string> allVersions = listMappings();
    for(std::map<std::string,std::string>::const_iterator iV = allVersions.begin();
       iV != allVersions.end(); ++iV){
      if( iV->second == containerName )
        m_versionBuffer->insert(std::make_pair( iV->first, std::make_pair(iV->second,"") ));
    }
  }
  {
    std::string clVer = classVersion;
    std::map<std::pair<std::string,std::string>, std::string > classVersions =  mappingForClassVersions();
    if(!all){
      if( clVer.empty() ){
        clVer = builder.classVersion( containerName );
      }
      std::string version("");
      std::map<std::pair<std::string,std::string>, std::string >::const_iterator iCv = classVersions.find(std::make_pair(clVer,containerName));
      if(iCv != classVersions.end()){
        version = iCv->second;
      }
      if( version.empty()){
        throw pool::ObjectRelationalException("No mapping found for class version \""+clVer+"\"",
                                              "ObjectRelationalMappingUtilities::loadMappingInformation");
        
      }
      m_versionBuffer->insert( std::make_pair(version,std::make_pair( containerName, clVer) ));
    } else {
      for(std::map<std::pair<std::string,std::string>, std::string >::const_iterator iCv = classVersions.begin();
          iCv != classVersions.end(); iCv++){
        if( iCv->first.second == containerName ){
          std::map<std::string,std::pair<std::string,std::string> >::iterator iV = m_versionBuffer->find( iCv->second );
          if( iV == m_versionBuffer->end() ){
            throw pool::ObjectRelationalException("Database inconsistency detected.",
                                                  "ObjectRelationalMappingUtilities::loadMappingInformation");
          }
          iV->second.second = iCv->first.first;
        }
      }
    }
  }
  for( std::map<std::string,std::pair<std::string,std::string> >::const_iterator iV = m_versionBuffer->begin();
       iV != m_versionBuffer->end(); iV++){
    ObjectRelationalMapping* mapping = builder.retrieveMapping( iV->first );
    m_mappingBuffer->insert( std::make_pair( iV->first,mapping)  );
  }
  return true;
}

bool pool::ObjectRelationalMappingUtilities::storeMappingInformation( bool materialize ){
  initialize();
  coral::ISessionProxy* session = this->session();
  if(!session) throw pool::ObjectRelationalException("No valid connection to database.","ObjectRelationalMappingUtilities::storeMappingInformation");
  pool::MappingBuilder builder( session->nominalSchema() );
  if(!m_mappingBuffer || !m_versionBuffer) return false;
  for( std::map<std::string,std::pair<std::string,std::string> >::const_iterator iV = m_versionBuffer->begin();
       iV != m_versionBuffer->end(); ++iV ){
    std::map<std::string,ObjectRelationalMapping*>::const_iterator iM = m_mappingBuffer->find( iV->first );
    if(!builder.storeMapping( iV->second.first ,*(iM->second), iV->second.second, materialize )){
      throw pool::ObjectRelationalException("Could not store the mapping.","ObjectRelationalMappingUtilities::storeMappingInformation");
    };
  }
  return true;
}

void
pool::ObjectRelationalMappingUtilities::prettyPrint(const std::map<std::string,std::string>& data,
                                                    const std::string& firstColumn,const std::string& secondColumn, std::ostream& out ){
  size_t firstColMax = firstColumn.length();
  size_t secondColMax = secondColumn.length();
  // first find the max lenghts
  for(std::map<std::string,std::string>::const_iterator iR = data.begin();iR!=data.end();++iR){
    if(iR->first.length() > firstColMax) firstColMax = iR->first.length();
    if(iR->second.length() > secondColMax) secondColMax = iR->second.length();
  }

  // then print formatted
  std::string space("   ");
  out << std::setiosflags(std::ios_base::left);
  out << space << std::setw(firstColMax)<<firstColumn;
  out << space << std::setw(secondColMax)<<secondColumn;
  out << std::endl;
  out << space << std::setfill('-')<<std::setw(firstColMax)<<"";
  out << space <<std::setw(secondColMax)<<"";
  out << std::endl;
  out << std::setfill(' ');
  for(std::map<std::string,std::string>::const_iterator iR = data.begin();iR!=data.end();++iR){
    out << space << std::setw(firstColMax)<<iR->first;
    out << space << std::setw(secondColMax)<<iR->second;
    out << std::endl;
  }
}

void
pool::ObjectRelationalMappingUtilities::prettyPrint(const std::map<std::pair<std::string,std::string>,std::string>& data,
                                                    const std::string& firstColumn,
                                                    const std::string& secondColumn,
                                                    const std::string& thirdColumn,
                                                    std::ostream& out ){
  size_t firstColMax = firstColumn.length();
  size_t secondColMax = secondColumn.length();
  size_t thirdColMax = thirdColumn.length();
  // first find the max lenghts
  for(std::map<std::pair<std::string,std::string>, std::string>::const_iterator iR = data.begin();iR!=data.end();++iR){
    if(iR->first.first.length() > firstColMax) firstColMax = iR->first.first.length();
    if(iR->first.second.length() > secondColMax) secondColMax = iR->first.second.length();
    if(iR->second.length() > thirdColMax) thirdColMax = iR->second.length();
  }

  // then print formatted
  std::string space("   ");
  out << std::setiosflags(std::ios_base::left);
  out << space << std::setw(firstColMax)<<firstColumn;
  out << space << std::setw(secondColMax)<<secondColumn;
  out << space << std::setw(thirdColMax)<<thirdColumn;
  out << std::endl;
  out << space << std::setfill('-')<<std::setw(firstColMax)<<"";
  out << space <<std::setw(secondColMax)<<"";
  out << space <<std::setw(thirdColMax)<<"";
  out << std::endl;
  out << std::setfill(' ');
  for(std::map<std::pair<std::string,std::string>, std::string>::const_iterator iR = data.begin();iR!=data.end();++iR){
    out << space << std::setw(firstColMax)<<iR->first.first;
    out << space << std::setw(secondColMax)<<iR->first.second;
     out << space << std::setw(thirdColMax)<<iR->second;
   out << std::endl;
  }
}

void
pool::ObjectRelationalMappingUtilities::prettyPrint(const std::map<std::pair<std::string,std::string>,std::pair<std::string,std::string> >& data,
                 const std::string& firstColumn,const std::string& secondColumn,
                 const std::string& thirdColumn,const std::string& fourthColumn, std::ostream& out ){
  
  size_t firstColMax = firstColumn.length();
  size_t secondColMax = secondColumn.length();
  size_t thirdColMax = thirdColumn.length();
  size_t fourthColMax = fourthColumn.length();
  // first find the max lenghts
  for(std::map<std::pair<std::string,std::string>, std::pair<std::string,std::string> >::const_iterator iR = data.begin();iR!=data.end();++iR){
    if(iR->first.first.length() > firstColMax) firstColMax = iR->first.first.length();
    if(iR->first.second.length() > secondColMax) secondColMax = iR->first.second.length();
    if(iR->second.first.length() > thirdColMax) thirdColMax = iR->second.first.length();
    if(iR->second.second.length() > fourthColMax) fourthColMax = iR->second.second.length();
  }

  // then print formatted
  std::string space("   ");
  out << std::setiosflags(std::ios_base::left);
  out << space << std::setw(firstColMax)<<firstColumn;
  out << space << std::setw(secondColMax)<<secondColumn;
  out << space << std::setw(thirdColMax)<<thirdColumn;
  out << space << std::setw(fourthColMax)<<fourthColumn;
  out << std::endl;
  out << space << std::setfill('-')<<std::setw(firstColMax)<<"";
  out << space <<std::setw(secondColMax)<<"";
  out << space <<std::setw(thirdColMax)<<"";
  out << space <<std::setw(fourthColMax)<<"";
  out << std::endl;
  out << std::setfill(' ');
  for(std::map<std::pair<std::string,std::string>,std::pair<std::string,std::string> >::const_iterator iR = data.begin();iR!=data.end();++iR){
    out << space << std::setw(firstColMax)<<iR->first.first;
    out << space << std::setw(secondColMax)<<iR->first.second;
     out << space << std::setw(thirdColMax)<<iR->second.first;
     out << space << std::setw(fourthColMax)<<iR->second.second;
   out << std::endl;
  }
}

std::string pool::ObjectRelationalMappingUtilities::keyOfMappingList(){
  static std::string s_ret("mapping_version");
  return s_ret;
}

std::string pool::ObjectRelationalMappingUtilities::valueOfMappingList(){
  static std::string s_ret("container_name");
  return s_ret;
}

std::string pool::ObjectRelationalMappingUtilities::keyFirstOfClassVersionList(){
  static std::string s_ret("class_version");
  return s_ret;
}

std::string pool::ObjectRelationalMappingUtilities::keySecondOfClassVersionList(){
  static std::string s_ret("container_name");
  return s_ret;
}

std::string pool::ObjectRelationalMappingUtilities::valueOfClassVersionList(){
  static std::string s_ret("mapping_version");
  return s_ret;
}

std::string pool::ObjectRelationalMappingUtilities::keyOfContainerTableMap(){
  static std::string s_ret("table_name");
  return s_ret;
}

std::string pool::ObjectRelationalMappingUtilities::valueOfContainerTableMap(){
  static std::string s_ret("container_name");
  return s_ret;
}

