#include <iostream>
#include <stdexcept>
#include <memory>

#include "StorageSvc/IStorageSvc.h"
#include "StorageSvc/IStorageExplorer.h"
#include "StorageSvc/DbType.h"
#include "StorageSvc/DbOption.h"
#include "StorageSvc/FileDescriptor.h"
#include "StorageSvc/poolDb.h"


class ExtractFileIdentifierApplication {
public:
  ExtractFileIdentifierApplication( int argc, char* argv[] );
  ~ExtractFileIdentifierApplication(){}
  bool parseArguments();
  void execute();
  void printSyntax();
private:
  const std::string executableName;
  std::string technologyName;
  std::vector< std::string > fileNames;
};

ExtractFileIdentifierApplication::ExtractFileIdentifierApplication( int argc, char* argv[] ):
  executableName( std::string( argv[0] ) ),
  technologyName( pool::ROOT_StorageType.storageName() ),
  fileNames()
{
  for ( int i = 1; i < argc; ++i )
    fileNames.push_back( std::string( argv[i] ) );
}


bool
ExtractFileIdentifierApplication::parseArguments()
{
  std::vector< std::string > theFiles;
  const std::vector< std::string >& args = fileNames;
  unsigned int excludedArgument = 0;
  for ( unsigned int iArg = 0; iArg < args.size(); ++iArg ) {
    if ( iArg > 0 && iArg == excludedArgument )
      continue;
    const std::string& arg = args[iArg];
    if ( arg == "-t" ) {
      unsigned int nextArgumentIndex = iArg + 1;
      if ( nextArgumentIndex < args.size() ) {
        excludedArgument = nextArgumentIndex;
        technologyName = args[nextArgumentIndex];
      }
    }
    else {
      theFiles.push_back( arg );
    }
  }

  fileNames = theFiles;

  // Check now the values.
  if ( fileNames.empty() ) return false;
  pool::DbType theType = pool::DbType::getType( technologyName );
  if ( theType.match( pool::TEST_StorageType ) ) {
    throw std::runtime_error( "Unsupported file type : " + technologyName );
  }
  return true;
}


void
ExtractFileIdentifierApplication::execute()
{
  // Do the job here
  pool::IStorageSvc* storageSvc = pool::createStorageSvc(0, "StorageSvc");
  if ( ! storageSvc ) {
    throw std::runtime_error( "Could not create a StorageSvc object" );
  }
  storageSvc->addRef();
  void* pVoid = 0;
  pool::DbStatus sc = storageSvc->queryInterface( pool::IStorageExplorer::interfaceID(),&pVoid);
  pool::IStorageExplorer* storageExplorer = (pool::IStorageExplorer*)pVoid;
  if ( !( sc.isSuccess() && storageExplorer ) ) {
    storageSvc->release();
    throw std::runtime_error( "Could not retrieve a IStorageExplorer interface" );
  }

  // Retrieve the fids.
  try {
    pool::Session* session = 0;
    long technologyId = pool::DbType::getType( technologyName ).majorType();
    if ( ! storageExplorer->startSession( poolDb::READ, technologyId, session ).isSuccess() ) {
      throw std::runtime_error( "Could not start a new session" );
    }
    if( technologyId == pool::ROOT_StorageType.majorType() ) {
       // Set ERROR_LEVEL ROOT domain option to kError (==3000)
       // to disable warnings about unknown classes when opening the file
       // ----  (NOT needed when nostreamers option is used)
       pool::DbOption        err_lvl("ERROR_LEVEL", "", 3000); 
       storageExplorer->setDomainOption(session, err_lvl);
       
       // Tell ROOT to not autoload dictionaries (speedup)
       pool::DbOption        no_streamers("FILE_READSTREAMERINFO", "", 0);
       //  (seems to break ROOT 5.20 ability to read 5.18 files)
       //storageExplorer->setDomainOption(session, no_streamers);
    }
    for ( std::vector< std::string >::const_iterator iFile = fileNames.begin();
          iFile != fileNames.end(); ++iFile ) {
      const std::string& pfn = *iFile;
      std::string fid = "";
      pool::FileDescriptor fd( fid, pfn );
      if ( ! storageExplorer->connect( session, poolDb::READ, fd ).isSuccess() ) {
        throw std::runtime_error( "Could not open the file \"" + pfn + "\"" );
      }
      if ( ! storageExplorer->dbParam( fd, "FID", fid ).isSuccess() ) {
        storageExplorer->disconnect( fd );
        throw std::runtime_error( "Could not retrieve the FID from file \"" + pfn + "\"" );
      }
      std::cout << fid << " (" << pfn << ")" << std::endl;
      storageExplorer->disconnect( fd );
    }
    storageExplorer->endSession( session );
  }
  catch ( std::exception& error ) {
    std::cerr << error.what() << std::endl;
  }
  catch ( ... ) {
    std::cerr << "Funny error ..." << std::endl;
  }

  storageExplorer->release();
  storageSvc->release();
}


void
ExtractFileIdentifierApplication::printSyntax()
{
  std::cerr << "Syntax : " << executableName << " [-t technologyType] files" << std::endl;
}




int main( int argc, char* argv[] )
{
  try {
    ExtractFileIdentifierApplication app( argc, argv );
    if ( app.parseArguments() ) {
      app.execute();
    }
    else app.printSyntax();
  }
  catch ( std::exception& error ) {
    std::cerr << error.what() << std::endl;
    return 1;
  }
  catch ( ... ) {
    std::cerr << "Funny error ..." << std::endl;
    return 1;
  }
  return 0;
}
