#include "ObjectRelationalReadDataBuffer.h"
#include "RelationalStorageServiceException.h"
#include "ContainerHandlerFactory.h"
#include "IContainerHandler.h"
#include "POOLCore/Context.h"
#include "POOLCore/IBlobStreamingService.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingElement.h"
#include "ObjectRelationalAccess/ObjectRelationalClassUtils.h"
#include "ObjectRelationalAccess/ObjectRelationalDefaultMappingRules.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITable.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/AttributeListException.h"
#include "CoralBase/AttributeListSpecification.h"
#include "CoralBase/AttributeSpecification.h"
#include "CoralBase/Blob.h"
#include "Reflex/Member.h"

#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
using namespace ROOT;
#endif

pool::RelationalStorageService::ObjectRelationalReadDataBuffer::ObjectRelationalReadDataBuffer( pool::Context* context):
  m_context(context),
  m_spec( 0 ),
  m_offsets(),
  m_metadata(),
  m_parentMetadata(0),
  m_parentMetadataOffset(0),
  m_table( 0 ),
  m_dictionary( 0 ),
  m_offsetToParent( 0 ),
  m_keyItems( 0 ),
  m_mapping( 0 ),
  m_keyData( false ),
  m_containerHandler( 0 ),
  m_arrayBuffers(),
  m_whereData( 0 ),
  m_whereClause(""),
  m_blobReaders(),
  m_poolPointerBuffers(),
  m_dependentObjectBuffers()
{}



pool::RelationalStorageService::ObjectRelationalReadDataBuffer::~ObjectRelationalReadDataBuffer()
{
  for ( std::vector< pool::RelationalStorageService::ObjectRelationalReadDataBuffer* >::iterator iBuffer = m_arrayBuffers.begin();
        iBuffer != m_arrayBuffers.end(); ++iBuffer ) {
    delete *iBuffer;
  }
  m_arrayBuffers.clear();
  for ( std::vector< pool::RelationalStorageService::ObjectRelationalReadDataBuffer* >::iterator iBuffer = m_poolPointerBuffers.begin();
        iBuffer != m_poolPointerBuffers.end(); ++iBuffer ) {
    delete *iBuffer;
  }
  m_poolPointerBuffers.clear();
  for ( std::vector< pool::RelationalStorageService::ObjectRelationalReadDataBuffer* >::iterator iBuffer = m_poolPolymorphicPointerBuffers.begin();
        iBuffer != m_poolPolymorphicPointerBuffers.end(); ++iBuffer ) {
    delete *iBuffer;
  }
  m_poolPolymorphicPointerBuffers.clear();
  for ( std::vector< pool::RelationalStorageService::ObjectRelationalReadDataBuffer* >::iterator iBuffer = m_dependentObjectBuffers.begin();
        iBuffer != m_dependentObjectBuffers.end(); ++iBuffer ) {
    delete *iBuffer;
  }
  m_dependentObjectBuffers.clear();

  if ( m_whereData ) delete m_whereData;

  if ( m_containerHandler ) delete m_containerHandler;

  if ( m_spec ) m_spec->release();
  for( std::map< int,pool::IBlobReader* >::iterator iBlobReader=m_blobReaders.begin();
       iBlobReader!=m_blobReaders.end();
       iBlobReader++)
  {
    delete iBlobReader->second;
  }
}

bool
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::build( coral::ISchema& schema,
                                                                       const pool::ObjectRelationalMappingElement& mapping,
                                                                       const TypeH& dict )
{
  m_spec = new coral::AttributeListSpecification;
  m_table = &(schema.tableHandle( mapping.tableName() ) );
  m_whereData = new coral::AttributeList;
  // resolve possible typedef chains
  TypeH dictionary = ObjectRelationalClassUtils::resolvedType( dict );
  m_dictionary = dictionary;
  m_mapping = &mapping;

  this->buildBuffer( mapping, dictionary, schema, 0, false );
  return true;
}

pool::RelationalStorageService::ObjectRelationalReadDataBuffer&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::addDependency()
{
  ObjectRelationalReadDataBuffer* newBuffer = new ObjectRelationalReadDataBuffer(m_context);
  newBuffer->m_metadata = m_metadata;
  newBuffer->m_parentMetadata = m_parentMetadata;
  newBuffer->m_parentMetadataOffset = m_parentMetadataOffset;
  m_dependentObjectBuffers.push_back( newBuffer );
  return *newBuffer;
}

int
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::buildBuffer( const pool::ObjectRelationalMappingElement& mapping,
                                                                             const TypeH& dictionary,
                                                                             coral::ISchema& schema,
                                                                             int offset,
                                                                             bool keyData )
{
  int retVal = 0;
  if ( mapping.elementType() == pool::ObjectRelationalMappingElement::Primitive ) { // Primitives
    if ( !( dictionary.IsFundamental() || dictionary.TypeInfo() == typeid(std::string) || dictionary.IsEnum() ) ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + mapping.variableName() +
                                                                               "\" is not declared as primitive in the mapping element",
                                                                               "ObjectRelationalReadDataBuffer::buildBuffer" );
    }
    if(m_spec->index( mapping.columnNames()[0] )==-1){
      const std::type_info* attrType = &dictionary.TypeInfo();
      if(dictionary.IsEnum()) attrType = &typeid(int);
      m_spec->extend( mapping.columnNames()[0], *attrType );
    }
    m_offsets.push_back( std::make_pair(offset,-1) );
    ++retVal;
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::Blob ) { // Blobs
    if(m_spec->index( mapping.columnNames()[0] )==-1){
      m_spec->extend<coral::Blob>( mapping.columnNames()[0] );
    }
    m_offsets.push_back( std::make_pair(offset,-1) );
    ++retVal;
    /// book also the corresponding Blob reader
    pool::IBlobReader* blobReader = 0;
    std::map< int,pool::IBlobReader* >::const_iterator iBlobReader=m_blobReaders.find(offset);
    if( iBlobReader!=m_blobReaders.end() )
    {
      blobReader = iBlobReader->second;
    }
    if(!blobReader)
    {
      pool::IBlobStreamingService* blobStreamingService = m_context->blobStreamer();
      if ( !blobStreamingService ) {
        throw pool::RelationalStorageService::RelationalStorageServiceException( "Cannot read blob variable \""+mapping.variableName()+"\": No blob streaming Service installed.","ObjectRelationalReadDataBuffer::buildBuffer" );
      }
      // fix me: version is currently not used...
      blobReader = blobStreamingService->newReader(TypeH::ByName(mapping.variableType()),"");
      m_blobReaders.insert(std::make_pair(offset,blobReader));
    }
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::Object ||
            mapping.elementType() == pool::ObjectRelationalMappingElement::DependentObject ||
            mapping.elementType() == pool::ObjectRelationalMappingElement::PoolReference ) { // Objects
    // Get the id columns first
    std::vector< std::string > columnNames = mapping.columnNames();
    if ( mapping.elementType() == pool::ObjectRelationalMappingElement::PoolReference ) {
      columnNames.pop_back();
      columnNames.pop_back();
    }
    unsigned int iCol=0;
    for ( std::vector< std::string >::const_iterator iColumn = columnNames.begin();
          iColumn != columnNames.end(); ++iColumn, ++iCol ) {
      if(m_spec->index( *iColumn )==-1){
        const std::type_info* colType = &typeid(int);
        m_spec->extend( *iColumn, *colType );
        m_whereData->extend( *iColumn, *colType  );
      }
    }

    if ( !ObjectRelationalClassUtils::isTypeObject( dictionary ) ) {
      // processing of the non-objects (top level: primitives, containers, arrays, ptr's, etc)
      ObjectRelationalMappingElement::const_iterator iArrElement = mapping.find( dictionary.Name(Reflex::SCOPED) );
      if(iArrElement==mapping.end()) throw pool::RelationalStorageService::RelationalStorageServiceException( "Mapping for array \"" + dictionary.Name(Reflex::SCOPED) +
                                                                                                              "\" has not been found.",
                                                                                                              "ObjectRelationalReadDataBuffer::buildBuffer" );
      retVal += this->buildBuffer( iArrElement->second, dictionary, schema, offset, keyData );
    } else {
      if ( mapping.elementType() == pool::ObjectRelationalMappingElement::PoolReference ) {

        // Book the OID columns
        size_t index = columnNames.size();
        int stubOffset = -( 1 + offset );
        m_spec->extend< int >( mapping.columnNames()[index] );
        m_offsets.push_back( std::make_pair(stubOffset,-1) );
        ++index;
        m_spec->extend< int >( mapping.columnNames()[index] );
        m_offsets.push_back( std::make_pair(stubOffset,-1) );
      
      } else {
      
        // update dictionary to include base classes members
        dictionary.UpdateMembers(); 
        // Loop over the fields of the class.
        for(unsigned int i=0;i<dictionary.DataMemberSize();i++){
          Reflex::Member field = dictionary.DataMemberAt(i);
      
          // Ignore the transients
          if ( field.IsTransient() ) continue;

          // Get the member type and resolve possible typedef chains
          TypeH field_type = ObjectRelationalClassUtils::resolvedType( field.TypeOf() );
          if ( ! field_type ) {
            throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for field \"" +
                                                                                     field.Name() + "\" of class \"" +
                                                                                     dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                                     "ObjectRelationalReadDataBuffer::buildBuffer" );
          }

          // check if the member is from a class in the inheritance tree
          std::string scope("");
          TypeH declaringType = ObjectRelationalClassUtils::resolvedType( field.DeclaringType());
          if( declaringType != dictionary ){
            scope = declaringType.Name(Reflex::SCOPED|Reflex::FINAL);
          }
          // Get the field name
          std::string fieldName = pool::ObjectRelationalDefaultMappingRules::scopedVariableName( field.Name(), scope );

          // Retrieve the relevant mapping element
          pool::ObjectRelationalMappingElement::const_iterator iFieldElement = mapping.find( fieldName );
          if ( iFieldElement == mapping.end() ) {
            throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + fieldName + "\" not found in the mapping element",
                                                                                     "ObjectRelationalReadDataBuffer::buildBuffer" );
          }
          const pool::ObjectRelationalMappingElement& fieldElement = iFieldElement->second;
  
          retVal += this->buildBuffer( fieldElement, field_type, schema, offset + field.Offset(), keyData );
        }
        
      }
    }
    
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::InlineCArray ) { // Inline c-arrays
    // Get the id columns first
    std::vector< std::string > columnNames = mapping.columnNames();
    for ( std::vector< std::string >::const_iterator iColumn = columnNames.begin();
          iColumn != columnNames.end(); ++iColumn ) {
      if(m_spec->index( *iColumn )==-1){
        m_spec->extend< int >( *iColumn );
      }
    }
    
      // Get the member type and resolve possible typedef chains
    TypeH elementType = ObjectRelationalClassUtils::resolvedType( dictionary.ToType() );
    if ( ! elementType ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException(
        "Missing dictionary information for the element of array \"" +dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"","ObjectRelationalReadDataBuffer::buildBuffer" );
    }
    // Loop over the elements of the array.
    for ( unsigned int i=0;i<dictionary.ArrayLength();i++){

      // Form the element name
      std::string arrayElementLabel =  ObjectRelationalDefaultMappingRules::variableNameForArrayIndex( mapping.variableName(),i);

      // Retrieve the relevant mapping element
      pool::ObjectRelationalMappingElement::const_iterator iElement = mapping.find( arrayElementLabel );
      if ( iElement == mapping.end() ) {
        throw pool::RelationalStorageService::RelationalStorageServiceException( "Array element \"" + arrayElementLabel + "\" not found in the mapping element",
                                                                                 "ObjectRelationalReadDataBuffer::buildBuffer" );
      }
      const pool::ObjectRelationalMappingElement& element = iElement->second;

      retVal += this->buildBuffer( element, elementType, schema, offset + i*elementType.SizeOf(), keyData );
    }
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::Array ) { // Containers
    if ( !ObjectRelationalClassUtils::isTypeContainer(dictionary) ){ 
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + mapping.variableName() +
                                                                               " of type "+ mapping.variableType() +
                                                                               "\" has been declared as array in the mapping element",
                                                                               "ObjectRelationalReadDataBuffer::buildBuffer" );
    }
    this->buildArrayBuffer( mapping, dictionary, schema, offset, keyData );
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::PoolArray ) { // Pool Arrays
    if ( !ObjectRelationalClassUtils::isTypePVector(dictionary) ){ 
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + mapping.variableName() +
                                                                               " of type "+ mapping.variableType() +
                                                                               "\" has been declared as array in the mapping element",
                                                                               "ObjectRelationalReadDataBuffer::buildBuffer" );
    }
    this->buildArrayBuffer( mapping, dictionary, schema, offset, keyData );
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::CArray ) { // C-arrays
    if ( ! dictionary.IsArray() ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + mapping.variableName() +
                                                                               " of type "+ mapping.variableType() +
                                                                               "\" has been declared as a C-array in the mapping element",
                                                                               "ObjectRelationalReadDataBuffer::buildBuffer" );
    }
    this->buildCArrayBuffer( mapping, dictionary, schema, offset, keyData );
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::PoolPointer ) { // Pool pointer
    if ( ! ObjectRelationalClassUtils::isTypePoolPointer( dictionary ) ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + mapping.variableName() +
                                                                               " of type "+ mapping.variableType() +
                                                                               "\" has been declared as Pool Pointer in the mapping element",
                                                                               "ObjectRelationalReadDataBuffer::buildBuffer" );
    }
    this->buildPoolPointerBuffer( mapping, dictionary, schema, offset, keyData );
  }
  else if ( mapping.elementType() == pool::ObjectRelationalMappingElement::PoolPolymorphicPointer ) { // Pool Poly Pointers
    if ( ! ObjectRelationalClassUtils::isTypePoolPolymorphicPointer( dictionary )) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + mapping.variableName() +
                                                                               " of type "+ mapping.variableType() +
                                                                               "\" has been declared as a Pool PolymorphicPointer in the mapping element",
                                                                               "ObjectRelationalReadDataBuffer::buildBuffer" );
    }
    int metadataOffset = m_metadata.size();
    if(m_spec->index( mapping.columnNames()[0] )==-1){
      m_spec->extend( mapping.columnNames()[0], typeid(std::string) );
    }
    if(m_spec->index( mapping.columnNames()[1] )==-1){
      m_spec->extend( mapping.columnNames()[1], typeid(int) );
    }
    m_offsets.push_back( std::make_pair(offset,metadataOffset) );
    m_metadata.push_back( std::make_pair(std::string(""),0) ); // reserving a metadata slot
    this->buildPoolPolymorphicPointerBuffer( *m_mapping, dictionary, offset, metadataOffset );
  }
  else { // Do not treat other types yet...
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Cannot treat anything else than objects, primitives, arrays and containers for the moment...",
                                                                             "ObjectRelationalReadDataBuffer::buildBuffer" );
  }

  return retVal;
}




bool
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::buildArrayBuffer( const pool::ObjectRelationalMappingElement& mapping,
                                                                                  const TypeH& dictionary,
                                                                                  coral::ISchema& schema,
                                                                                  int offset,
                                                                                  bool keyData )
{
  pool::RelationalStorageService::ObjectRelationalReadDataBuffer* newBuffer = new pool::RelationalStorageService::ObjectRelationalReadDataBuffer(m_context);
  m_arrayBuffers.push_back( newBuffer );

  newBuffer->m_mapping = &mapping;
  newBuffer->m_dictionary = dictionary;
  newBuffer->m_metadata = m_metadata;
  newBuffer->m_parentMetadata = m_parentMetadata;
  newBuffer->m_parentMetadataOffset = m_parentMetadataOffset;
  newBuffer->m_offsetToParent = offset;
  newBuffer->m_table = &(schema.tableHandle( mapping.tableName() ) );
  newBuffer->m_containerHandler = pool::RelationalStorageService::ContainerHandlerFactory::newContainerHandler( dictionary );

  newBuffer->m_keyData = keyData;
  newBuffer->m_spec = new coral::AttributeListSpecification;
  newBuffer->m_whereData = new coral::AttributeList;

  // Get the id columns first
  const std::vector< std::string >& columnNames = mapping.columnNames();
  std::vector< std::string >::const_iterator iColumn = columnNames.begin();
  if( mapping.isDependentTree() ){
    newBuffer->m_spec->extend< int >( *iColumn );
    newBuffer->m_whereData->extend< int >( *iColumn );
    newBuffer->m_whereClause += *iColumn + " = :" + *iColumn;
    iColumn++;
  }
  bool posCol = true;
  for ( ;iColumn != columnNames.end(); ++iColumn ) {
    newBuffer->m_spec->extend< int >( *iColumn );
    if( !posCol ){
      newBuffer->m_whereData->extend< int >( *iColumn );
      if ( ! ( newBuffer->m_whereClause.empty() ) ) newBuffer->m_whereClause += " AND ";
      newBuffer->m_whereClause += *iColumn + " = :" + *iColumn;
    }
    posCol = false;
  }

  TypeH field_type;
  if( ObjectRelationalClassUtils::isTypeAssociativeContainer(dictionary)) { // associative containers.    

    TypeH key_type = ObjectRelationalClassUtils::containerKeyType(dictionary);
    TypeH keyResolvedType = ObjectRelationalClassUtils::resolvedType(key_type);

    if ( ! key_type ||!keyResolvedType ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for the key type of the container \"" +
                                                                               dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                               "ObjectRelationalReadDataBuffer::buildArrayBuffer" );
    }

    std::string keyName = key_type.Name();
    // Retrieve the relevant mapping element
    pool::ObjectRelationalMappingElement::const_iterator iFieldElement = mapping.find( keyName );
    if ( iFieldElement == mapping.end() ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + keyName + "\" not found in the mapping element",
                                                                               "ObjectRelationalReadDataBuffer::buildArrayBuffer" );
    }
    newBuffer->m_keyItems = newBuffer->buildBuffer( iFieldElement->second, keyResolvedType, schema, 0, true );
    field_type = ObjectRelationalClassUtils::containerDataType(dictionary);
  } else {
    field_type = ObjectRelationalClassUtils::containerValueType(dictionary);
  }

  TypeH fieldResolvedType = ObjectRelationalClassUtils::resolvedType(field_type);
  // Check the component type
  if ( ! field_type ||!fieldResolvedType ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for the content type of the container \"" +
                                                                             dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                             "ObjectRelationalReadDataBuffer::buildArrayBuffer" );
  }
  
  std::string fieldName = field_type.Name();
  // Retrieve the relevant mapping element
  pool::ObjectRelationalMappingElement::const_iterator iFieldElement = mapping.find( fieldName );
  if ( iFieldElement == mapping.end() ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + fieldName + "\" not found in the mapping element",
                                                                             "ObjectRelationalReadDataBuffer::buildArrayBuffer" );
  }
  newBuffer->buildBuffer( iFieldElement->second, fieldResolvedType, schema, 0, false );
  
  return true;
}

bool
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::buildCArrayBuffer( const pool::ObjectRelationalMappingElement& mapping,
                                                                                   const TypeH& dictionary,
                                                                                   coral::ISchema& schema,
                                                                                   int offset,
                                                                                   bool keyData )
{
  pool::RelationalStorageService::ObjectRelationalReadDataBuffer* newBuffer = new pool::RelationalStorageService::ObjectRelationalReadDataBuffer(m_context);
  m_arrayBuffers.push_back( newBuffer );

  newBuffer->m_mapping = &mapping;
  newBuffer->m_dictionary = dictionary;
  newBuffer->m_metadata = m_metadata;
  newBuffer->m_parentMetadata = m_parentMetadata;
  newBuffer->m_parentMetadataOffset = m_parentMetadataOffset;
  newBuffer->m_offsetToParent = offset;
  newBuffer->m_table = &(schema.tableHandle( mapping.tableName() ) );
  newBuffer->m_containerHandler = pool::RelationalStorageService::ContainerHandlerFactory::newContainerHandler( dictionary );

  newBuffer->m_keyData = keyData;
  newBuffer->m_spec = new coral::AttributeListSpecification;
  newBuffer->m_whereData = new coral::AttributeList;

  // Get the id columns first
  const std::vector< std::string >& columnNames = mapping.columnNames();
  std::vector< std::string >::const_iterator iColumn = columnNames.begin();
  if( mapping.isDependentTree() ){
    newBuffer->m_spec->extend< int >( *iColumn );
    newBuffer->m_whereData->extend< int >( *iColumn );
    newBuffer->m_whereClause += *iColumn + " = :" + *iColumn;
    iColumn++;
  }
  bool posCol = true;
  for ( ;iColumn != columnNames.end(); ++iColumn ) {
    newBuffer->m_spec->extend< int >( *iColumn );
    if( !posCol ){
      newBuffer->m_whereData->extend< int >( *iColumn );
      if ( ! ( newBuffer->m_whereClause.empty() ) ) newBuffer->m_whereClause += " AND ";
      newBuffer->m_whereClause += *iColumn + " = :" + *iColumn;
    }
    posCol = false;
  }

  // Check the array type
  TypeH arrayType = dictionary.ToType();
  TypeH arrayResolvedType = ObjectRelationalClassUtils::resolvedType(arrayType);
  // Check the component type
  if ( ! arrayType || !arrayResolvedType ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for the element type of the array \"" +
                                                                             dictionary.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                             "ObjectRelationalReadDataBuffer::buildCArrayBuffer" );
  }

  std::string arrayTypeName = arrayType.Name();

  // Retrieve the relevant mapping element
  pool::ObjectRelationalMappingElement::const_iterator iElement = mapping.find( arrayTypeName );
  if ( iElement == mapping.end() ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Field \"" + arrayTypeName + "\" not found in the mapping element",
                                                                             "ObjectRelationalReadDataBuffer::buildCArrayBuffer" );
  }
  newBuffer->buildBuffer( iElement->second, arrayResolvedType, schema, 0, false );
  
  return true;
}

bool
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::buildPoolPointerBuffer( const pool::ObjectRelationalMappingElement& mapping,
                                                                                        const ROOT::Reflex::Type& dictionary,
                                                                                        coral::ISchema& schema,
                                                                                        int offset,
                                                                                        bool keyData )
{
  pool::RelationalStorageService::ObjectRelationalReadDataBuffer* newBuffer = new pool::RelationalStorageService::ObjectRelationalReadDataBuffer(m_context);
  m_poolPointerBuffers.push_back( newBuffer );

  newBuffer->m_mapping = &mapping;
  newBuffer->m_dictionary = dictionary;
  newBuffer->m_metadata = m_metadata;
  newBuffer->m_parentMetadata = m_parentMetadata;
  newBuffer->m_parentMetadataOffset = m_parentMetadataOffset;
  newBuffer->m_offsetToParent = offset;
  newBuffer->m_table = &(schema.tableHandle( mapping.tableName() ) );
  newBuffer->m_keyData = keyData;
  newBuffer->m_spec = new coral::AttributeListSpecification;
  newBuffer->m_whereData = new coral::AttributeList;

  // Get the id columns first
  const std::vector< std::string >& columnNames = mapping.columnNames();
  std::vector< std::string >::const_iterator iColumn = columnNames.begin();
  if( mapping.isDependentTree() ){
    newBuffer->m_spec->extend< int >( *iColumn );
    newBuffer->m_whereData->extend< int >( *iColumn );
    newBuffer->m_whereClause += *iColumn + " = :" + *iColumn;
    iColumn++;
  }
  for ( ;iColumn != columnNames.end(); ++iColumn ) {
    newBuffer->m_spec->extend< int >( *iColumn );
    newBuffer->m_whereData->extend< int >( *iColumn );
    if ( ! ( newBuffer->m_whereClause.empty() ) ) newBuffer->m_whereClause += " AND ";
    newBuffer->m_whereClause += *iColumn + " = :" + *iColumn;
  }

  // Check the  type
  ROOT::Reflex::Type ptrType = dictionary.TemplateArgumentAt(0);
  ROOT::Reflex::Type ptrResolvedType = ObjectRelationalClassUtils::resolvedType(ptrType);
  // Check the component type
  if ( ! ptrType || !ptrResolvedType ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for the type of the pointer \"" +
                                                                             dictionary.Name(ROOT::Reflex::SCOPED|ROOT::Reflex::FINAL) + "\"",
                                                                             "ObjectRelationalReadDataBuffer::buildPoolPointerBuffer" );
  }

  std::string ptrTypeName = ptrType.Name();
  // Retrieve the relevant mapping element
  pool::ObjectRelationalMappingElement::const_iterator iElement = mapping.find( ptrTypeName );
  if ( iElement == mapping.end() ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Item \"" + ptrTypeName + "\" not found in the mapping element",
                                                                             "ObjectRelationalReadDataBuffer::buildPoolPointerBuffer" );
  }
  newBuffer->buildBuffer( iElement->second, ptrResolvedType, schema, 0, false );  
  return true;
}

bool
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::buildPoolPolymorphicPointerBuffer( const pool::ObjectRelationalMappingElement& topLevelMapping,
                                                                                                   const ROOT::Reflex::Type& dictionary,
                                                                                                   int offset,
                                                                                                   int metadataOffset)
{
  pool::RelationalStorageService::ObjectRelationalReadDataBuffer* newBuffer = new pool::RelationalStorageService::ObjectRelationalReadDataBuffer(m_context);
  m_poolPolymorphicPointerBuffers.push_back( newBuffer );

  newBuffer->m_mapping = &topLevelMapping;
  newBuffer->m_dictionary = dictionary;
  newBuffer->m_offsetToParent = offset;
  newBuffer->m_parentMetadata = &m_metadata;
  newBuffer->m_parentMetadataOffset = metadataOffset;

  // Check the  type
  ROOT::Reflex::Type ptrType = dictionary.TemplateArgumentAt(0);
  ROOT::Reflex::Type ptrResolvedType = ObjectRelationalClassUtils::resolvedType(ptrType);
  // Check the component type
  if ( ! ptrType || !ptrResolvedType ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing dictionary information for the type of the pointer \"" +
                                                                             dictionary.Name(ROOT::Reflex::SCOPED|ROOT::Reflex::FINAL) + "\"",
                                                                             "ObjectRelationalReadDataBuffer::buildPoolPointerBuffer" );
  }
  return true;
}

const coral::AttributeListSpecification&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::dataSpecification() const
{
  return *m_spec;
}

const std::vector< std::pair<int,int> >&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::offsets() const
{
  return m_offsets;
}

const std::vector< std::pair<std::string,int> >&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::metadata() const {
  return m_metadata;
}

std::vector< std::pair<std::string,int> >&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::metadata() {
  return m_metadata;
}

const std::vector< std::pair<std::string,int> >*
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::parentMetadata() const{
  return m_parentMetadata;
}

int
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::parentMetadataOffset() const{
  return m_parentMetadataOffset;
}

int
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::keyItems() const
{
  return m_keyItems;
}

int
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::offsetToParent() const
{
  return m_offsetToParent;
}

coral::ITable&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::table()
{
  return *m_table;
}

TypeH
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::dictionary() const
{
  return m_dictionary;
}


const pool::ObjectRelationalMappingElement&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::mapping() const
{
  return *m_mapping;
}


pool::RelationalStorageService::IContainerHandler&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::containerHandler()
{
  return *m_containerHandler;
}


bool
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::isKeyData() const
{
  return m_keyData;
}


pool::RelationalStorageService::ObjectRelationalReadDataBuffer::iterator
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::begin()
{
  return m_arrayBuffers.begin();
}

pool::RelationalStorageService::ObjectRelationalReadDataBuffer::iterator
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::end()
{
  return m_arrayBuffers.end();
}

const std::vector<pool::RelationalStorageService::ObjectRelationalReadDataBuffer* >&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::poolPointerBuffers() const{
  return m_poolPointerBuffers;
}

const std::vector< pool::RelationalStorageService::ObjectRelationalReadDataBuffer* >&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::poolPolymorphicPointerBuffers() const {
  return m_poolPolymorphicPointerBuffers;
}

const std::string&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::whereClauseForId() const
{
  return m_whereClause;
}


coral::AttributeList&
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::whereDataForId()
{
  return *m_whereData;
}

void
pool::RelationalStorageService::ObjectRelationalReadDataBuffer::readBlob(int offset, const coral::Blob& data, void* startAddress)
{
  pool::IBlobReader* blobReader = 0;
  std::map< int,pool::IBlobReader* >::const_iterator iBlobReader=m_blobReaders.find(offset);
  if( iBlobReader!=m_blobReaders.end() )
  {
    blobReader = iBlobReader->second;
  }
  if(!blobReader)
  {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Cannot read from blob: The reader has not been booked.","ObjectRelationalReadDataBuffer::writeBlob" );

  }
  blobReader->read(data,startAddress);
}
