#ifndef OBJECTRELATIONALREADDATABUFFER_H
#define OBJECTRELATIONALREADDATABUFFER_H

#include <vector>
#include <map>
#include <string>
#include "Reflex/Type.h"

#include "RVersion.h"

namespace coral {

  class ITable;
  class ISchema;
  class AttributeList;
  class AttributeListSpecification;
  class Blob;
}


#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
 typedef ROOT::Reflex::Type TypeH;
#else
 typedef Reflex::Type TypeH;
#endif
 

namespace pool {
  class ObjectRelationalMappingElement;
  class IBlobReader;
  class Context;

  namespace RelationalStorageService {

    class IContainerHandler;

    class ObjectRelationalReadDataBuffer {
    public:
      /// Constuctor
      explicit ObjectRelationalReadDataBuffer(pool::Context* context);

      /// Destructor
      ~ObjectRelationalReadDataBuffer();

      /// Builds the structure given a mapping element and the dictionary information
      bool build( coral::ISchema& schema,
                  const pool::ObjectRelationalMappingElement& mapping,
                  const TypeH& dictionary );

      /// adds a new buffer to the dependency list
      ObjectRelationalReadDataBuffer& addDependency();

      /// Returns the data specification
      const coral::AttributeListSpecification& dataSpecification() const;

      /// Returns the offset vector
      const std::vector<std::pair<int, int> >& offsets() const;

      /// Returns the metadata vector
      const std::vector< std::pair<std::string,int> >& metadata() const;
      /// Returns the metadata vector
      std::vector< std::pair<std::string,int> >& metadata();

      /// Returns the parent metadata
      const std::vector< std::pair<std::string,int> >* parentMetadata() const;
      
      /// Returns the parent metadata offset
      int parentMetadataOffset() const;
      
      /// Returns the number of items belonging to the key (in case of an associative container)
      int keyItems() const;

      /// Returns the offset w.r.t. the parent object
      int offsetToParent() const;

      /// Returns the associoted table handle
      coral::ITable& table();

      /// Returns the associated dictionary information
      TypeH dictionary() const;

      /// Returns the relevant mapping element
      const pool::ObjectRelationalMappingElement& mapping() const;

      /// Returns the container handler
      IContainerHandler& containerHandler();

      /// Returns a flag indicating whether the starting address is the key of an associative container
      bool isKeyData() const;

      /// Returns the where clause for pinning the id
      const std::string& whereClauseForId() const;

      /// Returns the data for the where clause.
      coral::AttributeList& whereDataForId();

      /// Returns the original version of the specified data, using the blob reader registered in the pool context 
      void readBlob(int offset, const coral::Blob& data, void* startAddress);

      public:
      typedef std::vector< ObjectRelationalReadDataBuffer* >::iterator iterator;
      iterator begin();
      iterator end();
      const std::vector<ObjectRelationalReadDataBuffer* >& poolPointerBuffers() const;
      const std::vector< ObjectRelationalReadDataBuffer* >& poolPolymorphicPointerBuffers() const;

    private:
      /// builds a buffer entry
      int buildBuffer( const pool::ObjectRelationalMappingElement& mapping,
                       const TypeH& dictionary,
                       coral::ISchema& schema,
                       int offset,
                       bool keyData );

      /// builds an array buffer
      bool buildArrayBuffer( const pool::ObjectRelationalMappingElement& mapping,
                             const TypeH& dictionary,
                             coral::ISchema& schema,
                             int offset,
                             bool keyData );

      /// builds an array buffer
      bool buildCArrayBuffer( const pool::ObjectRelationalMappingElement& mapping,
                              const TypeH& dictionary,
                              coral::ISchema& schema,
                              int offset,
                              bool keyData );

      /// builds a pool ptr buffer
      bool buildPoolPointerBuffer( const pool::ObjectRelationalMappingElement& mapping,
                                   const ROOT::Reflex::Type& dictionary,
                                   coral::ISchema& schema,
                                   int offset,
                                   bool keyData );
      
      /// builds a Pool PolyPtr buffer
      bool buildPoolPolymorphicPointerBuffer( const pool::ObjectRelationalMappingElement& topLevelMapping,
                                              const ROOT::Reflex::Type& dictionary,
                                              int offset,
                                              int metadataOffset);
      private:
      pool::Context* m_context;
      coral::AttributeListSpecification* m_spec;
      std::vector< std::pair<int,int> > m_offsets;
      std::vector< std::pair<std::string,int> > m_metadata;
      std::vector< std::pair<std::string,int> >*  m_parentMetadata;
      int m_parentMetadataOffset;
      coral::ITable* m_table;
      TypeH m_dictionary;
      int m_offsetToParent;
      int m_keyItems;
      const pool::ObjectRelationalMappingElement* m_mapping;
      bool m_keyData;
      IContainerHandler* m_containerHandler;
      std::vector< ObjectRelationalReadDataBuffer* > m_arrayBuffers;
      coral::AttributeList* m_whereData;
      std::string m_whereClause;
      std::map< int, pool::IBlobReader* > m_blobReaders;
      std::vector< ObjectRelationalReadDataBuffer* > m_poolPointerBuffers;
      std::vector< ObjectRelationalReadDataBuffer* > m_poolPolymorphicPointerBuffers;
      std::vector< ObjectRelationalReadDataBuffer* > m_dependentObjectBuffers;
    };

  }

}

#endif
