#include "ObjectWriter.h"
#include "ObjectRelationalDataBuffer.h"
#include "MappingManager.h"
#include "MappingSequence.h"
#include "IContainerHandler.h"
#include "IContainerIteratorHandler.h"
#include "RelationalStorageServiceNames.h"
#include "RelationalStorageServiceException.h"
#include "ObjectRelationalAccess/ObjectRelationalMapping.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingElement.h"
#include "ObjectRelationalAccess/ObjectRelationalClassUtils.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/ITableDataEditor.h"
#include "RelationalAccess/IBulkOperation.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/AttributeSpecification.h"
#include "CoralBase/Blob.h"
#include "StorageSvc/DataCallBack.h"
#include "StorageSvc/DbDataHandlerGuard.h"
#include "POOLCore/Reference.h"
#include "POOLCore/Token.h"
#include "Reflex/Member.h"
#include <sstream>

pool::RelationalStorageService::ObjectWriter::ObjectWriter( coral::ISchema& schema,
                                                            const Reflex::Type& dictionaryEntry,
                                                            MappingManager& mappingManager,
                                                            MappingSequence& mappingSequence,
                                                            pool::DbDatabase* databaseHandle ):
  DbIOHandler(databaseHandle),
  m_schema( schema ),
  m_dataBuffer( 0 ),
  m_refDict( Reflex::Type::ByTypeInfo( typeid(pool::Reference) ) ),
  m_mappingManager(mappingManager),
  m_mappingSequence(mappingSequence)
{
  m_dataBuffer = new pool::RelationalStorageService::ObjectRelationalDataBuffer(DbIOHandler::m_context);
  const ObjectRelationalMappingElement& mappingElement = m_mappingManager.mappingForClass( dictionaryEntry ).element();
  m_dataBuffer->build( schema, mappingElement, dictionaryEntry );

}


pool::RelationalStorageService::ObjectWriter::~ObjectWriter()
{
  delete m_dataBuffer;
}

bool
pool::RelationalStorageService::ObjectWriter::write( int oid,
                                                     const void* data,
                                                     pool::DataCallBack& callback )
{
  return write(oid,*m_dataBuffer,data, callback);
}

bool
pool::RelationalStorageService::ObjectWriter::write( int oid,
                                                     ObjectRelationalDataBuffer& dataBuffer,
                                                     const void* data,
                                                     pool::DataCallBack& callback )
{
  coral::AttributeList& buffer = dataBuffer.data();

  // Write first the row of the top level table
  coral::AttributeList::iterator iAttribute = buffer.begin();

  // Skip the identity columns
  for ( size_t iColumn = 0; iColumn < dataBuffer.mapping().columnNames().size(); ++iColumn ) ++iAttribute;

  // Copy the data
  std::vector<std::pair<int,int> >::const_iterator iOffset = dataBuffer.offsets().begin();
  while ( iAttribute != buffer.end() ) {
    int offset = iOffset->first;
    if ( offset < 0 ) {
      pool::Token::OID_t oid;
      const void* refAddress = static_cast< const char* >( data ) - ( 1 + offset );
      // Call here the ref handler to get the oid...
      pool::DbDataHandlerGuard handlerGuard( &callback, m_databaseContext );
      callback.notify( pool::DataCallBack::PUT, pool::DataCallBack::REFERENCE, m_refDict, const_cast< void* >( refAddress ), &oid );
      iAttribute->data<int>() =  oid.first;
      ++iAttribute;
      ++iOffset;
      iAttribute->data<int>() =  oid.second;
    }
    else {
      if( iOffset->second < 0 ) {
        if( iAttribute->specification().type() == typeid(coral::Blob) ){
          iAttribute->bind<coral::Blob>( dataBuffer.writeBlob( offset, static_cast< const char* >( data ) + offset ) );
        } else {
          iAttribute->setValueFromAddress( static_cast< const char* >( data ) + offset );
          if(!checkData(*iAttribute)){
            std::stringstream msg;
            msg << "Data provided for column \""<<iAttribute->specification().name()<<"\" is not valid for RDBMS storage.";
            throw pool::RelationalStorageService::RelationalStorageServiceException(msg.str(),"ObjectWriter::write");
          }
        }
      } else {
        std::pair<std::string,std::string> ptrMetadata = getDataForPolyPtr( dataBuffer.metadata()[ iOffset->second ].first, static_cast< const char* >( data ) + offset);
        iAttribute->data<std::string>() = ptrMetadata.first;
        ++iAttribute;
        int rid = 0;
        if( !ptrMetadata.second.empty()){
          rid = m_mappingSequence.incrementOid( ptrMetadata.second );
          dataBuffer.metadata()[ iOffset->second ].second = rid;
        }
        iAttribute->data<int>() = rid;
      }
    }
    ++iAttribute;
    ++iOffset;
  }

  // Set the oid. For the moment set only the first identity column. In principle there should be a plugin taking
  // care of the OID -> identity values mapping.
  coral::AttributeList::iterator iIdAttr = buffer.begin();
  if(dataBuffer.mapping().isDependentTree()){
    iIdAttr->data< int >() = (dataBuffer.parentMetadata()->operator[](dataBuffer.parentMetadataOffset())).second;
    ++iIdAttr;
  }
  iIdAttr->data< int >() =  oid;
  
  // Write the entry into the table
  dataBuffer.table().dataEditor().insertRow( buffer );

  std::vector<int> identity( 1, oid );

  // Loop over the arrays now...
  for ( pool::RelationalStorageService::ObjectRelationalDataBuffer::iterator iBuffer = dataBuffer.begin();
        iBuffer != dataBuffer.end(); ++iBuffer ) {
    if ( ! this->writeArray( identity, **iBuffer,
                             static_cast< const char* >( data ) + (*iBuffer)->offsetToParent(),
                             callback ) ) return false;
  }
  // Loop over the pool pointers now...
  for ( std::vector<ObjectRelationalDataBuffer*>::const_iterator iBuffer = dataBuffer.poolPointerBuffers().begin();
        iBuffer != dataBuffer.poolPointerBuffers().end(); ++iBuffer ) {
    if ( ! this->writePoolPointer( oid, *(*iBuffer),
                                   static_cast< const char* >( data ) + (*iBuffer)->offsetToParent(),
                                   callback ) ) return false;
  }
  // Loop over the pool polymorphic pointers now...
  for ( std::vector<ObjectRelationalDataBuffer*>::const_iterator iBuffer = dataBuffer.poolPolymorphicPointerBuffers().begin();
        iBuffer != dataBuffer.poolPolymorphicPointerBuffers().end(); ++iBuffer ) {
    if ( ! this->writePoolPolymorphicPointer( oid, *(*iBuffer),
                                              static_cast< const char* >( data ) + (*iBuffer)->offsetToParent(),
                                              callback ) ) return false;
  }
  return true;
}

bool
pool::RelationalStorageService::ObjectWriter::writeArray( const std::vector<int>& identity,
                                                          pool::RelationalStorageService::ObjectRelationalDataBuffer& buffer,
                                                          const void* address,
                                                          pool::DataCallBack& callback )
{
  const Reflex::Type& dictionary = buffer.dictionary();

  std::vector< int > id( identity.size() + 1, 0 );
  for ( size_t i = 0; i < identity.size(); ++i ) id[i+1] = identity[i];
  
  // Retrieve the container type
  Reflex::Type keyType;
  if ( buffer.keyItems() > 0 ) keyType = dictionary.TemplateArgumentAt(0);

  // Use the iterator to loop over the elements of the container.
  pool::RelationalStorageService::IContainerHandler& containerHandler = buffer.containerHandler();

  size_t containerSize = containerHandler.size( address );
  size_t persistentSize = containerHandler.persistentSize( address );
  if ( containerSize == 0 || containerSize < persistentSize ) return true;
  size_t startElementIndex = containerHandler.startElementIndex( address );

  pool::RelationalStorageService::IContainerIteratorHandler* iteratorHandler = containerHandler.iterate( address );
  const Reflex::Type& iteratorDereferenceReturnType = iteratorHandler->returnType();

  const std::vector< std::pair<int,int> >& offsets = buffer.offsets();
  coral::AttributeList& data = buffer.data();
  int keyItems = buffer.keyItems();
  bool useBulkInsert = ( buffer.size() == 0 ) ? true : false;
  coral::IBulkOperation* bulkInserter = 0;
  if ( useBulkInsert ) {
    bulkInserter = buffer.table().dataEditor().bulkInsert(data,containerSize-startElementIndex);
    if ( ! bulkInserter ) return false;
  }

  for ( size_t iObject = startElementIndex; iObject < containerSize; ++iObject ) {
    id[0] = iObject;
    void* objectReference = iteratorHandler->object();
    void* componentData = objectReference;
    void* keyData = 0;

    // Treat the array element.
    coral::AttributeList::iterator iAttribute = data.begin();

    
    // Skip temporarily the refid column (for dependent objects)
    if( buffer.mapping().isDependentTree() ){
      ++iAttribute;
    }    
    // Skip temporarily the id columns
    for ( std::vector<int>::const_iterator iId = id.begin(); iId != id.end(); ++iId ) {
      ++iAttribute;
    }
    
    // Copy the data
    std::vector<std::pair<int,int> >::const_iterator iOffset = offsets.begin();
    if ( keyType ) { // treat the key object first
      Reflex::Member firstField = iteratorDereferenceReturnType.MemberByName( "first" );
      if ( ! firstField ) {
        throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not retrieve the \"first\" field for the class \"" +
                                                                                 iteratorDereferenceReturnType.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                                 "ObjectWriter::writeArray" );
      }
      keyData = static_cast< char* >( objectReference ) + firstField.Offset();
      for ( int i = 0; i < keyItems; ++i ) {
        int offset = iOffset->first;
        if ( offset < 0 ) {
          pool::Token::OID_t oid;
          const void* refAddress = static_cast< const char* >( keyData ) - ( 1 + offset );
          // Call here the ref handler to get the oid...
          callback.notify( pool::DataCallBack::PUT, pool::DataCallBack::REFERENCE, m_refDict, const_cast< void* >( refAddress ), &oid );
          iAttribute->data<int>() =  oid.first;
          ++iAttribute;
          ++iOffset;
          iAttribute->data<int>() = oid.second;
        }
        else {
          if( iOffset->second < 0 ) {
            if( iAttribute->specification().type() == typeid(coral::Blob) ){
              iAttribute->bind<coral::Blob>( buffer.writeBlob( iOffset->first, static_cast< const char* >( keyData ) + iOffset->first ) );
            } else {
              iAttribute->setValueFromAddress( static_cast< const char* >( keyData ) + iOffset->first );
              if(!checkData(*iAttribute)){
                std::stringstream msg;
                msg << "Data provided for column \""<<iAttribute->specification().name()<<"\" is not valid for RDBMS storage.";
                throw pool::RelationalStorageService::RelationalStorageServiceException(msg.str(),"ObjectWriter::writeArray");
              }
            }
          } else {
            std::pair<std::string,std::string> ptrMetadata = getDataForPolyPtr( buffer.metadata()[ iOffset->second ].first, static_cast< const char* >( keyData ) + iOffset->first);
            iAttribute->data<std::string>() = ptrMetadata.first;
            ++iAttribute;
            int rid = 0;
            if( !ptrMetadata.second.empty()){
              rid = m_mappingSequence.incrementOid( ptrMetadata.second );
              buffer.metadata()[ iOffset->second ].second = rid;
            }
            iAttribute->data<int>() = rid;
          }
        }
        ++iAttribute;
        ++iOffset;
      }
      Reflex::Member secondField = iteratorDereferenceReturnType.MemberByName( "second" );
      if ( ! secondField ) {
        throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not retrieve the \"second\" field for the class \"" +
                                                                                 iteratorDereferenceReturnType.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                                 "ObjectWriter::writeArray" );
      }
      componentData = static_cast< char* >( objectReference ) + secondField.Offset();
    }
    while ( iAttribute != data.end() ) {
      int offset = iOffset->first;
      if ( offset < 0 ) {
        pool::Token::OID_t oid;
        const void* refAddress = static_cast< const char* >( componentData ) - ( 1 + offset );
        // Call here the ref handler to get the oid...
        callback.notify( pool::DataCallBack::PUT, pool::DataCallBack::REFERENCE, m_refDict, const_cast< void* >( refAddress ), &oid );
        iAttribute->data<int>() = oid.first;
        ++iAttribute;
        ++iOffset;
        iAttribute->data<int>() = oid.second;
      }
      else {
        if( iOffset->second < 0 ) {
          if( iAttribute->specification().type() == typeid(coral::Blob) ){
            iAttribute->bind<coral::Blob>( buffer.writeBlob( iOffset->first, static_cast< const char* >( componentData ) + iOffset->first ) );
          } else {
            iAttribute->setValueFromAddress( static_cast< const char* >( componentData ) + iOffset->first );
            if(!checkData(*iAttribute)){
              std::stringstream msg;
              msg << "Data provided for column \""<<iAttribute->specification().name()<<"\" is not valid for RDBMS storage.";
              throw pool::RelationalStorageService::RelationalStorageServiceException(msg.str(),"ObjectWriter::writeArray");
            }
          }
        } else {
          std::pair<std::string,std::string> ptrMetadata = getDataForPolyPtr( buffer.metadata()[ iOffset->second ].first, static_cast< const char* >( componentData ) + iOffset->first);
          iAttribute->data<std::string>() = ptrMetadata.first;
          ++iAttribute;
          int rid = 0;
          if( !ptrMetadata.second.empty()){
            rid = m_mappingSequence.incrementOid( ptrMetadata.second );
            buffer.metadata()[ iOffset->second ].second = rid;
          }
          iAttribute->data<int>() = rid;
        }
      }
      ++iAttribute;
      ++iOffset;
    }

    // Fill in the identity
    iAttribute = data.begin();
    if( buffer.mapping().isDependentTree() ){
      iAttribute->data< int >() = (buffer.parentMetadata()->operator[](buffer.parentMetadataOffset())).second;
      ++iAttribute;
    }    
    for ( std::vector<int>::const_iterator iId = id.begin(); iId != id.end(); ++iId ) {
      iAttribute->data< int >() = *iId;
      ++iAttribute;
    }

    // Write the entry into the table
    if ( useBulkInsert ) {
      bulkInserter->processNextIteration();
    }
    else {
      buffer.table().dataEditor().insertRow( data );
    }

    // Loop over the arrays now...
    for ( pool::RelationalStorageService::ObjectRelationalDataBuffer::iterator iBuffer = buffer.begin();
          iBuffer != buffer.end(); ++iBuffer ) {
      const void* theAddress = componentData;
      if ( (*iBuffer)->isKeyData() ) {
        theAddress = keyData;
      }
      if ( ! this->writeArray( id, **iBuffer,
                               static_cast< const char* >( theAddress ) + (*iBuffer)->offsetToParent(),
                               callback ) ) return false;
    }
    
    // Loop over the pool pointers now...
    for ( std::vector<ObjectRelationalDataBuffer*>::const_iterator iBuffer = buffer.poolPointerBuffers().begin();
          iBuffer != buffer.poolPointerBuffers().end(); ++iBuffer ) {
      const void* theAddress = componentData;
      if ( (*iBuffer)->isKeyData() ) {
        theAddress = keyData;
      }
      if ( ! this->writePoolPointer( id.back(), *(*iBuffer),
                                     static_cast< const char* >( theAddress ) + (*iBuffer)->offsetToParent(),
                                     callback ) ) return false;
    }
    // Loop over the pool polymorphic pointers now...
    for ( std::vector<ObjectRelationalDataBuffer*>::const_iterator iBuffer = buffer.poolPolymorphicPointerBuffers().begin();
          iBuffer != buffer.poolPolymorphicPointerBuffers().end(); ++iBuffer ) {
      const void* theAddress = componentData;
      if ( (*iBuffer)->isKeyData() ) {
        theAddress = keyData;
      }
      if ( ! this->writePoolPolymorphicPointer( id.back(), *(*iBuffer),
                                                static_cast< const char* >( theAddress ) + (*iBuffer)->offsetToParent(),
                                                callback ) ) return false;
    }
  
    // Increment the iterator
    iteratorHandler->increment();
  }
  containerHandler.finalize( const_cast<void*>(address) );

  if ( useBulkInsert ) {
    bulkInserter->flush();
  }

  if(bulkInserter) delete bulkInserter;

  delete iteratorHandler;

  return true;
}

bool
pool::RelationalStorageService::ObjectWriter::writePoolPointer( int oid,
                                                                ObjectRelationalDataBuffer& dataBuffer,
                                                                const void* data,
                                                                pool::DataCallBack& callback )
{
  ROOT::Reflex::Object ptrObj(dataBuffer.dictionary(),const_cast<void*>(data));

#if ROOT_VERSION_CODE < ROOT_VERSION(5,21,3)
  bool isNull = *static_cast<bool*>(ptrObj.Invoke("operator!").Address());
#else
  bool isNull;
  ptrObj.Invoke("operator!",isNull);
#endif

  if(!isNull){
#if ROOT_VERSION_CODE < ROOT_VERSION(5,21,3)
    void* ptrData = ptrObj.Invoke("operator*").Address();
#else
    void* ptrData;
    ptrObj.Invoke("operator*",ptrData);
#endif
    return write(oid,dataBuffer,ptrData,callback);  
  }
  return true;
}

bool
pool::RelationalStorageService::ObjectWriter::writePoolPolymorphicPointer( int oid,
                                                                           ObjectRelationalDataBuffer& dataBuffer,
                                                                           const void* data,
                                                                           pool::DataCallBack& callback )
{
  if( !ObjectRelationalClassUtils::isTypePoolPolymorphicPointer( dataBuffer.dictionary() ) ) return false;
  ROOT::Reflex::Object ptrObj(dataBuffer.dictionary(),const_cast<void*>(data));

#if ROOT_VERSION_CODE < ROOT_VERSION(5,21,3)
  bool isNull = *static_cast<bool*>(ptrObj.Invoke("operator!").Address());
#else
  bool isNull;
  ptrObj.Invoke("operator!",isNull);
#endif

  if(!isNull){
#if ROOT_VERSION_CODE < ROOT_VERSION(5,21,3)
    std::type_info* ptrTypeInfo = *static_cast<std::type_info**>(ptrObj.Invoke("typeInfo").Address());
#else
    std::type_info* ptrTypeInfo = 0;
    ptrObj.Invoke("typeInfo",ptrTypeInfo);
#endif
    Reflex::Type ptrType = Reflex::Type::ByTypeInfo(*ptrTypeInfo);
    if( ptrType ){
      const pool::ObjectRelationalMappingElement& classMapping =  m_mappingManager.mappingForDependingClass( ptrType, m_dataBuffer->dictionary(), true );
      pool::RelationalStorageService::ObjectRelationalDataBuffer&  newDataBuffer = dataBuffer.addDependency();
      newDataBuffer.build( m_schema, classMapping, ptrType );
#if ROOT_VERSION_CODE < ROOT_VERSION(5,21,3)
      void* ptrData = ptrObj.Invoke("operator*").Address();
#else
      void* ptrData;
      ptrObj.Invoke("operator*",ptrData);
#endif
      return write(oid,newDataBuffer,ptrData,callback);
    } else {
      return false;
    }
  } 
  return true;
}

pool::RelationalStorageService::ObjectRelationalDataBuffer&
pool::RelationalStorageService::ObjectWriter::dataBuffer()
{
  return *m_dataBuffer;
}

const Reflex::Type&
pool::RelationalStorageService::ObjectWriter::refDictionary()
{
  return m_refDict;
}

std::pair<std::string,std::string>
pool::RelationalStorageService::ObjectWriter::getDataForPolyPtr(const std::string& typeName, const void* address){
  std::pair<std::string,std::string> ret(std::string(""),std::string(""));
  Reflex::Type type = Reflex::Type::ByName( typeName );
  if( !ObjectRelationalClassUtils::isTypePoolPolymorphicPointer( type ) ) {
    return ret;
  }
  ROOT::Reflex::Object ptrObj(type,const_cast<void*>(address));

#if ROOT_VERSION_CODE < ROOT_VERSION(5,21,3)
  bool isNull = *static_cast<bool*>(ptrObj.Invoke("operator!").Address());
#else
  bool isNull;
  ptrObj.Invoke("operator!",isNull);
#endif

  if(!isNull){
#if ROOT_VERSION_CODE < ROOT_VERSION(5,21,3)
    std::type_info* ptrTypeInfo = *static_cast<std::type_info**>(ptrObj.Invoke("typeInfo").Address());
#else
    std::type_info* ptrTypeInfo = 0;
    ptrObj.Invoke("typeInfo",ptrTypeInfo);
#endif
    Reflex::Type ptrType = Reflex::Type::ByTypeInfo(*ptrTypeInfo);
    if( ptrType ){
      const pool::ObjectRelationalMappingElement& classMapping =  m_mappingManager.mappingForDependingClass( ptrType, m_dataBuffer->dictionary(), true );
      ret.first = ptrType.Name(Reflex::SCOPED);
      ret.second = classMapping.tableName();
    }
    return ret;
  }
  ret.first = RelationalStorageServiceNames::polyPtrNullLabel();
  return ret;
}

pool::RelationalStorageService::MappingManager&
pool::RelationalStorageService::ObjectWriter::mappingManager(){
  return m_mappingManager;
}

pool::RelationalStorageService::MappingSequence&
pool::RelationalStorageService::ObjectWriter::mappingSequence(){
  return m_mappingSequence;
}

bool
pool::RelationalStorageService::ObjectWriter::checkData( const coral::Attribute& data ){
  return data.isValidData();
}

