#include "PVectorContainerHandler.h"
#include "PVectorContainerIteratorHandler.h"
#include "RelationalStorageServiceException.h"
#include "ObjectRelationalAccess/ObjectRelationalClassUtils.h"

pool::RelationalStorageService::PVectorContainerHandler::PVectorContainerHandler( const Reflex::Type& dictionary ):
  m_type( dictionary ),
  m_iteratorReturnType(),
  m_isAssociative( false ),
  m_collEnv(),
  m_collProxy(0),
  m_persistentSizeAttributeOffset(0),
  m_vecAttributeOffset(0)
{
  Reflex::Member privateVectorAttribute = m_type.DataMemberByName("m_vec");
  if(privateVectorAttribute){
    m_vecAttributeOffset = privateVectorAttribute.Offset();
    Reflex::Member method = privateVectorAttribute.TypeOf().MemberByName("createCollFuncTable");
    if(method){
#if ROOT_VERSION_CODE < ROOT_VERSION(5,21,3)
      m_collProxy = static_cast<Reflex::CollFuncTable*>(method.Invoke().Address());
#else
      method.Invoke(m_collProxy);
#endif
    }
  }
  Reflex::Member persistentSizeAttribute = m_type.DataMemberByName("m_persistentSize");
  if( persistentSizeAttribute ){
    m_persistentSizeAttributeOffset = persistentSizeAttribute.Offset();
  }
  
  // find the iterator return type as the member type_value of the containers
  Reflex::Type valueType = ObjectRelationalClassUtils::containerValueType( m_type );
  m_iteratorReturnType = ObjectRelationalClassUtils::resolvedType( valueType );
  
}

pool::RelationalStorageService::PVectorContainerHandler::~PVectorContainerHandler(){
  if(m_collProxy){
    delete m_collProxy;
  }
}

size_t
pool::RelationalStorageService::PVectorContainerHandler::size( const void* address )
{
  if(!m_collProxy){
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing \"createCollFuncTable\" method for the container \"" +
                                                                             m_type.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                             "PVectorContainerHandler" );
  }
#if ROOT_VERSION_CODE >= ROOT_VERSION(5,22,0)
  m_collEnv.fObject = static_cast<char*>(const_cast<void*>(address))+m_vecAttributeOffset;
#else
  m_collEnv.object = static_cast<char*>(const_cast<void*>(address))+m_vecAttributeOffset;
#endif
  size_t transientSize = *(static_cast<size_t*>(m_collProxy->size_func(&m_collEnv)));
  return transientSize;
}

size_t
pool::RelationalStorageService::PVectorContainerHandler::startElementIndex( const void* address ){
  void* persistentSizeAddress = static_cast<char*>(const_cast<void*>(address))+m_persistentSizeAttributeOffset;
  size_t persistentSize = *static_cast<size_t*>(persistentSizeAddress);
  size_t transientSize = *(static_cast<size_t*>(m_collProxy->size_func(&m_collEnv)));
  size_t startElement = 0;
  if(persistentSize < transientSize) startElement = persistentSize;
  return startElement;
}

size_t pool::RelationalStorageService::PVectorContainerHandler::persistentSize( const void* address ){
  void* persistentSizeAddress = static_cast<char*>(const_cast<void*>(address))+m_persistentSizeAttributeOffset;
  size_t persistentSize = *static_cast<size_t*>(persistentSizeAddress);
  return persistentSize;  
}

pool::RelationalStorageService::IContainerIteratorHandler*
pool::RelationalStorageService::PVectorContainerHandler::iterate( const void* address )
{
#if ROOT_VERSION_CODE >= ROOT_VERSION(5,22,0)
  m_collEnv.fObject = static_cast<char*>(const_cast<void*>(address))+m_vecAttributeOffset;
#else
  m_collEnv.object = static_cast<char*>(const_cast<void*>(address))+m_vecAttributeOffset;
#endif
  if ( ! m_iteratorReturnType ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Missing the dictionary information for the value_type member of the container \"" +m_type.Name(Reflex::SCOPED|Reflex::FINAL) + "\"","PVectorContainerHandler" );
  }
  return new pool::RelationalStorageService::PVectorContainerIteratorHandler( m_collEnv,
                                                                              m_collProxy,
                                                                              m_iteratorReturnType,
                                                                              startElementIndex(address) );
}


void
pool::RelationalStorageService::PVectorContainerHandler::appendNewElement( void* address, void* data )
{
#if ROOT_VERSION_CODE >= ROOT_VERSION(5,22,0)
  m_collEnv.fObject = static_cast<char*>(address)+m_vecAttributeOffset;
  m_collEnv.fSize = 1;
  m_collEnv.fStart = data;
#else
  m_collEnv.object = static_cast<char*>(address)+m_vecAttributeOffset;
  m_collEnv.size = 1;
  m_collEnv.start = data;
#endif
  m_collProxy->feed_func(&m_collEnv);
}

void
pool::RelationalStorageService::PVectorContainerHandler::clear( const void* address )
{
#if ROOT_VERSION_CODE >= ROOT_VERSION(5,22,0)
  m_collEnv.fObject = static_cast<char*>(const_cast<void*>(address))+m_vecAttributeOffset;
#else
  m_collEnv.object = static_cast<char*>(const_cast<void*>(address))+m_vecAttributeOffset;
#endif
  m_collProxy->clear_func(&m_collEnv);
}

Reflex::Type
pool::RelationalStorageService::PVectorContainerHandler::iteratorReturnType() const
{
  return m_iteratorReturnType;
}

void 
pool::RelationalStorageService::PVectorContainerHandler::finalize( void* address ) 
{
  size_t transSize = size( address );
  void* persistentSizeAttributeAddress = static_cast<char*>(address)+m_persistentSizeAttributeOffset;
  *static_cast<size_t*>(persistentSizeAttributeAddress) = transSize;
}
