#include "PoolDatabaseManager.h"
#include "PoolBaseContainerManager.h"
#include "PoolContainerManagerFactory.h"
#include "PoolPolymorphicContainerManager.h"
#include "MappingManager.h"
#include "MappingInformation.h"
#include "RelationalStorageServiceNames.h"
#include "RelationalStorageServiceException.h"
#include "MappingSequence.h"
#include "ObjectRelationalAccess/ObjectRelationalTableNameRationalizer.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingSchema.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingElement.h"
#include "RelationalAccess/ISessionProxy.h"
#include "RelationalAccess/ITransaction.h"
#include "RelationalAccess/ITypeConverter.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ICursor.h"
#include "RelationalAccess/ITablePrivilegeManager.h"
#include "RelationalAccess/ITableDataEditor.h"
#include "RelationalAccess/TableDescription.h"
#include "RelationalAccess/IBulkOperation.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/AttributeSpecification.h"
#include "CoralBase/MessageStream.h"
#include "StorageSvc/DbTypeInfo.h"
#include "StorageSvc/DbColumn.h"
#include "POOLCore/Token.h"
#include <sstream>
#include <memory>

#include "RVersion.h"

#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
  typedef ROOT::Reflex::Type TypeH;
#else
  typedef Reflex::Type TypeH;
#endif



namespace pool {
  namespace RelationalStorageService {
    static bool isClassPrimitive( const std::string& className )
    {
      return ( className == "bool" ||
               className == "char" ||
               className == "unsigned char" ||
               className == "short" ||
               className == "unsigned short" ||
               className == "int" ||
               className == "unsigned int" ||
               className == "long" ||
               className == "unsigned long" ||
               className == "float" ||
               className == "double" ||
               className == "long double" ||
               className == "std::string" ||
               className == "string" );
    }
  }
}


pool::RelationalStorageService::PoolDatabaseManager::PoolDatabaseManager( coral::ISessionProxy* session ):
  m_session( session ),
  m_containersUpToDate( false ),
  m_containers(),
  m_mappingSequence( 0 ),
  m_nextContainerId(0)
{
  coral::ISchema& nominalSchema = m_session->nominalSchema();
  m_mappingSequence = new pool::RelationalStorageService::MappingSequence( nominalSchema );
}


void
pool::RelationalStorageService::PoolDatabaseManager::reopen( coral::ISessionProxy* session )
{
  m_session = session;
  coral::ISchema& nominalSchema = m_session->nominalSchema();
  delete m_mappingSequence;
  m_mappingSequence = new pool::RelationalStorageService::MappingSequence( nominalSchema );
  m_containersUpToDate = false;
}


pool::RelationalStorageService::PoolDatabaseManager::~PoolDatabaseManager()
{
  delete m_mappingSequence;
  for ( std::map< std::string, pool::RelationalStorageService::PoolBaseContainerManager* >::iterator iContainer = m_containers.begin();
        iContainer != m_containers.end(); ++iContainer ){
    delete iContainer->second;
  }
}


bool
pool::RelationalStorageService::PoolDatabaseManager::existsDatabase()
{
  bool ret = false;
  try{
    ret = m_session->nominalSchema().existsTable( pool::RelationalStorageService::RelationalStorageServiceNames::topLevelDatabaseTableName() );
  } catch (const coral::Exception& e){
    coral::MessageStream log( pool::RelationalStorageService::serviceName() );
    log << coral::Error <<e.what()<<coral::MessageStream::endmsg;
    ret = false;
  }
  return ret;
}



void
pool::RelationalStorageService::PoolDatabaseManager::dropDatabase()
{
  try {
    if ( ! m_containersUpToDate ) this->refreshContainers();
    for ( std::map< std::string, pool::RelationalStorageService::PoolBaseContainerManager* >::iterator iContainer = m_containers.begin();
          iContainer != m_containers.end(); ++iContainer ) {
      pool::RelationalStorageService::PoolBaseContainerManager* containerManager = iContainer->second;
      containerManager->dropContainer();
      delete containerManager;
    }
    m_containersUpToDate = false;
    m_containers.clear();

    // Drop the tables.
    coral::ISchema& nominalSchema = m_session->nominalSchema();
    nominalSchema.dropIfExistsTable( pool::RelationalStorageService::RelationalStorageServiceNames::containerHeaderTableName() );
    nominalSchema.dropIfExistsTable( pool::RelationalStorageService::RelationalStorageServiceNames::topLevelDatabaseTableName() );
  } catch (const coral::Exception& e){
    coral::MessageStream log( pool::RelationalStorageService::serviceName() );
    log << coral::Error <<e.what()<<coral::MessageStream::endmsg;
  }    
}



void
pool::RelationalStorageService::PoolDatabaseManager::createDatabase()
{
  try {
    // Clear the internal states of the mapping objects
    if(!m_mappingSequence->existsTable()) m_mappingSequence->createTable();
    m_mappingSequence->clear();

    // Create the top level pool database table
    coral::ITypeConverter& typeConverter = m_session->typeConverter();
    std::string oldSqlTypeForString = typeConverter.sqlTypeForCppType( coral::AttributeSpecification::typeNameForType<std::string>() );
    pool::Guid guid;
    pool::Guid::create( guid );
    std::string guidRepresentation = guid.toString();
    std::ostringstream sqlStringType;
    sqlStringType << "CHAR(" << guidRepresentation.size() << ")";
    typeConverter.setSqlTypeForCppType( sqlStringType.str(),
                                        coral::AttributeSpecification::typeNameForType<std::string>() );

    coral::TableDescription descriptionDB( "CORAL" );
    descriptionDB.setName(pool::RelationalStorageService::RelationalStorageServiceNames::topLevelDatabaseTableName());
    descriptionDB.insertColumn( pool::RelationalStorageService::RelationalStorageServiceNames::guidVariableNameInTopLevelDatabaseTable(),
                                coral::AttributeSpecification::typeNameForType<std::string>() );
    descriptionDB.setNotNullConstraint( pool::RelationalStorageService::RelationalStorageServiceNames::guidVariableNameInTopLevelDatabaseTable() );
    descriptionDB.insertColumn( pool::RelationalStorageService::RelationalStorageServiceNames::numberOfContainersVariableNameInTopLevelDatabaseTable(),
                                coral::AttributeSpecification::typeNameForType<unsigned long>() );
    descriptionDB.setNotNullConstraint( pool::RelationalStorageService::RelationalStorageServiceNames::numberOfContainersVariableNameInTopLevelDatabaseTable() );

    coral::ISchema& nominalSchema = m_session->nominalSchema();
    coral::ITable& tableDB = nominalSchema.createTable( descriptionDB );
    tableDB.privilegeManager().grantToPublic( coral::ITablePrivilegeManager::Select );
    coral::AttributeList dataToInsert;
    dataToInsert.extend<std::string>(pool::RelationalStorageService::RelationalStorageServiceNames::guidVariableNameInTopLevelDatabaseTable());
    dataToInsert.extend<unsigned long>(pool::RelationalStorageService::RelationalStorageServiceNames::numberOfContainersVariableNameInTopLevelDatabaseTable());
    dataToInsert[pool::RelationalStorageService::RelationalStorageServiceNames::guidVariableNameInTopLevelDatabaseTable()].data<std::string>() =  guidRepresentation;
    dataToInsert[pool::RelationalStorageService::RelationalStorageServiceNames::numberOfContainersVariableNameInTopLevelDatabaseTable()].data<unsigned long>() = 0;
    tableDB.dataEditor().insertRow( dataToInsert );

    typeConverter.setSqlTypeForCppType( oldSqlTypeForString,
                                        coral::AttributeSpecification::typeNameForType<std::string>() );

    // Create the table holding the container headers
    coral::TableDescription descriptionCont( "CORAL" );

    descriptionCont.setName(pool::RelationalStorageService::RelationalStorageServiceNames::containerHeaderTableName());
    descriptionCont.insertColumn( pool::RelationalStorageService::RelationalStorageServiceNames::idVariableNameInContainerHeaderTable(),
                                  coral::AttributeSpecification::typeNameForType<int>() );
    descriptionCont.setNotNullConstraint( pool::RelationalStorageService::RelationalStorageServiceNames::idVariableNameInContainerHeaderTable() );
    descriptionCont.insertColumn( pool::RelationalStorageService::RelationalStorageServiceNames::containerNameVariableNameInContainerHeaderTable(),
                                  coral::AttributeSpecification::typeNameForType<std::string>());
    descriptionCont.setNotNullConstraint( pool::RelationalStorageService::RelationalStorageServiceNames::containerNameVariableNameInContainerHeaderTable() );
    descriptionCont.insertColumn( pool::RelationalStorageService::RelationalStorageServiceNames::containerTypeVariableNameInContainerHeaderTable(),
                                  coral::AttributeSpecification::typeNameForType<std::string>() );
    descriptionCont.setNotNullConstraint( pool::RelationalStorageService::RelationalStorageServiceNames::containerTypeVariableNameInContainerHeaderTable() );
    descriptionCont.insertColumn( pool::RelationalStorageService::RelationalStorageServiceNames::tableNameVariableNameInContainerHeaderTable(),
                                  coral::AttributeSpecification::typeNameForType<std::string>() );
    descriptionCont.insertColumn( pool::RelationalStorageService::RelationalStorageServiceNames::classNameVariableNameInContainerHeaderTable(),
                                  coral::AttributeSpecification::typeNameForType<std::string>() );
    descriptionCont.insertColumn( pool::RelationalStorageService::RelationalStorageServiceNames::mappingVersionVariableNameInContainerHeaderTable(),
                                  coral::AttributeSpecification::typeNameForType<std::string>() );
    descriptionCont.insertColumn( pool::RelationalStorageService::RelationalStorageServiceNames::numberOfWrittenObjectsVariableNameInContainerHeaderTable(),
                                  coral::AttributeSpecification::typeNameForType<unsigned long>() );
    descriptionCont.setNotNullConstraint( pool::RelationalStorageService::RelationalStorageServiceNames::numberOfWrittenObjectsVariableNameInContainerHeaderTable() );
    descriptionCont.insertColumn( pool::RelationalStorageService::RelationalStorageServiceNames::numberOfDeletedObjectsVariableNameInContainerHeaderTable(),
                                  coral::AttributeSpecification::typeNameForType<unsigned long>() );
    descriptionCont.setNotNullConstraint( pool::RelationalStorageService::RelationalStorageServiceNames::numberOfDeletedObjectsVariableNameInContainerHeaderTable() );
    // Define a primary key on the container id
    descriptionCont.setPrimaryKey( std::vector<std::string>( 1,
                                                             pool::RelationalStorageService::RelationalStorageServiceNames::idVariableNameInContainerHeaderTable() ) );
   // Define a unique constraint on the container name
    descriptionCont.setUniqueConstraint( pool::RelationalStorageService::RelationalStorageServiceNames::containerNameVariableNameInContainerHeaderTable() );

    nominalSchema.createTable( descriptionCont ).privilegeManager().grantToPublic( coral::ITablePrivilegeManager::Select );
  } catch (const coral::Exception& e){
    coral::MessageStream log( pool::RelationalStorageService::serviceName() );
    log << coral::Error <<e.what()<<coral::MessageStream::endmsg;
  }    
}


void
pool::RelationalStorageService::PoolDatabaseManager::refreshContainers()
{
  coral::MessageStream log( pool::RelationalStorageService::serviceName() );
  for ( std::map< std::string, pool::RelationalStorageService::PoolBaseContainerManager* >::iterator iContainer = m_containers.begin();
        iContainer != m_containers.end(); ++iContainer ) delete iContainer->second;
  m_containers.clear();

  coral::ISchema& nominalSchema = m_session->nominalSchema();
  if ( ! nominalSchema.existsTable( pool::RelationalStorageService::RelationalStorageServiceNames::containerHeaderTableName() ) ) {
    m_containersUpToDate = true;
    return;
  }
  
  // Fetch the headers from the database.
  coral::ITable& containerHeaderTable = nominalSchema.tableHandle( pool::RelationalStorageService::RelationalStorageServiceNames::containerHeaderTableName() );
  coral::IQuery* query = containerHeaderTable.newQuery();
  query->setRowCacheSize( 100 );
  query->defineOutputType( pool::RelationalStorageService::RelationalStorageServiceNames::idVariableNameInContainerHeaderTable(),
                           coral::AttributeSpecification::typeNameForType<int>());
  coral::ICursor& cursor = query->execute();
  int maxContId = 0;
  while ( cursor.next() ) {
    pool::RelationalStorageService::PoolBaseContainerManager* containerManager = 0;
    const coral::AttributeList& row = cursor.currentRow();
    int containerId = row[pool::RelationalStorageService::RelationalStorageServiceNames::idVariableNameInContainerHeaderTable()].data< int >();
    std::string containerName = row[pool::RelationalStorageService::RelationalStorageServiceNames::containerNameVariableNameInContainerHeaderTable()].data< std::string >();
    std::string containerType = row[pool::RelationalStorageService::RelationalStorageServiceNames::containerTypeVariableNameInContainerHeaderTable()].data< std::string >();

    if(maxContId < containerId) maxContId = containerId;

    // Homogeneous containers of primitives
    if ( containerType == pool::RelationalStorageService::RelationalStorageServiceNames::primitivesContainerTypeValueInContainerHeaderTable() ) {
      std::string tableNameForContainer = row[pool::RelationalStorageService::RelationalStorageServiceNames::tableNameVariableNameInContainerHeaderTable()].data< std::string >();
      containerManager = pool::RelationalStorageService::PoolContainerManagerFactory::newPrimitivesContainer( containerId,
                                                                                                              nominalSchema,
                                                                                                              tableNameForContainer );
      log << coral::Debug << "Container \""<<containerName<<"\" successfully open."<<coral::MessageStream::endmsg;
    } // Homogeneous containers
    else if ( containerType == pool::RelationalStorageService::RelationalStorageServiceNames::homogeneousContainerTypeValueInContainerHeaderTable() ) {
      std::string className = row[pool::RelationalStorageService::RelationalStorageServiceNames::classNameVariableNameInContainerHeaderTable()].data< std::string >();
      containerManager = pool::RelationalStorageService::PoolContainerManagerFactory::newHomogeneousContainer( containerId,
                                                                                                               containerName,
                                                                                                               className,
                                                                                                               nominalSchema,
                                                                                                               *m_mappingSequence);
      log << coral::Debug << "Container \""<<containerName<<"\" successfully open."<<coral::MessageStream::endmsg;
      
    } // Polymorphic
    else if ( containerType == pool::RelationalStorageService::RelationalStorageServiceNames::polymorphicContainerTypeValueInContainerHeaderTable() ) {
      std::string tableNameForContainer = row[pool::RelationalStorageService::RelationalStorageServiceNames::tableNameVariableNameInContainerHeaderTable()].data< std::string >();
      containerManager = pool::RelationalStorageService::PoolContainerManagerFactory::newPolymorphicContainer( containerId,
                                                                                                               containerName,
                                                                                                               nominalSchema,
                                                                                                               tableNameForContainer,
                                                                                                               *m_mappingSequence );
      log << coral::Debug << "Container \""<<containerName<<"\" successfully open."<<coral::MessageStream::endmsg;
    }
    if(containerManager) m_containers.insert( std::make_pair( containerName, containerManager ) );
  }
  m_nextContainerId = maxContId+1;
    
  delete query;

  m_containersUpToDate = true;
}


pool::RelationalStorageService::PoolBaseContainerManager&
pool::RelationalStorageService::PoolDatabaseManager::createHomogeneousContainer( const std::string& containerName,
                                                                                 const pool::DbTypeInfo& info )
{
  if ( ! m_containersUpToDate ) this->refreshContainers();

  pool::RelationalStorageService::PoolBaseContainerManager* containerManager = 0;

  // Get the new container id
  int containerId = m_nextContainerId;
  this->incrementNumberOfContainers();
  m_nextContainerId++;
  coral::ISchema& nominalSchema = m_session->nominalSchema();
  coral::ITable& containerHeaderTable = nominalSchema.tableHandle( pool::RelationalStorageService::RelationalStorageServiceNames::containerHeaderTableName() );

  // Retrieve the type information
  const pool::DbColumn* mainColumn = *( info.columns().begin() );
  std::string className = mainColumn->typeName();
  TypeH reflection = mainColumn->type();
  if ( reflection ) {
#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
    className = reflection.Name(ROOT::Reflex::SCOPED);
#else
    className = reflection.Name(Reflex::SCOPED);
#endif
  }


  coral::AttributeList containerHeaderRow;
  containerHeaderRow.extend<int>( pool::RelationalStorageService::RelationalStorageServiceNames::idVariableNameInContainerHeaderTable() );
  containerHeaderRow.extend<std::string>( pool::RelationalStorageService::RelationalStorageServiceNames::containerNameVariableNameInContainerHeaderTable() );
  containerHeaderRow.extend<std::string>( pool::RelationalStorageService::RelationalStorageServiceNames::containerTypeVariableNameInContainerHeaderTable() );
  containerHeaderRow.extend<std::string>( pool::RelationalStorageService::RelationalStorageServiceNames::tableNameVariableNameInContainerHeaderTable() );
  containerHeaderRow.extend<std::string>( pool::RelationalStorageService::RelationalStorageServiceNames::classNameVariableNameInContainerHeaderTable() );
  containerHeaderRow.extend<std::string>( pool::RelationalStorageService::RelationalStorageServiceNames::mappingVersionVariableNameInContainerHeaderTable() );
  containerHeaderRow.extend<unsigned long>( pool::RelationalStorageService::RelationalStorageServiceNames::numberOfWrittenObjectsVariableNameInContainerHeaderTable() );
  containerHeaderRow.extend<unsigned long>( pool::RelationalStorageService::RelationalStorageServiceNames::numberOfDeletedObjectsVariableNameInContainerHeaderTable() );
  containerHeaderRow[ pool::RelationalStorageService::RelationalStorageServiceNames::idVariableNameInContainerHeaderTable() ].data< int >() = containerId;
  containerHeaderRow[ pool::RelationalStorageService::RelationalStorageServiceNames::containerNameVariableNameInContainerHeaderTable() ].data< std::string >() = containerName;
  containerHeaderRow[ pool::RelationalStorageService::RelationalStorageServiceNames::classNameVariableNameInContainerHeaderTable() ].data< std::string >() = className;
  containerHeaderRow[ pool::RelationalStorageService::RelationalStorageServiceNames::numberOfWrittenObjectsVariableNameInContainerHeaderTable() ].data< unsigned long >() = 0;
  containerHeaderRow[ pool::RelationalStorageService::RelationalStorageServiceNames::numberOfDeletedObjectsVariableNameInContainerHeaderTable() ].data< unsigned long >() = 0;
  std::string containerType = "";


  // Check if the class is primitive
  if ( pool::RelationalStorageService::isClassPrimitive( className ) ) {

    // create the corresponding table.
    coral::TableDescription description( "CORAL" );
    description.insertColumn( pool::RelationalStorageService::RelationalStorageServiceNames::idVariableNameInPrimitivesContainerTable(),
                              coral::AttributeSpecification::typeNameForType<unsigned long>() );
    description.setNotNullConstraint( pool::RelationalStorageService::RelationalStorageServiceNames::idVariableNameInPrimitivesContainerTable() );
    std::string theClassName = className;
    if ( theClassName == "std::string" ) theClassName = coral::AttributeSpecification::typeNameForType<std::string>();
    description.insertColumn( pool::RelationalStorageService::RelationalStorageServiceNames::dataVariableNameInPrimitivesContainerTable(),
                              theClassName );
    description.setNotNullConstraint( pool::RelationalStorageService::RelationalStorageServiceNames::dataVariableNameInPrimitivesContainerTable() );
    description.setPrimaryKey( std::vector<std::string>( 1,
                                                         pool::RelationalStorageService::RelationalStorageServiceNames::idVariableNameInPrimitivesContainerTable() ) );

    // The table name for the container
    std::string tableNameForContainer = pool::RelationalStorageService::RelationalStorageServiceNames::tableNameForContainer( containerName );
    pool::ObjectRelationalTableNameRationalizer rationalizer( nominalSchema );
    tableNameForContainer = rationalizer.validNameForNewTable( tableNameForContainer );
    description.setName(tableNameForContainer);
    
    nominalSchema.createTable( description ).privilegeManager().grantToPublic( coral::ITablePrivilegeManager::Select );

    containerHeaderRow[ pool::RelationalStorageService::RelationalStorageServiceNames::tableNameVariableNameInContainerHeaderTable() ].data< std::string >() = tableNameForContainer;
    containerType = pool::RelationalStorageService::RelationalStorageServiceNames::primitivesContainerTypeValueInContainerHeaderTable();

    // Create the container manager object
    containerManager = pool::RelationalStorageService::PoolContainerManagerFactory::newPrimitivesContainer( containerId,
                                                                                                            nominalSchema,
                                                                                                            tableNameForContainer );
  }
  else {
  
    if ( ! reflection ) {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not find the dictionary information for the class \"" + className + "\"",
                                                                               "PoolDatabaseManager::createHomogeneousContainer" );
    }

    // Create the container manager object
    containerManager = pool::RelationalStorageService::PoolContainerManagerFactory::newHomogeneousContainer( containerId, containerName, className, nominalSchema, *m_mappingSequence );
    containerManager->createContainer();
    

    containerHeaderRow[ pool::RelationalStorageService::RelationalStorageServiceNames::tableNameVariableNameInContainerHeaderTable() ].data< std::string >() =  containerManager->tableName();
    containerHeaderRow[ pool::RelationalStorageService::RelationalStorageServiceNames::mappingVersionVariableNameInContainerHeaderTable() ].data< std::string >() = containerManager->mappingVersion();
    containerType = pool::RelationalStorageService::RelationalStorageServiceNames::homogeneousContainerTypeValueInContainerHeaderTable();
    
  }

  coral::MessageStream log( pool::RelationalStorageService::serviceName() );
  log << coral::Debug << "Cont header: table=\""<<containerHeaderRow[pool::RelationalStorageService::RelationalStorageServiceNames::tableNameVariableNameInContainerHeaderTable() ].data< std::string >()<<
    "\" name=\""<<containerHeaderRow[ pool::RelationalStorageService::RelationalStorageServiceNames::containerNameVariableNameInContainerHeaderTable() ].data< std::string >()<<
    "\" mapping=\""<<containerHeaderRow[ pool::RelationalStorageService::RelationalStorageServiceNames::mappingVersionVariableNameInContainerHeaderTable() ].data< std::string >()<<
    "\" id=\""<< containerHeaderRow[ pool::RelationalStorageService::RelationalStorageServiceNames::idVariableNameInContainerHeaderTable() ].data< int >()<<"\" "<<coral::MessageStream::endmsg;
  
  containerHeaderRow[ pool::RelationalStorageService::RelationalStorageServiceNames::containerTypeVariableNameInContainerHeaderTable() ].data< std::string >() = containerType;

  // Insert the container row in the container header table
  containerHeaderTable.dataEditor().insertRow( containerHeaderRow );  
  m_containers.insert( std::make_pair( containerName, containerManager ) );
  
  return *containerManager;
}


pool::RelationalStorageService::PoolBaseContainerManager&
pool::RelationalStorageService::PoolDatabaseManager::createPolymorphicContainer( const std::string& containerName )
{
  if ( ! m_containersUpToDate ) this->refreshContainers();
  coral::ISchema& nominalSchema = m_session->nominalSchema();
  coral::ITable& containerHeaderTable = nominalSchema.tableHandle( pool::RelationalStorageService::RelationalStorageServiceNames::containerHeaderTableName() );

  pool::RelationalStorageService::PoolBaseContainerManager* containerManager = 0;

  // Find a valid name for the container table
  //int containerId = this->incrementNumberOfContainers();
  int containerId =   m_nextContainerId;
  this->incrementNumberOfContainers();
  m_nextContainerId++;

  std::string containerType = pool::RelationalStorageService::RelationalStorageServiceNames::polymorphicContainerTypeValueInContainerHeaderTable();

  // Table name for the container
  std::string tableNameForContainer = pool::RelationalStorageService::RelationalStorageServiceNames::tableNameForContainer( containerName );
  pool::ObjectRelationalTableNameRationalizer rationalizer( nominalSchema );
  tableNameForContainer = rationalizer.validNameForNewTable( tableNameForContainer );

  // Create the container table
  coral::TableDescription description( "CORAL" );
  
  description.setName(tableNameForContainer);
  
  description.insertColumn( pool::RelationalStorageService::RelationalStorageServiceNames::oidVariableNameInPolymorphicContainerTable(),
                            coral::AttributeSpecification::typeNameForType<unsigned long>() );
  description.setNotNullConstraint( pool::RelationalStorageService::RelationalStorageServiceNames::oidVariableNameInPolymorphicContainerTable() );
  description.insertColumn( pool::RelationalStorageService::RelationalStorageServiceNames::classNameVariableNameInPolymorphicContainerTable(),
                            coral::AttributeSpecification::typeNameForType<std::string>() );
  description.setNotNullConstraint( pool::RelationalStorageService::RelationalStorageServiceNames::classNameVariableNameInPolymorphicContainerTable() );
  description.insertColumn( pool::RelationalStorageService::RelationalStorageServiceNames::mappingVersionVariableNameInPolymorphicContainerTable(),
                            coral::AttributeSpecification::typeNameForType<std::string>() );
  description.setNotNullConstraint( pool::RelationalStorageService::RelationalStorageServiceNames::mappingVersionVariableNameInPolymorphicContainerTable() );
  description.insertColumn( pool::RelationalStorageService::RelationalStorageServiceNames::targetOIDVariableNameInPolymorphicContainerTable(),
                            coral::AttributeSpecification::typeNameForType<unsigned long>() );
  description.setNotNullConstraint( pool::RelationalStorageService::RelationalStorageServiceNames::targetOIDVariableNameInPolymorphicContainerTable() );
  description.setPrimaryKey( std::vector<std::string>( 1,
                                                       pool::RelationalStorageService::RelationalStorageServiceNames::oidVariableNameInPolymorphicContainerTable() ) );

  nominalSchema.createTable( description ).privilegeManager().grantToPublic( coral::ITablePrivilegeManager::Select );

  // Create the container manager object
  containerManager = pool::RelationalStorageService::PoolContainerManagerFactory::newPolymorphicContainer( containerId,
                                                                                                           containerName,
                                                                                                           nominalSchema,
                                                                                                           tableNameForContainer,
                                                                                                           *m_mappingSequence );

  // Insert the container row in the container header table
  coral::AttributeList containerHeaderRow;
  containerHeaderRow.extend<int>( pool::RelationalStorageService::RelationalStorageServiceNames::idVariableNameInContainerHeaderTable() );
  containerHeaderRow.extend<std::string>( pool::RelationalStorageService::RelationalStorageServiceNames::containerNameVariableNameInContainerHeaderTable() );
  containerHeaderRow.extend<std::string>( pool::RelationalStorageService::RelationalStorageServiceNames::containerTypeVariableNameInContainerHeaderTable() );
  containerHeaderRow.extend<std::string>( pool::RelationalStorageService::RelationalStorageServiceNames::tableNameVariableNameInContainerHeaderTable() );
  containerHeaderRow.extend<unsigned long>( pool::RelationalStorageService::RelationalStorageServiceNames::numberOfWrittenObjectsVariableNameInContainerHeaderTable() );
  containerHeaderRow.extend<unsigned long>( pool::RelationalStorageService::RelationalStorageServiceNames::numberOfDeletedObjectsVariableNameInContainerHeaderTable() );

  containerHeaderRow[ pool::RelationalStorageService::RelationalStorageServiceNames::idVariableNameInContainerHeaderTable() ].data< int >() = containerId;
  containerHeaderRow[ pool::RelationalStorageService::RelationalStorageServiceNames::containerNameVariableNameInContainerHeaderTable() ].data< std::string >() = containerName;
  containerHeaderRow[ pool::RelationalStorageService::RelationalStorageServiceNames::containerTypeVariableNameInContainerHeaderTable() ].data< std::string >() = containerType;
  containerHeaderRow[ pool::RelationalStorageService::RelationalStorageServiceNames::tableNameVariableNameInContainerHeaderTable() ].data< std::string >() = tableNameForContainer;
  containerHeaderRow[ pool::RelationalStorageService::RelationalStorageServiceNames::numberOfWrittenObjectsVariableNameInContainerHeaderTable() ].data< unsigned long >() = 0;
  containerHeaderRow[ pool::RelationalStorageService::RelationalStorageServiceNames::numberOfDeletedObjectsVariableNameInContainerHeaderTable() ].data< unsigned long >() = 0;

  // Insert the container row in the container header table
  containerHeaderTable.dataEditor().insertRow( containerHeaderRow );

  m_containers.insert( std::make_pair( containerName, containerManager ) );
  return *containerManager;
}


bool
pool::RelationalStorageService::PoolDatabaseManager::existsContainer( const std::string& containerName )
{
  if ( ! m_containersUpToDate ) this->refreshContainers();
  return ( m_containers.find( containerName ) != m_containers.end() );
}


pool::RelationalStorageService::PoolBaseContainerManager&
pool::RelationalStorageService::PoolDatabaseManager::containerHandle( const std::string& containerName )
{
  if ( ! m_containersUpToDate ) this->refreshContainers();
  std::map< std::string, pool::RelationalStorageService::PoolBaseContainerManager* >::iterator iContainer = m_containers.find( containerName );
  if(iContainer==m_containers.end()){
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not find container named \"" + containerName + "\"",
                                                                               "PoolDatabaseManager::containerHandle" );
  }
  if(!iContainer->second){
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Container named \"" + containerName + "\" cannot be read.",
                                                                             "PoolDatabaseManager::containerHandle" );
  }
  return *( iContainer->second );
}


void
pool::RelationalStorageService::PoolDatabaseManager::freeContainerHandle( const std::string& containerName )
{
  std::map< std::string, pool::RelationalStorageService::PoolBaseContainerManager* >::iterator iContainer = m_containers.find( containerName );
  if ( iContainer != m_containers.end() ) {
    delete iContainer->second;
    m_containers.erase( iContainer );
    m_containersUpToDate = false;
  }
}

void
pool::RelationalStorageService::PoolDatabaseManager::endOfContainerTransaction( const std::string& containerName )
{
  std::map< std::string, pool::RelationalStorageService::PoolBaseContainerManager* >::iterator iContainer = m_containers.find( containerName );
  if ( iContainer != m_containers.end() ) {
    delete iContainer->second;
    m_containers.erase( iContainer );
  }
  if(m_containers.empty()) m_containersUpToDate = false;
}

unsigned long
pool::RelationalStorageService::PoolDatabaseManager::incrementNumberOfContainers()
{
  coral::ISchema& nominalSchema = m_session->nominalSchema();
  coral::ITable& dbTable = nominalSchema.tableHandle( pool::RelationalStorageService::RelationalStorageServiceNames::topLevelDatabaseTableName() );

  // Retrieve the current number of containers
  coral::IQuery* query = dbTable.newQuery();
  query->addToOutputList( pool::RelationalStorageService::RelationalStorageServiceNames::numberOfContainersVariableNameInTopLevelDatabaseTable() );
  query->defineOutputType( pool::RelationalStorageService::RelationalStorageServiceNames::numberOfContainersVariableNameInTopLevelDatabaseTable(), coral::AttributeSpecification::typeNameForType<unsigned long>());
  query->setForUpdate();
  coral::ICursor& cursor = query->execute();
  cursor.next();
  unsigned long numberOfContainers = cursor.currentRow().begin()->data<unsigned long>();
  delete query;

  // Increment this number in the database
  std::string setCLause = pool::RelationalStorageService::RelationalStorageServiceNames::numberOfContainersVariableNameInTopLevelDatabaseTable() +
    " = " + pool::RelationalStorageService::RelationalStorageServiceNames::numberOfContainersVariableNameInTopLevelDatabaseTable() + " + 1";
  dbTable.dataEditor().updateRows( setCLause, "" , coral::AttributeList() );

  return numberOfContainers;
}


void
pool::RelationalStorageService::PoolDatabaseManager::signalEndOfTransaction()
{
  for ( std::map< std::string, pool::RelationalStorageService::PoolBaseContainerManager* >::iterator iContainer = m_containers.begin();
        iContainer != m_containers.end(); ++iContainer ) delete iContainer->second;
  m_containers.clear();
  m_containersUpToDate = false;
  m_mappingSequence->clear();
}

/// Clean up the containers (deleting the homogenous and removing entries from the polimorphic) for a deleted mapping version
void
pool::RelationalStorageService::PoolDatabaseManager::dropContainer(const std::string& containerName){
  
  coral::ISchema& nominalSchema = m_session->nominalSchema();
  coral::ITable& cntTable = nominalSchema.tableHandle( pool::RelationalStorageService::RelationalStorageServiceNames::containerHeaderTableName() );
  std::auto_ptr<coral::IQuery> cntQuery(cntTable.newQuery());
  cntQuery->addToOutputList( pool::RelationalStorageService::RelationalStorageServiceNames::containerTypeVariableNameInContainerHeaderTable() );
  cntQuery->addToOutputList( pool::RelationalStorageService::RelationalStorageServiceNames::tableNameVariableNameInContainerHeaderTable() );
  std::string whereClause = pool::RelationalStorageService::RelationalStorageServiceNames::containerNameVariableNameInContainerHeaderTable()+" = :"+
    pool::RelationalStorageService::RelationalStorageServiceNames::containerNameVariableNameInContainerHeaderTable();
  coral::AttributeList whereData;
  whereData.extend<std::string>( pool::RelationalStorageService::RelationalStorageServiceNames::containerNameVariableNameInContainerHeaderTable() );
  whereData.begin()->data<std::string>() = containerName;
  cntQuery->setCondition(whereClause,whereData);
  coral::ICursor& cursor = cntQuery->execute();
  std::string containerType("");
  std::string containerTable("");
  while( cursor.next() ){
    const coral::AttributeList& row = cursor.currentRow();
    containerType = row[pool::RelationalStorageService::RelationalStorageServiceNames::containerTypeVariableNameInContainerHeaderTable()].data<std::string>();
    containerTable = row[pool::RelationalStorageService::RelationalStorageServiceNames::tableNameVariableNameInContainerHeaderTable()].data<std::string>();
  }
  if(containerType.empty()){
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not find container named \"" + containerName + "\"",
                                                                               "PoolDatabaseManager::dropContainer" );    
  }
  
  std::set<unsigned long> oidsToDelete;
  // Remove 'by hands' entries in LINK table for the deleted containers
  coral::ITable& linksTable = nominalSchema.tableHandle( "POOL_CONT___LINKS" );
  std::auto_ptr<coral::IQuery> query(linksTable.newQuery());
  query->addToOutputList( "OID" );
  query->addToOutputList( "VALUE" );
  query->defineOutputType( "OID", coral::AttributeSpecification::typeNameForType<unsigned long>() );
  query->defineOutputType( "VALUE", coral::AttributeSpecification::typeNameForType<std::string>() );
  coral::ICursor& resSet = query->execute();
  while( resSet.next() ){
    const coral::AttributeList& row = resSet.currentRow();
    unsigned long linkOid = row["OID"].data<unsigned long>();
    std::string linkString = row["VALUE"].data<std::string>();
    pool::Token* tok = new Token;
    tok->fromString(linkString);
    if( tok->contID() == containerName ) oidsToDelete.insert(linkOid);
  }
  for(std::set<unsigned long>::const_iterator iOid=oidsToDelete.begin();
      iOid != oidsToDelete.end(); iOid++){
    std::string whereClause = "OID = :OID";
    coral::AttributeList whereData;
    whereData.extend<unsigned long>( "OID" );
    whereData.begin()->data<unsigned long>() = *iOid;
    linksTable.dataEditor().deleteRows(whereClause,whereData);
  }
  
  long containersDeleted = cntTable.dataEditor().deleteRows(whereClause,whereData);
  if(containersDeleted){
    coral::ITable& dbTable = nominalSchema.tableHandle( pool::RelationalStorageService::RelationalStorageServiceNames::topLevelDatabaseTableName() );
    std::ostringstream setClause;
    setClause << pool::RelationalStorageService::RelationalStorageServiceNames::numberOfContainersVariableNameInTopLevelDatabaseTable() <<
      " = " << pool::RelationalStorageService::RelationalStorageServiceNames::numberOfContainersVariableNameInTopLevelDatabaseTable() <<
      " - " << containersDeleted;
    dbTable.dataEditor().updateRows( setClause.str(), "" , coral::AttributeList() );
  }

  if( containerType == pool::RelationalStorageService::RelationalStorageServiceNames::polymorphicContainerTypeValueInContainerHeaderTable()){
    // drop the polymorphic container table
    nominalSchema.dropTable( containerTable );
  }
  
  refreshContainers();
}




