#include "PoolHomogeneousContainerManager.h"
#include "ObjectUpdater.h"
#include "ObjectReader.h"
#include "MappingSequence.h"
#include "ExistingDataScanner.h"
#include "ObjectRelationalAccess/ObjectRelationalMapping.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingElement.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/ITableDataEditor.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ICursor.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/Exception.h"
#include "CoralBase/MessageStream.h"
#include "RelationalStorageServiceNames.h"
#include "RelationalStorageServiceException.h"

pool::RelationalStorageService::PoolHomogeneousContainerManager::PoolHomogeneousContainerManager( int containerId,
                                                                                                  const std::string& containerName,
                                                                                                  const std::string& className,
                                                                                                  coral::ISchema& schema,
                                                                                                  MappingSequence& mappingSequences ):
  pool::RelationalStorageService::PoolBaseContainerManager( containerId, pool::RelationalStorageService::RelationalStorageServiceNames::homogeneousContainerTypeValueInContainerHeaderTable(), schema ),
  m_containerName( containerName ),
  m_className( className ),
  m_dictionary( Reflex::Type::ByName(className) ),
  m_mappingManager( containerName, schema ),
  m_mappingSequences( mappingSequences ),
  m_initialized( false ),
  m_mapping( 0 ),  
  m_updater( 0 ),
  m_reader( 0 ),
  m_oidMin( 0 ),
  m_oidMax( 0 ),
  m_validOids()
{}

pool::RelationalStorageService::PoolHomogeneousContainerManager::~PoolHomogeneousContainerManager()
{
  m_validOids.clear();
  if ( m_updater ) delete m_updater;
  if ( m_reader ) delete m_reader;
}

void pool::RelationalStorageService::PoolHomogeneousContainerManager::initialize() const{
  if(!m_initialized){
    m_mapping = &m_mappingManager.mappingForClass( m_dictionary );
    m_initialized = true;
  }
}

std::string pool::RelationalStorageService::PoolHomogeneousContainerManager::className() const {
  return m_className;
}

std::string pool::RelationalStorageService::PoolHomogeneousContainerManager::mappingVersion() const {
  initialize();
  return m_mapping->version();
}

const std::string& pool::RelationalStorageService::PoolHomogeneousContainerManager::tableName() const {
  return mappingElement().tableName();
}

bool pool::RelationalStorageService::PoolHomogeneousContainerManager::createContainer(){
  if(!m_mappingSequences.existsTable()) m_mappingSequences.createTable();
  m_mapping = &m_mappingManager.mappingForClass( m_dictionary, true );
  m_initialized = true;
  return true;
}

const pool::ObjectRelationalMappingElement&
pool::RelationalStorageService::PoolHomogeneousContainerManager::mappingElement() const{
  initialize();
  if(m_mapping->className().empty()){
    throw RelationalStorageServiceException( "Mapping structure is empty.","PoolHomogeneousContainerManager::mappingElement" );
  }
  
  return m_mapping->element();
}

bool
pool::RelationalStorageService::PoolHomogeneousContainerManager::dropContainer()
{
  initialize();
  bool ret = true;
  try {
    if ( this->schema().existsTable( m_mapping->element().tableName() ) ) {
      int oid = 0;
      int oidOut = 0;
      Reflex::Type dict;
      while ( this->existsOid( oid, oidOut, dict ) ) {
        for( std::map<std::string, pool::ObjectRelationalMappingElement >::const_iterator iDep = m_mapping->dependentElements().begin();
             iDep != m_mapping->dependentElements().end(); ++iDep ){
          this->deleteAnObject( iDep->second, "", oidOut );          
        }
        this->deleteAnObject( m_mapping->element(), "", oidOut );
        m_validOids.erase( oidOut );
        oid = oidOut;
      }
    }
    ret = this->removeContainerEntry();
  } catch (const coral::Exception& e){
    coral::MessageStream log( pool::RelationalStorageService::serviceName() );
    log << coral::Error <<e.what()<<coral::MessageStream::endmsg;
    ret = false;
  }
  return ret;
}


bool
pool::RelationalStorageService::PoolHomogeneousContainerManager::writeObject( const void* data,
                                                                              int oid,
                                                                              pool::DataCallBack& callback,
                                                                              const std::string& type_name,
                                                                              const Reflex::Type& dictionary,
                                                                              pool::DbDatabase* databaseHandle )
{
  // Check on the type name
  if ( type_name != mappingElement().variableType() ) return false;
  if ( ! m_updater ) {
    m_updater = new pool::RelationalStorageService::ObjectUpdater( this->schema(),
                                                                   dictionary,
                                                                   m_mappingManager,
                                                                   m_mappingSequences,
                                                                   databaseHandle );
  }
  bool ret = true;
  try {
    this->incrementNumberOfObjectsWritten();
    ret = m_updater->write( oid, data, callback );
  } catch (const coral::Exception& e){
    coral::MessageStream log( pool::RelationalStorageService::serviceName() );
    log << coral::Error <<e.what()<<coral::MessageStream::endmsg;
    ret = false;
  }
  return ret;
}


bool
pool::RelationalStorageService::PoolHomogeneousContainerManager::updateObject( const void* data,
                                                                               int oid,
                                                                               pool::DataCallBack& callback,
                                                                               const std::string& type_name,
                                                                               const Reflex::Type& dictionary,
                                                                               pool::DbDatabase* databaseHandle )
{
  // Check on the type name
  if ( type_name != mappingElement().variableType() ) return false;
  if ( ! m_updater ) {
    m_updater = new pool::RelationalStorageService::ObjectUpdater( this->schema(),
                                                                   dictionary,
                                                                   m_mappingManager,
                                                                   m_mappingSequences,
                                                                   databaseHandle );
  }

  bool ret = true;
  try {
    ret = m_updater->update( oid, data, callback );
    //return ( ( this->deleteAnObject( m_mapping, "", oid ) ) && ( m_updater->write( oid, data, callback ) ) );
  } catch (const coral::Exception& e){
    coral::MessageStream log( pool::RelationalStorageService::serviceName() );
    log << coral::Error <<e.what()<<coral::MessageStream::endmsg;
    ret = false;
  }
  return ret;
}


bool
pool::RelationalStorageService::PoolHomogeneousContainerManager::readObject( void** data,
                                                                             int oid,
                                                                             pool::DataCallBack& callback,
                                                                             const std::string& type_name,
                                                                             const Reflex::Type& dictionary,
                                                                             pool::DbDatabase* databaseHandle )
{
  // Check on the type name
  if ( type_name != mappingElement().variableType() ) return false;
  void* objectData = *data;
  if ( ! objectData ) return false;
  if ( ! m_reader ) {
    m_reader = new pool::RelationalStorageService::ObjectReader( this->schema(),
                                                                 dictionary,
                                                                 m_mappingManager,
                                                                 databaseHandle );
  }
  bool ret = true;
  try {
    ret = m_reader->readObject( objectData, oid, false, callback );
  } catch (const coral::Exception& e){
    coral::MessageStream log( pool::RelationalStorageService::serviceName() );
    log << coral::Error <<e.what()<<coral::MessageStream::endmsg;
    ret = false;
  }
  return ret;
}


bool
pool::RelationalStorageService::PoolHomogeneousContainerManager::readAnyObject( void** data,
                                                                                int oid,
                                                                                pool::DataCallBack& callback,
                                                                                const std::string& type_name,
                                                                                const Reflex::Type& dictionary,
                                                                                pool::DbDatabase* databaseHandle )
{
  // Check on the type name
  if ( type_name != mappingElement().variableType() ) return false;
  void* objectData = *data;
  if ( ! objectData ) return false;
  if ( ! m_reader ) {
    m_reader = new pool::RelationalStorageService::ObjectReader( this->schema(),
                                                                 dictionary,
                                                                 m_mappingManager,
                                                                 databaseHandle );
  }

  bool ret = true;
  try {
    ret = m_reader->readObject( objectData, oid, true, callback );
  } catch (const coral::Exception& e){
    coral::MessageStream log( pool::RelationalStorageService::serviceName() );
    log << coral::Error <<e.what()<<coral::MessageStream::endmsg;
    ret = false;
  }
  return ret;
}


bool
pool::RelationalStorageService::PoolHomogeneousContainerManager::deleteObject( int oid )
{
  initialize();
  bool ret = true;
  try {
    // Increment the number of objects deleted
    this->incrementNumberOfObjectsDeleted();

    // delete the dependent objects
    for( std::map< std::string, pool::ObjectRelationalMappingElement >::const_iterator iDm = m_mapping->dependentElements().begin();
       iDm != m_mapping->dependentElements().end(); ++iDm ){
      coral::ITable& dependentTable = this->schema().tableHandle( iDm->second.tableName() );
      if(iDm->second.columnNames().size()<2){
        coral::MessageStream log( pool::RelationalStorageService::serviceName() );
        log << coral::Error <<"Mapping for dependent object \""<<iDm->second.variableName()<<"\" is incorrect."<<coral::MessageStream::endmsg;
        return false;
      }
      std::string oidClause = iDm->second.columnNames()[1] + " = :" + iDm->second.columnNames()[1];
      coral::AttributeList oidData;
      oidData.extend<int>( iDm->second.columnNames()[1]);
      oidData.begin()->data<int>() = oid;
      dependentTable.dataEditor().deleteRows( oidClause, oidData );
    }
    
    if ( this->deleteAnObject( mappingElement(), "", oid ) ) {
      m_validOids.erase( oid );
    } else {
      ret = false;
    }
  } catch (const coral::Exception& e){
    coral::MessageStream log( pool::RelationalStorageService::serviceName() );
    log << coral::Error <<e.what()<<coral::MessageStream::endmsg;
    ret = false;
  }
  return ret;
}


bool
pool::RelationalStorageService::PoolHomogeneousContainerManager::deleteAnObject( const pool::ObjectRelationalMappingElement& mapping,
                                                                                 const std::string& parentTableName,
                                                                                 int oid )
{
  
  for ( pool::ObjectRelationalMappingElement::const_iterator iMapping = mapping.begin();
        iMapping != mapping.end(); ++iMapping ) {
    pool::ObjectRelationalMappingElement::ElementType elementType = iMapping->second.elementType();
    if ( elementType == pool::ObjectRelationalMappingElement::Object ||
         elementType == pool::ObjectRelationalMappingElement::PoolReference ||
         elementType == pool::ObjectRelationalMappingElement::PoolArray ||
         elementType == pool::ObjectRelationalMappingElement::Array ||
         elementType == pool::ObjectRelationalMappingElement::CArray ||
         elementType == pool::ObjectRelationalMappingElement::PoolPointer ) {
      this->deleteAnObject( iMapping->second, mapping.tableName(), oid );
    }
  }

  if ( mapping.tableName() == parentTableName ) return true;
  size_t columnId = mapping.columnNames().size() - 1;
  if ( mapping.elementType() == pool::ObjectRelationalMappingElement::PoolReference ) columnId -= 2;
  std::string oidColumn = mapping.columnNames()[ columnId ];
  coral::AttributeList condition;
  condition.extend< int >( oidColumn );
  condition.begin()->data< int >() = oid;
  this->schema().tableHandle( mapping.tableName() ).dataEditor().deleteRows( oidColumn + " = :" + oidColumn,
                                                                             condition );
  return true;
}

bool
pool::RelationalStorageService::PoolHomogeneousContainerManager::updateContent(const std::vector<std::pair<std::string,std::string> >&, bool, bool dry)
{
  bool ret = true;
  try {
    ExistingDataScanner scanner(this->schema());
    scanner.prepareScan(mappingElement());
    scanner.scanAll();
    coral::MessageStream log( pool::RelationalStorageService::serviceName() );
    log << coral::Info << "       Header: objects stored="<<size()<<
                              " ( written="<<numberOfObjectsWritten()<<
                              " deleted="<<numberOfObjectsWritten()-size()<<
                              " )" << coral::MessageStream::endmsg; 
    log << coral::Info << "       Scan on the table "<< mappingElement().tableName() <<
                              ": object stored="<< scanner.writtenObjects()-scanner.deletedObjects() <<
                              " ( written="<<scanner.writtenObjects()<<
                              " deleted="<<scanner.deletedObjects()<<
                              " )" << coral::MessageStream::endmsg;
    std::string status("UPTODATE");
    if(scanner.writtenObjects()!=numberOfObjectsWritten() ||
       scanner.deletedObjects()!=(numberOfObjectsWritten()-size())) status = "OUTOFDATE";
    log << coral::Info << "       Status=[" <<status<<"]"<<coral::MessageStream::endmsg; 
    if(!dry) {
      log << coral::Info << "       Updating container. " << coral::MessageStream::endmsg;
      ret = this->updateNumberOfObjects(scanner.writtenObjects(),scanner.deletedObjects());
    }
  } catch (const coral::Exception& e){
    coral::MessageStream log( pool::RelationalStorageService::serviceName() );
    log << coral::Error <<e.what()<<coral::MessageStream::endmsg;
    ret = false;
  }
  return ret;
}

bool
pool::RelationalStorageService::PoolHomogeneousContainerManager::existsOid( int oid,
                                                                            int& oidOut,
                                                                            Reflex::Type& dictionary )
{
  bool ret = true;
  try {
    if ( ! ( oid >= m_oidMin && oid < m_oidMax ) ) {
      m_validOids.clear();

      int intervalSize = 10000; // Need to better define this !!!
      m_oidMin = oid;

      coral::IQuery* query = this->schema().tableHandle( mappingElement().tableName() ).newQuery();
      std::string idColumn = *( mappingElement().columnNames().begin() );
      query->addToOutputList( idColumn );
      query->defineOutputType( idColumn, coral::AttributeSpecification::typeNameForType< int >() );
      query->addToOrderList( idColumn );

      coral::AttributeList oidData;
      oidData.extend< int >( idColumn );
      coral::AttributeList::iterator iData = oidData.begin();
      iData->data< int >() = m_oidMin;
      std::string whereClause = idColumn + ">= :" + idColumn;
      query->setCondition( whereClause, oidData );
      query->setRowCacheSize( intervalSize );
      query->limitReturnedRows( intervalSize, 0 );

      int value = 0;
      coral::ICursor& cursor = query->execute();
      while ( cursor.next() ) {
        value = cursor.currentRow().begin()->data< int >();
        m_validOids.insert( value );
        if ( (int)m_validOids.size() > intervalSize ) {
          break;
        }
      }
      delete query;
      m_oidMax = value + 1;
    }

    std::set< int >::const_iterator iResult = m_validOids.lower_bound( oid );
    if ( iResult == m_validOids.end() ) return false;
    oidOut = *iResult;
  
    dictionary = m_dictionary;
  } catch (const coral::Exception& e){
    coral::MessageStream log( pool::RelationalStorageService::serviceName() );
    log << coral::Error <<e.what()<<coral::MessageStream::endmsg;
    ret = false;
  }
  return ret;
}

/// Ends a transaction
void
pool::RelationalStorageService::PoolHomogeneousContainerManager::endTransaction(){
  PoolBaseContainerManager::endTransaction();
  m_mapping = 0;
}

