#ifndef STORAGESVC_DBOBJECTCALLBACK_H
#define STORAGESVC_DBOBJECTCALLBACK_H

// POOL include files
#include "StorageSvc/DataCallBack.h"
// Reflection interface
#include "Reflex/Type.h"

/*
 * POOL namespace
 */
namespace pool {

  /** 
    * @class DbObjectCallBack DbObjectCallBack.h StorageSvc/DbObjectCallBack.h
    *
    * A default implementation of a DataCallBack object performing
    * the identity transformation (tranient shape = persistent shape)
    */
  class DbObjectCallBack : public DataCallBack  {
  public:
    /// Constructor takes the transient class type
    explicit DbObjectCallBack( TypeH transientType );
    /// Destuctor
    ~DbObjectCallBack();
    /// Returns the type of the transient object
    TypeH transientType() const;
    /// copy
    DataCallBack* clone() const;
    /// Passes the pointer to the transient data buffer
    void setObject( const void* ptr );
    /// Retrieves the pointer to the transient data buffer
    const void* object() const;
    /// Passes the persistent shape
    void setShape( ShapeH shape );
    /// Retrieves the persistent shape
    ShapeH shape() const;
    /// Sets the data handler to be used
    void setHandler( pool::DataHandler* pSvc );
    /// Access optional supplied data handler    
    DataHandler* handler() const;

    /// Callback when a read/write sequence should be started
    /** @param action_type    [IN] Action type: PUT/GET
      * @param data           [IN] Reference to persistent data location (on Write)
      * @param context        [IN] Data context for subsequent calls
      *
      * @return DbStatus indicating success or failure.
      */
    DbStatus start( CallType  action_type,
                          void*     data,
                          void**    context);

    /// Callback to retrieve the absolute address of a column
    /** @param action_type    [IN] Action type: PUT/GET
      * @param col_identifier [IN] Reference to persistent column to be processed.
      * @param col_number     [IN] Number of persistent column.
      * @param context        [IN] Data context returned by "start"
      * @param data          [OUT] Reference to data location
      *
      * @return DbStatus indicating success or failure.
      *         If SKIP is returned, the column will not be processed.
      */
    DbStatus bind(CallType        action_type,
                        const DbColumn* col_identifier,
                        int             col_number,
                        void*           context,
                        void**          data );

    /// Callback when a read/write sequence should be finalized
    /** @param action    [IN] Action type: PUT/GET
      * @param context   [IN] Data context returned by "start"
      *
      * @return DbStatus indicating success or failure.
      */
    DbStatus end( CallType action, void* context );
    
    /// Callback to handle on the fly object manipulation during save/load operations.
    /** @param action_type    [IN] Action type: PUT/GET
      * @param data_type      [IN] DataType of the object to be processed
      * @param reflection     [IN] Reflection class information if availible
      * @param object_ptr     [IN] Pointer to the transient object.
      * @param data       [IN/OUT] Optional data for processing the request.
      *
      * @return DbStatus indicating success or failure.
      */
    DbStatus notify(CallType action_type,
                          int      data_type,
                          TypeH    reflection,
                          void*    object_ptr,
                          void*    data );
  protected:
    TypeH                              m_transientType;
    DataHandler*                       m_dataHandler;
    const Shape*                       m_shape;
    const void*                        m_objectData;

    // No assignment operator
    DbObjectCallBack& operator=( const DbObjectCallBack& );
    // No copy constructor
    DbObjectCallBack( const DbObjectCallBack& );
  };
}

#include "StorageSvc/DbObjectCallBack.inl"

#endif
