// $Id: DbObjectSwitch.h,v 1.13 2007/11/28 16:19:32 frankb Exp $
//====================================================================
//
//  @author      M.Frank
//====================================================================
#ifndef POOL_DBOBJECTSWITCH_H
#define POOL_DBOBJECTSWITCH_H 1

// Framework include files
#include "StorageSvc/pool.h"
#include "StorageSvc/DbType.h"

// C++ header files
#include <typeinfo>

/*
 * POOL namespace declaration
 */
namespace pool  {

  /** @class DbObjectSwitch DbObjectSwitch.h StorageSvc/DbObjectSwitch.h
    *
    * Base class definition of cast and destructor opertaration 
    * for object handles.
    *
    * @author  M.Frank
    * @version 1.0
    */
  struct DbObjectSwitch  {
    /// Definition of the destructor prototype
    typedef void (*Dtor)(void*);
    /// Internal data structure
    struct Data {
      Data(void* p, void** t, DbType* l) : from(p), to(t), typ(l) {}
      void*   from;
      void**  to;
      DbType* typ;
    } m_data;
    /// Constructor with initializing arguments
    DbObjectSwitch(void* p, void** t, DbType* l) : m_data(p, t, l)  { }
    /// Execute object cast
    virtual void*  cast() const = 0;
    /// Destruct the object
    virtual Dtor   destructor() const = 0;
    /// Type identifier of the target type for the cast operation
    virtual const  std::type_info& toType() const = 0;
    /// Type identifier of the source type for the cast operation
    virtual const  std::type_info& fromType() const = 0;

  protected:
    /// Destructor (called only by sub-classes)
    virtual ~DbObjectSwitch()   {     }
  };



  /** @class DbObjectSwitchImp_Cast  DbObjectSwitchImp.h StorageSvc/DbObjectSwitch.h
    *
    * Small helper class to implement type save cast operations for
    * object handles.
    *
    * @author  M.Frank
    * @version 1.0
    */
  template <class T> struct DbObjectSwitchImp_Cast {
    /// Cast operation if cast is at all possible
    template <class Q> static T* cast(Q* p) { return dynamic_cast<T*>(p);}
    /// Cast operation from void*
    static T* cast(void* p)                 { return p;                  }
  };

  /** @class DbObjectSwitchImp_Destructor DbObjectSwitchImp.h StorageSvc/DbObjectSwitch.h
    *
    * Small helper class to implement type save delete operations for
    * object handles.
    *
    * @author  M.Frank
    * @version 1.0
    */
  template<class T> struct DbObjectSwitchImp_Destructor {
    /// Definition of the function prototype
    typedef void (*Dtor)(void*);
    /// Destruction function
    static void _destruct(DbObject* pObj)  { ((T*)pObj)->~T();             }
    /// Access to the destructor function
    static Dtor destructor()               { return _destruct;             }
  };

#ifndef POOL_HAVE_BASE
  template<> struct DbObjectSwitchImp_Destructor<DbObject> {      
    typedef void (*Dtor)(void*);
    static Dtor destructor()               { return 0;                     }  
  };
#endif


  /** @class DbObjectSwitchImp DbObjectSwitchImp.h StorageSvc/DbObjectSwitch.h
    *
    * Implementation of cast and destructor opertaration for object handles.
    *
    * @author  M.Frank
    * @version 1.0
    */
  template <class FROM, class TO> struct DbObjectSwitchImp : public DbObjectSwitch {

    /// Constructor with initializing arguments
    DbObjectSwitchImp(const FROM* p, TO*& t, DbType& l)
     : DbObjectSwitch((void*)p,(void**)&t,&l)        {                       }
    /// Type identifier of the target type for the cast operation
    virtual const std::type_info& toType() const    { return typeid(TO);     }
    /// Type identifier of the source type for the cast operation
    virtual const std::type_info& fromType() const  { return typeid(FROM);   }
    /// Access to the destruction function
    virtual Dtor  destructor() const { return DbObjectSwitchImp_Destructor<TO>::destructor();  }
    /// Cast operation
    virtual void* cast() const { return DbObjectSwitchImp_Cast<TO>::cast((FROM*)m_data.from);  }
  };

}       // End namespace pool
#endif  // POOL_DBOBJECTSWITCH_H
