#include "StorageSvc/DbHeap.h"
#include "StorageSvc/DbObject.h"
#include "StorageSvc/DbContainer.h"
#include "StorageSvc/DbObjectGuard.h"
#include "StorageSvc/DbInstanceCount.h"

#include <memory>
#include <cassert>

#ifdef WIN32
#include <windows.h>
#else // OS wrappers around WIN32 heap functions
static void* HeapCreate (int,   int, int  ) { return 0;             }
static void* HeapAlloc  (void*, int, int  ) { return 0;             }
static void  HeapFree   (void*, int, void*) {                       }
static bool  HeapDestroy(void*)             { return true;          }
#endif

namespace pool { class DbAllocations {}; }
using namespace pool;

static size_t GUARDSIZE = 64;

static DbInstanceCount::Counter* s_count = 
  DbInstanceCount::getCounter(typeid(DbAllocations));
namespace {
  template <int i> struct DbMemory { int value() const { return i; } };
  template<int i> static inline DbInstanceCount::Counter*
    __cnt() {
      static DbInstanceCount::Counter* s_cnt = 
        DbInstanceCount::getCounter(typeid(DbMemory<i>));
      return s_cnt;
    }
  template <int i> struct __alloc {
    typedef std::allocator< char[i] > allocator;
    inline static void* alloc()  {
      static allocator all;
      __cnt<i>()->increment();
      return all.allocate(1);
    }
    inline static DbStatus free(void* p)  {
      static allocator all;
      __cnt<i>()->decrement();
      all.deallocate(p, 1);
      return Success;
    }
  };
}

inline static DbObjectGuard* GUARD(const void* ptr)  {
  return (DbObjectGuard*) (((char*)ptr)-GUARDSIZE);
}

/// Standard Constructor
DbHeap::DbHeap() : m_handle(0),m_identifier(0),m_initial(0),m_total(0) {
}

/// Standard Destructor
DbHeap::~DbHeap() {
  release();
}

/// Create the heap
DbStatus DbHeap::create(void* id, size_t init, size_t total) {
  if ( 0 == m_handle )    {
    m_handle = ::HeapCreate(0,init,total);
    if ( 0 != m_handle )    {
      m_identifier  = id;
      m_initial     = init;
      m_total       = total;
      return Success;
    }
  }
  return Error;
}

/// Release the created heap
DbStatus DbHeap::release() {
  if ( 0 != m_handle )    {
    if ( ::HeapDestroy(m_handle) )    {
      m_handle      = 0;
      m_identifier  = 0;
      m_initial     = 0;
      m_total       = 0;
      return Success;
    }
  }
  return Error;
}

/// Access the object guard size
size_t DbHeap::guardSize() {
  return GUARDSIZE;
}

/// Number of chunks currently allocated from the heap
DbContainer& DbHeap::container(const DbObject* pObj)  {
  return GUARD(pObj)->container();
}

long long int DbHeap::numInstances()    {
  return s_count->value();
}

//#define ALLOC(x) case x: return __alloc<x>::alloc();
#define ALLOC(x) case x: __cnt<x>()->increment(); return ::operator new(x);

#define __A(x) ALLOC( x##0 ) ALLOC( x##1 ) ALLOC( x##2 ) ALLOC( x##3 ) \
               ALLOC( x##4 ) ALLOC( x##5 ) ALLOC( x##6 ) ALLOC( x##7 ) \
               ALLOC( x##8 ) ALLOC( x##9 )

#define _AA(x) __A( x##0 ) __A( x##1 ) __A( x##2 ) __A( x##3 ) \
               __A( x##4 ) __A( x##5 ) __A( x##6 ) __A( x##7 ) \
               __A( x##8 ) __A( x##9 )

/// Allocate a chunk of raw memory from the local heap
void* DbHeap::allocateRawMemory(void* handle, size_t size) {
  s_count->increment();
  if ( handle ) {
    return ::HeapAlloc(handle,0,size);
  }
#if 0
  switch(size) {
    ALLOC(1)
    ALLOC(2)
    ALLOC(3)
    ALLOC(4)
    ALLOC(5)
    ALLOC(6)
    ALLOC(7)
    ALLOC(8)
    ALLOC(9)
    __A(1)
    __A(2)
    __A(3)
    __A(4)
    __A(5)
    __A(6)
    __A(7)
    __A(8)
    __A(9)
    _AA(1)
    _AA(2)
    _AA(3)
    _AA(4)
    _AA(5)
    _AA(6)
    _AA(7)
    _AA(8)
    _AA(9)
  }
#endif
  // last resort:
  return ::operator new(size);
}
#undef ALLOC
#undef __A
#undef _AA

//#define ALLOC(x) case x: return __alloc<x>::free(ptr);
#define ALLOC(x) case x: __cnt<x>()->decrement(); ::operator delete(ptr); return Success;

#define __A(x) ALLOC( x##0 ) ALLOC( x##1 ) ALLOC( x##2 ) ALLOC( x##3 ) \
               ALLOC( x##4 ) ALLOC( x##5 ) ALLOC( x##6 ) ALLOC( x##7 ) \
               ALLOC( x##8 ) ALLOC( x##9 )

#define _AA(x) __A( x##0 ) __A( x##1 ) __A( x##2 ) __A( x##3 ) \
               __A( x##4 ) __A( x##5 ) __A( x##6 ) __A( x##7 ) \
               __A( x##8 ) __A( x##9 )

/// Free a chunk of raw memory from the local heap
DbStatus DbHeap::freeRawMemory(void* handle, void* ptr, size_t /* size */) {
  s_count->decrement();
  if ( handle ) {
    ::HeapFree(handle,0,ptr);
    return Success;
  }
#if 0
  switch(size) {
    ALLOC(1)
    ALLOC(2)
    ALLOC(3)
    ALLOC(4)
    ALLOC(5)
    ALLOC(6)
    ALLOC(7)
    ALLOC(8)
    ALLOC(9)
    __A(1)
    __A(2)
    __A(3)
    __A(4)
    __A(5)
    __A(6)
    __A(7)
    __A(8)
    __A(9)
    _AA(1)
    _AA(2)
    _AA(3)
    _AA(4)
    _AA(5)
    _AA(6)
    _AA(7)
    _AA(8)
    _AA(9)
  }
#endif
  ::operator delete(ptr);
  return Success;
}
#undef ALLOC
#undef __A
#undef _AA

/// Free an object, which was initially allocated from the heap
DbStatus DbHeap::free(void* ptr, DbContainer* /* cntH */ ) {
  if ( ptr )  {
    DbObjectGuard* g = GUARD(ptr);
    size_t size = g->size();
    g->~DbObjectGuard();
    return freeRawMemory(0, GUARD(ptr), size);
  }
  return Error;
}

/// Access the container of an object
DbObject* DbHeap::allocate( size_t                    size, 
                            DbContainer*              cntH,
                            const DbLink*             lnkH,
                            DbObjectHandle<DbObject>* objH)
{
  assert(sizeof(DbObjectGuard)<=GUARDSIZE);
  char* ptr = (char*)allocateRawMemory(0, size+GUARDSIZE);
  DbObjectGuard* g = ::new(ptr) DbObjectGuard(0, size+GUARDSIZE);
  ptr += GUARDSIZE;
  DbObject* pObj = (DbObject*)ptr;
  if ( lnkH ) g->oid()       = *lnkH;
  if ( cntH ) g->container() = *cntH;
  if ( objH ) objH->_setObject(pObj);
  return pObj;
}
